/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/03/28 20:37:57 $
 */

package kaos.core.csi;

import kaos.core.csi.ontology.OntologyRepositoryService;
import kaos.core.util.Logger;
import kaos.core.util.NamespaceValidator;
import kaos.ontology.util.ConceptTranslator;
import kaos.ontology.util.SerializableOntModelImpl;
import kaos.ontology.vocabulary.ActionConcepts;

import java.io.Serializable;
import java.net.URISyntaxException;
import java.util.List;
import java.util.Map;

/**
 * This class contains information describing an Action instance. $Revision: 1.29 $
 */
public class ActionInstanceDescriptionImpl extends OntInstanceDescriptionImpl implements ActionInstanceDescription, Cloneable, Serializable
{
    private static final long serialVersionUID      = -5894865716501809678L;

    private String            _conceptMappingName   = null;
    private String            _propertyPopulatorKey = null;

    /*
     * Create an instance of the ActionInstanceDescription described by the given arguments.
     * @param actionId String containing the id of the given action.
     * @param actionClassName String containing the ontological name of the given action.
     * @param actor String containing the id of the actor performing the action.
     * @throws URISyntaxException if the argument to create an URI is syntactically incorrect .
     */
    public ActionInstanceDescriptionImpl(String actionId, String actionClassName, String actor) throws URISyntaxException
    {
        super(actionId, actionClassName);

        if (actor == null)
            throw new URISyntaxException(actor, "Actor id is null!");

        actor = NamespaceValidator.validateURI(actor);
        OntPropertyDescription performedByProperty = new OntPropertyDescriptionImpl(ActionConcepts.performedBy);
        performedByProperty.addValue(actor);

        addProperty(performedByProperty);

        _conceptMappingName = ConceptTranslator.KAOS_MAPPING;
    }

    /*
     * Create an instance of the ActionInstanceDescription described by the given arguments.
     * @param actionId String containing the id of the given action.
     * @param actionClassName String containing the ontological name of the given action.
     * @param actor String containing the id of the actor performing the action.
     * @param properties HashMap containing the ontological names and their values of properties of the given action. currently supports String, OntInstanceDescription, and
     * OntPropertyDescription values
     * @throws URISyntaxException if the argument to create an URI is syntactically incorrect .
     */
    public ActionInstanceDescriptionImpl(String actionId, String actionClassName, String actor, Map<String, Object> properties) throws URISyntaxException
    {
        super(actionId, actionClassName);

        if (actor == null)
            throw new URISyntaxException(actor, "Actor id is null!");

        if (properties == null)
            return;

        NamespaceValidator.validateURI(actor);
        OntPropertyDescription performedByProperty = new OntPropertyDescriptionImpl(ActionConcepts.performedBy);
        performedByProperty.addValue(actor);
        addProperty(performedByProperty);

        for (String propertyName : properties.keySet())
        {
            Object value = properties.get(propertyName);
            OntPropertyDescription propDesc;
            if (value instanceof OntPropertyDescription)
                propDesc = (OntPropertyDescription)value;
            else
            {
                propDesc = new OntPropertyDescriptionImpl(propertyName);
                if (value instanceof String)
                    propDesc.addValue((String)value);
                else if (value instanceof OntInstanceDescription)
                    propDesc.setValueDescForValue(this.getName(), value);
                else if (value instanceof List)
                    for (Object val : (List)value)
                        propDesc.addValue(val.toString());
                else
                {
                    Logger.getLogger(this).logMessage("Invalid value type for property: " + propertyName + "=" + value, Logger.ERROR);
                    throw new URISyntaxException(value.toString(), "Invalid value type for property: " + propertyName);
                }
            }
            
            addProperty(propDesc);
        }

        _conceptMappingName = ConceptTranslator.KAOS_MAPPING;
    }

    /*
     * Create an instance of the ActionInstanceDescription described by the given OWL ontology model.
     * @param model ontology model the action id, class name, actor id and, optionally, properties.
     * @param ontService optional reference to the ontology service to populate value descriptions from the set of loaded ontologies
     * @throws URISyntaxException if the argument to create an URI is syntactically incorrect
     */
    public ActionInstanceDescriptionImpl(SerializableOntModelImpl model, OntologyRepositoryService ontService) throws URISyntaxException
    {
        super(model, ontService);
        _conceptMappingName = ConceptTranslator.KAOS_MAPPING;
    }

    /*
     * Create an instance of the ActionInstanceDescription described by the given OWL ontology XML string.
     * @param model ontology model the action id, class name, actor id and, optionally, properties.
     * @param ontService optional reference to the ontology service to populate value descriptions from the set of loaded ontologies
     * @throws URISyntaxException if the argument to create an URI is syntactically incorrect
     */
    public ActionInstanceDescriptionImpl(String owlString, OntologyRepositoryService ontService) throws URISyntaxException
    {
        super(owlString, ontService);
        _conceptMappingName = ConceptTranslator.KAOS_MAPPING;
    }

    /*
     * Create an instance of the ActionInstanceDescription described by the given OWL ontology string.
     * @param model Ontology model the action id, class name, actor id and, optionally, properties.
     * @throws URISyntaxException if the argument to create an URI is syntactically incorrect
     */
    public ActionInstanceDescriptionImpl(String owlString) throws URISyntaxException
    {
        super(owlString);
        _conceptMappingName = ConceptTranslator.KAOS_MAPPING;
    }

    /*
     * @see kaos.core.csi.ActionInstanceDescription#getActor()
     */
    public String getActor()
    {
        return (String)getProperty(ActionConcepts.performedBy()).getValues().toArray()[0];
    }

    /*
     * @see kaos.core.csi.ActionInstanceDescription#getActionClassName()
     */
    public String getActionClassName()
    {
        return (String)getTypes().toArray()[0];
    }

    public void setConceptMappingName(String name)
    {
        _conceptMappingName = name;
    }

    /*
     * @see kaos.core.csi.ActionInstanceDescription#getConceptMappingName()
     */
    public String getConceptMappingName()
    {
        return _conceptMappingName;
    }

    /*
     * @see kaos.core.csi.ActionInstanceDescription#getPropertyPopulatorKey()
     */
    public String getPropertyPopulatorKey()
    {
        return _propertyPopulatorKey;
    }

    /*
     * @see kaos.core.csi.ActionInstanceDescription#setPropertyPopulatorKey(java.lang.String)
     */
    public void setPropertyPopulatorKey(String aKey)
    {
        _propertyPopulatorKey = aKey;
    }

    /*
     * @see kaos.core.csi.ActionInstanceDescription#toString()
     */
    public String toString()
    {
        StringBuffer buffer = new StringBuffer();
        buffer.append("\nAction Instance Description with id - " + getName());
        buffer.append("\n action type - " + getActionClassName());
        buffer.append("\n performed by - " + getActor());

        buffer.append("\n properties - \n");

        for (OntPropertyDescription currentProperty : getProperties())
        {
            if (currentProperty.getName().equals(ActionConcepts.performedBy()))
                continue;

            buffer.append(currentProperty);
        }

        return buffer.toString();
    }

    public boolean equals(Object o)
    {
        return super.equals(o);
    }

    public int hashCode()
    {
        return super.hashCode();
    }

    /*
     * Implementation of the clone method
     */
    public ActionInstanceDescription clone()
    {
        ActionInstanceDescription myClone = null;
        
        try
        {
            myClone =  new ActionInstanceDescriptionImpl(getName(), getActionClassName(), getActor());
            
            for (String currentType: this.getTypes())
                myClone.addType(currentType);
            
            for (OntPropertyDescription currentProperty: getProperties())
                myClone.addProperty((OntPropertyDescription)currentProperty.clone());
            myClone.setConceptMappingName(getConceptMappingName());
            myClone.setPropertyPopulatorKey(getPropertyPopulatorKey());
        }
        catch (URISyntaxException e)
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        return myClone;
        
    }
}