/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2012/04/27 21:41:12 $
 */
 
package kaos.core.csi;

import kaos.core.csi.registration.Property;

import javax.agent.service.transport.MessageListener;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Defines the operations to access information about a KAoS Actor.
 *
 * @author KAoS Team
 * $Revision: 1.13 $
 */
public interface KAoSActor extends MessageListener
{
    // ****** NAMED PROPERTY MANAGMENT API 
   
    /**
     * Obtain the actor's globally unique identifier.
     *
     * @return String representing the actor's guid.
     */
    public String getGUID();
        
    /**
     * Obtain the actor's name.
     *
     * @return String representing the actor's name.
     */
    public String getName();
    
    /** 
     * Returns a List of String containing the names of domains the actor is registered in.  
     * Note: When the actor first registers with the DS, this List is checked and, if not empty, the actor 
     * will be registered in the specified domains.  Subsequently, if the actor wants to register 
     * in additional domains, then the actor has to lookup the desired domain and register in it 
     * via the Domain.registerAgent() method, which will update this List after successful registration.
     *
     * @return       List of domain names contained by this KAoSActorDescription.
     */
    public List<String> getDomainNames();
  
    /**
     * Obtain the collection of ontological types of this KAoSActorDescription. 
     *
     * Presently, the types are defined by constants in kaos.ontology.vocabulary.ActorConcepts.  
     *
     * @return Map containing mappings of ontological type names -> OWL descriptions of the types or null, if ontological types have not been set.
     */
    public Map<String, String> getEntityOntologicalTypes();
    
    /**
     * Obtain the collection of just names of ontological types of KAoS entity described by this KAoSActorDescription. 
     *
     * Presently, the types are defined by constants in kaos.ontology.jena.ActorConcepts.  
     *
     * @return Enumeration containing names of ontological types or null, if there no types specified in the KAoSActorDescription.
     */
    public Set<String> getEntityOntologicalTypeNames();
        
    /**
     * Obtain guarded/unguarded status of this KAoSActor.
     *
     * @return boolean indicating whether the KAoSActor is guarded. 
     */
    public boolean isGuarded();
   
    /**
     * Obtain registered/unregistered status of this KAoSActor.
     *
     * @return boolean indicating whether the KAoSActor has been registered in the KAoS DS. 
     */
    public boolean isRegistered();
    
    /**
     * Set the  registered/unregistered status of this KAoSActor.
     *
     * @param  registered boolean indicating whether the KAoSActor has been registered in the KAoS DS. 
     */
    public void setRegistered(boolean registered);
    
    /**
     * Obtain the collection of AgentCapabilities of this KAoSActor. 
     *
     * @return List of AgentCapability describing the capabilities  of the actor.
     */
    public List<AgentCapability> getCapabilities();
    
    /** 
     * Get the name of the concept mapping between the entity system's ontological concepts and KAoS ontology. 
     *
     * @return       String specifying the name of the mapping to be used to translate ontological names from
     *               the entity system to KAoS ontologies.  The default name is "KAoS" indicating KAoS ontologies
     *               to be used, no translation necessary.
     */
    public String getConceptMappingName();
    
           
    ////    ******  GENERIC PROPERTY MANAGMENT API 
    
    /**
     * Returns a Map containing a mapping of actor properties. The key is the Property name, the value is the Property.
     * These properties are stored in the Directory Service.
     *
     * @return Map of actor's properties.
     */
    public Map<String, Property> getProperties ();
    
    /**
     * Return the property, whose name is specified
     *
     * @param name  String containing the name of the property to be returned.
     */
    public Property getProperty(String name);
    
    /**
     * Return the value of the given property
     *
     * @param property  Property, whose value is to be returned.
     *
     * @return Object   containing the given Property value.
     */
    public Object getPropertyValue(Property property);
    
    ////    ******  TRANSPORT API 
    
    /**
     * Return the List of valid Transports.
     *
     * @return List     containing the supported Transport.
     */
    public List<Transport> getTransports();
    
    /**
     * Return the Transport, whose name is specified.
     * 
     * @param name          String containing the Transport name to be retrieved.
     *
     * @return Transport    whose name was specified.
     */
    public Transport getTransport(String name);
 }
