/**
 * Last Modified by: $Author: mvignati $ On: $Date: 2012/01/26 19:30:06 $
 */

package kaos.core.csi;

import kaos.ontology.util.SerializableOntModelImpl;

import java.net.URISyntaxException;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This interface contains methods allowing to create and investigate a description of an ontology instance. $Revision: 1.10 $
 */
public interface OntInstanceDescription extends Cloneable
{
    /**
     * Add the ontology type of this instance.
     * 
     * @param name String containing the uri of the ontology type (class).
     * @throws URISyntaxException when the name is not a correct URI
     */
    public void addType(String name) throws URISyntaxException;

    /**
     * Set property for the instance. The method will remove the previous property is it exists.
     * 
     * @param newProperty OntPropertyDescription with the property description.
     */
    public void setProperty(OntPropertyDescription newProperty);

    /**
     * Set property for the instance; providing name and value. The method will remove the previous property is it exists.
     * 
     * @param newProperty string with the property name
     * @param vale string with the property vale
     * @return the created OntPropertyDescription
     */
    public OntPropertyDescription setProperty(String newProperty, String value);

    /**
     * Add property to the instance The method will will check if the property already exist and if yes combine the values. If not; just add this property.
     * 
     * @param newProperty OntPropertyDescription with the property description.
     */
    public void addProperty(OntPropertyDescription newProperty);

    /**
     * Add property to the instance; providing name and value. The method will create OntPropertyDescription and add it to the instance if it does not
     * exist. If the property exist it will combine the values
     * 
     * @param newProperty string with the property name
     * @param value string with the property value
     * @return the created OntPropertyDescription
     */
    public OntPropertyDescription addProperty(String newProperty, String value);

    /**
     * Add property to the instance; providing name, value and value description. The method will create OntPropertyDescription and add it to the instance
     * if it does not exist. If the property exist it will combine the values
     * 
     * @param newProperty string with the property name
     * @param propertyValue string with the property value
     * @param propertyValueDesc description for the value
     * @return the created OntPropertyDescription
     */
    public OntPropertyDescription addProperty(String newProperty, String propertyValue, Object propertyValueDesc);

    /**
     * Get name of the instance
     * 
     * @return String representing the name in URI format.
     */
    public String getName();

    /**
     * Get types of the instance
     * 
     * @return Set containing the strings with URIs of the types.
     */
    public Set<String> getTypes();

    /**
     * Get first type of the instance
     * 
     * @return URI of the types.
     */
    public String getFirstType();

    /**
     * Get properties of the instance
     * 
     * @return Set containing the OntPropertyDescriptions.
     */
    public Set<OntPropertyDescription> getProperties();

    /**
     * Get properties of the instance
     * 
     * @return Map containing the OntPropertyDescriptions.
     */
    public Map<String, OntPropertyDescription> getPropertiesMap();

    /**
     * Get property of the instance
     * 
     * @param propertyName name of the property.
     * @return OntPropertyDescription.
     */
    public OntPropertyDescription getProperty(String propertyName);

    /**
     * Get values for property of the instance
     * 
     * @param propertyName name of the property.
     * @return list of values
     */
    public List<String> getAllValuesForProperty(String propertyURI);

    /**
     * Get values descriptions for property of the instance
     * 
     * @param propertyName name of the property.
     * @return list of values descriptions
     */
    public List<Object> getAllValueDescriptionsForProperty(String propertyURI);

    /**
     * Get the first value for property of the instance
     * 
     * @param propertyName name of the property.
     * @return value of the property
     */
    public String getFirstValueForProperty(String propertyURI);

    /**
     * Get value description for the first value of the property of the instance
     * 
     * @param propertyName name of the property.
     * @return value description
     */
    public Object getDescriptionOfFirstValueForProperty(String propertyURI);
    
    /**
     * Get the given property if nested in the train of interrelated instances. If the property is not found in this instance, recursively search for the
     * property in OntInstanceDescriptions if nested within this one.
     * 
     * @param propertyName - URI for the property to find
     * @return - the property description
     */
    public OntPropertyDescription getNestedProperty(String propertyName);

    /**
     * Remove the ontology type of this instance.
     * 
     * @param name String containing the uri of the ontology type (class).
     */
    public void removeType(String name);

    /**
     * Get property from the instance description
     * 
     * @param propertyName property name.
     */
    public void removeProperty(String propertyName);

    /**
     * Get the Jena-based Ontology Model object for this instance.
     * 
     * @return Jena-based Ontology Model
     */
    public SerializableOntModelImpl getInstanceOntModel();

    /**
     * Get the OWL representation for this instance.
     * 
     * @return string containing the OWL representation
     */
    public String getAsOWL();

    /**
     * Clone the instance 
     * 
     * @return the new instance
     */
    public OntInstanceDescription clone();
}