/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/04/02 07:12:11 $
 */

package kaos.core.csi;

import com.hp.hpl.jena.ontology.DatatypeProperty;
import com.hp.hpl.jena.ontology.Individual;
import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.rdf.model.*;

import kaos.core.csi.ontology.OntologyRepositoryService;
import kaos.core.util.NamespaceValidator;
import kaos.ontology.util.OntModelUtils;
import kaos.ontology.util.SerializableOntModelImpl;
import kaos.ontology.vocabulary.DefaultNamespaces;
import kaos.ontology.vocabulary.OWLConcepts;
import kaos.ontology.vocabulary.RDFConcepts;
import kaos.ontology.vocabulary.RDFSConcepts;

import java.io.Serializable;
import java.net.URI;
import java.net.URISyntaxException;
import java.rmi.dgc.VMID;
import java.util.*;

/**
 * Represents an instance of an ontology class. Each property's value may have description, which can be another OntInstanceDescriptionImpl thus allowing
 * for unlimited nesting of OntInstanceDescriptionImpl's.
 * 
 * @see OntInstanceDescription $Revision: 1.47 $
 */
public class OntInstanceDescriptionImpl implements OntInstanceDescription, Cloneable, Serializable
{
    /**
     * Comment for <code>serialVersionUID</code>
     */
    private static final long                   serialVersionUID = 1L;

    private String                              instanceNameURI;
    private Set<String>                         myTypes;
    private Map<String, OntPropertyDescription> myProperties;

    /**
     * Basic constructor
     * 
     * @param instanceName - instance name
     * @param className - initial class of instance
     * @throws URISyntaxException when the class or instance name are not URI
     */
    public OntInstanceDescriptionImpl(String instanceName, String className) throws URISyntaxException
    {
        super();

        if (instanceName == null)
            instanceName = DefaultNamespaces.defaultKAoSInstancesNamespace + (new VMID()).toString();

        if (className == null)
            throw new URISyntaxException(className, "Class name is null!");

        // Testing if correct URIs
        URI testInstanceName = new URI(instanceName);
        URI testclassName = new URI(className);

        instanceNameURI = instanceName;
        myTypes = new HashSet<String>();
        myTypes.add(className);
        myProperties = new HashMap<String, OntPropertyDescription>();
    }

    /**
     * Constructor creating an instance from OWL. Only property value descriptions (OntInstances referenced by Object Properties) defined in this OWL
     * String will be included. Note: the OWL may only contain a single 'root' instance definition in other words, only one instance that is not referenced
     * as a value description If your owl defines multiple 'root' instances, then use the SerializedOntModelImpl constructor and set the
     * SerializedOntModelImpl.ontologyStructureId to the one instance you want this class to represent.
     * 
     * @param owlString - String containing valid OWL that defines an ontology instance
     * @throws URISyntaxException
     */
    public OntInstanceDescriptionImpl(String owlString) throws URISyntaxException
    {
        this(owlString, (OntologyRepositoryService)null);
    }

    /**
     * Constructor creating an instance from OWL. Property value descriptions (OntInstances referenced by Object Properties) not defined in the OWL file
     * will be looked-up in the OntologyRepositoryService
     * 
     * @param owlString - String containing valid OWL that defines an ontology instance
     * @param myOntologyRepositoryService - Ontology service used to lookup the value descriptions not defined in the owlString, may be null in which case the
     *            OntologyRepositoryService lookup is skipped
     * @throws URISyntaxException
     */
    public OntInstanceDescriptionImpl(String owlString, OntologyRepositoryService myOntologyRepositoryService) throws URISyntaxException
    {
        super();
        SerializableOntModelImpl instanceModel = OntModelUtils.constructOntModelForRootInstance(owlString);
        initFromOntModel(instanceModel, myOntologyRepositoryService, true, null);
    }

    /**
     * Constructor creating an instance based on ontology model containing OWL defining this instance. This only looks in the provided ontology for
     * property value descriptions.
     * 
     * @param instanceModel - Ontology model wrapper with the id of instance you want in the ontologyStructureId (see
     *            SerializableOntModelImpl.setOntologyStructureId(String)
     */
    public OntInstanceDescriptionImpl(SerializableOntModelImpl instanceModel) throws URISyntaxException
    {
        super();

        initFromOntModel(instanceModel, null, true, null);
    }

    /**
     * Constructor creating an instance based on ontology model containing OWL defining this instance. This only looks in the provided ontology for
     * property value descriptions.
     * 
     * @param instanceModel - Ontology model wrapper with the id of instance you want in the ontologyStructureId (see
     *            SerializableOntModelImpl.setOntologyStructureId(String)
     * @param findValueDescriptions - indicates if to parse descriptions of the property values
     */
    public OntInstanceDescriptionImpl(SerializableOntModelImpl instanceModel, boolean findValueDescriptions) throws URISyntaxException
    {
        super();

        initFromOntModel(instanceModel, null, findValueDescriptions, null);
    }

    /**
     * Constructor creating an instance based on ontology model containing OWL defining this instance. Property value descriptions 
     * (OntInstances referenced by Object Properties) not defined in the OWL file will be looked-up in the OntologyRepositoryService
     * 
     * @param instanceModel - Ontology model wrapper with the id of instance you want in the ontologyStructureId
     * @param myOntologyRepositoryService - Ontology service used to lookup the value descriptions not defined in the owlString,
     *             may be null in which case the OntologyRepositoryService lookup is skipped
     */
    public OntInstanceDescriptionImpl(SerializableOntModelImpl instanceModel, OntologyRepositoryService myOntologyRepositoryService) 
         throws URISyntaxException
    {
        super();

        initFromOntModel(instanceModel, myOntologyRepositoryService, true, null);
    }

    /**
     * Constructor creating an instance based on ontology model containing OWL defining this instance. Property value descriptions 
     * (OntInstances referenced by Object Properties) not defined in the OWL file will be looked-up in the OntologyRepositoryService
     * 
     * @param instanceModel - Ontology model wrapper with the id of instance you want in the ontologyStructureId
     * @param myOntologyRepositoryService - Ontology service used to lookup the value descriptions not defined in the owlString, 
     *            may be null in which case the OntologyRepositoryService lookup is skipped
     */
    public OntInstanceDescriptionImpl(SerializableOntModelImpl instanceModel, OntologyRepositoryService myOntologyRepositoryService,
            Map<String, List<OntPropertyDescription>> createdValueDescs) throws URISyntaxException
    {
        super();

        initFromOntModel(instanceModel, myOntologyRepositoryService, true, createdValueDescs);
    }

    /**
     * Populates the member variables (e.g. name, superclases, properties, property value descriptions) of this instance from the instanceModel provided.
     * 
     * @param instanceModel
     * @param myOntologyRepositoryService
     * @param findValueDescriptions
     * @throws URISyntaxException
     */
    private void initFromOntModel(SerializableOntModelImpl instanceModel, OntologyRepositoryService myOntologyRepositoryService, 
            boolean findValueDescriptions, Map<String, List<OntPropertyDescription>> createdValueDescs) throws URISyntaxException
    {
        myTypes = new HashSet<String>();
        myProperties = new HashMap<String, OntPropertyDescription>();

        instanceNameURI = instanceModel.getOntologyStructureId();

        if (instanceNameURI == null)
            throw new URISyntaxException("", "Instance name is null!");

        if (instanceNameURI == "")
            throw new URISyntaxException("", "Instance name is empty!");

        if (!NamespaceValidator.validateURI(instanceNameURI).equals(instanceNameURI))
            throw new URISyntaxException("", "Instance name is not URI! " + instanceNameURI);

        // Get instance from model
        Individual instance = instanceModel.getIndividual(instanceNameURI);

        if (instance == null)
            return;// throw new URISyntaxException(instanceNameURI, "Instance name is not valid!");

        // Get types of the instance
        for (Iterator<Resource> instanceTypeIT = instance.listRDFTypes(true); instanceTypeIT.hasNext();)
        {
            Resource currentClsRes = instanceTypeIT.next();

            if (!currentClsRes.canAs(OntClass.class))
                continue;

            OntClass currentCls = (OntClass)currentClsRes.as(OntClass.class);

            if (currentCls.getURI().equals(OWLConcepts.Thing()))
                continue;

            myTypes.add(currentCls.getURI());
        }

        // Will contain unique properties of the instance
        Set<Property> mySnapshotProperties = new HashSet<Property>();

        // Read the properties of this instance
        for (StmtIterator statementIT = instance.listProperties(); statementIT.hasNext();)
        {
            Statement currentSatement = statementIT.nextStatement();
            Property currentProperty = currentSatement.getPredicate();

            mySnapshotProperties.add(currentProperty);
        }

        // Get properties values for the instance
        for (Property _currentProperty : mySnapshotProperties)
        {
            String predicateURI = _currentProperty.getURI();

            if (predicateURI.equals(RDFConcepts._type_) || predicateURI.startsWith(RDFSConcepts._range_))
                continue;
            
            OntPropertyDescription currentProperty = null;

            if (_currentProperty.canAs(DatatypeProperty.class))
            {
                DatatypeProperty currentDatatypeProperty = (DatatypeProperty)_currentProperty.as(DatatypeProperty.class);
                currentProperty = new OntPropertyDescriptionImpl(currentDatatypeProperty.getURI());

                // Add values of the datatype property
                for (NodeIterator propertyValuesIT = instanceModel.listObjectsOfProperty(instance, currentDatatypeProperty); propertyValuesIT.hasNext();)
                {
                    String currentValue = propertyValuesIT.next().toString();

                    // Remove the datatype annotation if present
                    if (currentValue.indexOf('^') >= 0)
                        currentValue = currentValue.substring(0, currentValue.indexOf('^'));

                    currentProperty.addValue(currentValue);
                }
            }
            else
            // if (s.getPredicate().canAs(ObjectProperty.class))
            {
                currentProperty = new OntPropertyDescriptionImpl(_currentProperty.getURI());

                // Add values of the object property
                for (NodeIterator propertyValuesIT = instance.listPropertyValues(_currentProperty); propertyValuesIT.hasNext();)
                {
                    RDFNode rdn = (RDFNode)propertyValuesIT.next();
                    String currentPropertyValue = "";

                    if (rdn.canAs(Individual.class))
                    {
                        Individual currentPropertyValueIndividual = (Individual)rdn.as(Individual.class);
                        currentPropertyValue = currentPropertyValueIndividual.getURI();
                    }
                    else if (rdn.asNode().isLiteral())
                        currentPropertyValue = rdn.asNode().getLiteral().getLexicalForm();
                    else
                        currentPropertyValue = null;

                    if (currentPropertyValue == null)
                        continue;

                    if (currentPropertyValue.startsWith("#"))
                        currentPropertyValue = currentPropertyValue.substring(1);

                    if (currentProperty.getValues().contains(currentPropertyValue))
                        continue;

                    // Add value to the property
                    currentProperty.addValue(currentPropertyValue);

                    // Check if to parse description of the value
                    if (!findValueDescriptions)
                        continue;

                    // Find if this value has a description
                    try
                    {
                        // First check whether value description is defined locally in this ontology
                        Resource valueDescResource = instanceModel.getOntResource(currentPropertyValue);
                        if (valueDescResource != null)
                        {
                            // instance is described in this ontology
                            if (createdValueDescs == null)
                                createdValueDescs = new HashMap<String, List<OntPropertyDescription>>();
                            
                            List<OntPropertyDescription> propsWithValueDesc = createdValueDescs.get(currentPropertyValue);
                            if (propsWithValueDesc == null)
                            {
                                propsWithValueDesc = new ArrayList<OntPropertyDescription>();
                                propsWithValueDesc.add(currentProperty);
                                createdValueDescs.put(currentPropertyValue, propsWithValueDesc);
                            }
                            else
                            {
                                // instance in cache, so it's already being defined
                                Object cachedValueDesc = propsWithValueDesc.get(0).getValueDescForValue(currentPropertyValue);
                                if (cachedValueDesc != null)
                                    // instance defined, so just use it
                                    currentProperty.setValueDescForValue(currentPropertyValue, cachedValueDesc);
                                else
                                    // instance not defined yet, so add this property to the list so it gets the description too
                                    propsWithValueDesc.add(currentProperty);
                                
                                continue;
                            }

                            String currentOntIdInModel = instanceModel.getOntologyStructureId();
                            instanceModel.setOntologyStructureId(valueDescResource.getURI());
                            OntInstanceDescription _propertyValueDesc = 
                                new OntInstanceDescriptionImpl(instanceModel, myOntologyRepositoryService, createdValueDescs);
                            instanceModel.setOntologyStructureId(currentOntIdInModel);

                            if (_propertyValueDesc != null)
                            {
                                if (_propertyValueDesc.getTypes().size() > 0)
                                {
                                    currentProperty.setValueDescForValue(currentPropertyValue, _propertyValueDesc);
                                    for (OntPropertyDescription opd : propsWithValueDesc)
                                    {
                                        if (opd.getValueDescForValue(currentPropertyValue) == null)
                                            opd.setValueDescForValue(currentPropertyValue, _propertyValueDesc);
                                    }
                                }
                            }
                        }
                        else
                        {
                            // Value description not defined locally, if a reference to an OntologyRepositoryService is provided, 
                            // lookup the value desc in the OntologyRepositoryService
                            SerializableOntModelImpl currentPropertyValueDesc = null;

                            if (myOntologyRepositoryService != null)
                            {
                                try
                                {
                                    currentPropertyValueDesc = myOntologyRepositoryService.getConstructedOntologyStructure(currentPropertyValue);
                                }
                                catch (Exception e)
                                {
                                    e.printStackTrace();
                                }

                                if (currentPropertyValueDesc != null)
                                {
                                    try
                                    {
                                        OntInstanceDescription _propertyValueDesc = new OntInstanceDescriptionImpl(currentPropertyValueDesc,
                                                myOntologyRepositoryService);

                                        if (_propertyValueDesc != null)
                                            currentProperty.setValueDescForValue(currentPropertyValue, _propertyValueDesc);
                                    }
                                    catch (URISyntaxException e)
                                    {
                                        e.printStackTrace();
                                    }
                                }
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        ex.printStackTrace();
                    }
                }
            }

            myProperties.put(currentProperty.getName(), currentProperty);
        }
    }

    public boolean equals(Object obj)
    {
        if (obj instanceof OntInstanceDescriptionImpl)
        {
            OntInstanceDescriptionImpl other = (OntInstanceDescriptionImpl)obj;
            boolean test1 = other.instanceNameURI.equals(instanceNameURI);
            boolean test2 = other.myProperties.equals(myProperties);
            boolean test3 = other.myTypes.equals(myTypes);

            return (other.instanceNameURI.equals(instanceNameURI) && other.myProperties.equals(myProperties) && other.myTypes.equals(myTypes));
        }
        return false;
    }

    public int hashCode()
    {
        return instanceNameURI.hashCode() & myProperties.hashCode() & myTypes.hashCode();
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#addProperty(kaos.core.csi.OntPropertyDescription)
     */
    public void addProperty(OntPropertyDescription newProperty)
    {
        // Check whether the property already exists
        OntPropertyDescription oldProperty = (OntPropertyDescription)myProperties.get(newProperty.getName());

        if (oldProperty == null)
            setProperty(newProperty); // No such existing property, so just set the new property
        else
            // Property already exists, so add the new values and value descriptions to the existing property
            for (String newPropValue : newProperty.getValues())
            {
                oldProperty.addValue(newPropValue);
                Object newPropValueDesc = newProperty.getValueDescForValue(newPropValue);

                if (newPropValueDesc != null)
                    // TODO: SHOULD BE ADD NOT SET
                    oldProperty.setValueDescForValue(newPropValue, newPropValueDesc);
            }
    }

    /*
     * @see kaos.core.csi.OntInstanceDescription#addProperty(String, String)
     */
    public OntPropertyDescription addProperty(String newPropertyName, String value)
    {
        OntPropertyDescription newProperty = null;
        try
        {
            newProperty = new OntPropertyDescriptionImpl(newPropertyName);
            newProperty.addValue(value);
        }
        catch (URISyntaxException e)
        {
            return null;
        }

        addProperty(newProperty);

        return newProperty;
    }

    public OntPropertyDescription addProperty(String propertyURI, String propertyValue, Object propertyValueDesc)
    {
        OntPropertyDescription propertyDesc = this.getProperty(propertyURI);

        if (propertyDesc == null)
            try
            {
                propertyDesc = new OntPropertyDescriptionImpl(propertyURI);
            }
            catch (URISyntaxException use)
            {
                use.printStackTrace();
                return null;
            }

        if (propertyValueDesc != null)
            propertyDesc.setValueDescForValue(propertyValue, propertyValueDesc);
        else
            propertyDesc.addValue(propertyValue);

        addProperty(propertyDesc);
        
        return propertyDesc;
    }

    public void setProperty(OntPropertyDescription newProperty)
    {
        myProperties.put(newProperty.getName(), newProperty);
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#addType(java.lang.String)
     */
    public void addType(String className) throws URISyntaxException
    {
        myTypes.add(className);
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#getName()
     */
    public String getName()
    {
        return instanceNameURI;
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#getTypes()
     */
    public Set<String> getTypes()
    {
        return myTypes;
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#getFirstType()
     */
    public String getFirstType()
    {
        if (myTypes.size() < 1)
            return null;

        return (String)myTypes.toArray()[0];
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#getProperties()
     */
    public Set<OntPropertyDescription> getProperties()
    {
        Set<OntPropertyDescription> myResult = new HashSet<OntPropertyDescription>();

        // Check if there are any properties
        if (myProperties.values() == null)
            return myResult;

        // Gather properties in a set
        for (Iterator<OntPropertyDescription> myPropertiesIt = myProperties.values().iterator(); myPropertiesIt.hasNext();)
            myResult.add(myPropertiesIt.next());

        return myResult;
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#getPropertiesMap()
     */
    public HashMap<String, OntPropertyDescription> getPropertiesMap()
    {
        HashMap<String, OntPropertyDescription> myResult = new HashMap<String, OntPropertyDescription>(myProperties);

        return myResult;
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#getProperty(java.lang.String)
     */
    public OntPropertyDescription getProperty(String propertyName)
    {
        return (OntPropertyDescription)myProperties.get(propertyName);
    }

    /**
     * Get the given property if nested in the path of interrelated instances. If the property is not found in this instance, recursively search for the
     * property in OntInstanceDescriptions if nested within this one.
     * 
     * @param propertyName - URI for the property to find
     * @return - the property description
     */
    public OntPropertyDescription getNestedProperty(String propertyName)
    {
        OntPropertyDescription ret = null;

        OntPropertyDescription currentProperty = getProperty(propertyName);

        if (currentProperty != null)
            return currentProperty;

        for (Iterator<OntPropertyDescription> propIT = getProperties().iterator(); propIT.hasNext();)
        {
            currentProperty = propIT.next();

            for (Iterator<String> propVIT = currentProperty.getValues().iterator(); propVIT.hasNext();)
            {
                String currentValue = propVIT.next();

                if (currentProperty.getValueDescForValue(currentValue) instanceof OntInstanceDescription)
                {
                    OntInstanceDescriptionImpl currentValueDesc = (OntInstanceDescriptionImpl)currentProperty.getValueDescForValue(currentValue);

                    ret = currentValueDesc.getNestedProperty(propertyName);

                    if (ret != null)
                        return ret;
                }
            }
        }

        return ret;
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#removeType(java.lang.String)
     */
    public void removeType(String name)
    {
        myTypes.remove(name);
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#removeProperty(java.lang.String)
     */
    public void removeProperty(String propertyName)
    {
        myProperties.remove(propertyName);
    }

    public String getFirstValueForProperty(String propertyURI)
    {
        OntPropertyDescription propertyDesc = getProperty(propertyURI);

        if (propertyDesc != null)
        {
            Set<String> values = propertyDesc.getValues();

            if (values != null)
            {
                for (Iterator<String> valuesIt = values.iterator(); valuesIt.hasNext();)
                    return valuesIt.next();
            }
        }

        return null;
    }

    public Object getDescriptionOfFirstValueForProperty(String propertyURI)
    {
        String firstValue = getFirstValueForProperty(propertyURI);

        if (firstValue == null)
            return null;

        OntPropertyDescription propertyDesc = getProperty(propertyURI);

        return propertyDesc.getValueDescForValue(firstValue);
    }

    public List<Object> getAllValueDescriptionsForProperty(String propertyURI)
    {
        OntPropertyDescription propertyDesc = getProperty(propertyURI);

        if (propertyDesc != null)
        {
            Set<String> values = propertyDesc.getValues();

            if (values != null)
            {
                List<Object> valDescs = new ArrayList<Object>(values.size());

                for (String valStr : values)
                {
                    Object valDesc = propertyDesc.getValueDescForValue(valStr);

                    if (valDesc != null)
                        valDescs.add(valDesc);
                }

                return valDescs;
            }
        }

        return null;
    }

    public List<String> getAllValuesForProperty(String propertyURI)
    {
        OntPropertyDescription propertyDesc = getProperty(propertyURI);

        if (propertyDesc != null)
        {
            Set<String> values = propertyDesc.getValues();

            if (values != null)
                return new ArrayList<String>(values);
        }

        return null;
    }

    public OntPropertyDescription setValueForProperty(String propertyURI, OntInstanceDescription propertyValue)
    {
        if (propertyValue == null)
            return null;

        return setValueForProperty(propertyURI, propertyValue.getName(), propertyValue);
    }

    @Override
    public OntPropertyDescription setProperty(String newProperty, String value)
    {
        return setValueForProperty(newProperty, value, null);
    }
    
    private OntPropertyDescription setValueForProperty(String propertyURI, String propertyValue, Object propertyValueDesc)
    {
        OntPropertyDescription propertyDesc = null;

        try
        {
            propertyDesc = new OntPropertyDescriptionImpl(propertyURI);
        }
        catch (URISyntaxException use)
        {
            use.printStackTrace();
            return null;
        }

        propertyDesc.addValue(propertyValue);

        if (propertyValueDesc != null)
            propertyDesc.setValueDescForValue(propertyValue, propertyValueDesc);

        setProperty(propertyDesc);
        
        return propertyDesc;
    }

   
    
    /**
     * Implementation of the clone method
     * 
     * @return clone of the object
     */
    public OntInstanceDescription clone()
    {
        OntInstanceDescription myClone;
        
        try
        {
            myClone = new OntInstanceDescriptionImpl(getName(), getFirstType());
            
            for (String currentType: this.getTypes())
                myClone.addType(currentType);
            
            for (OntPropertyDescription currentProperty: getProperties())
                myClone.addProperty((OntPropertyDescription)currentProperty.clone());
        }
        catch (URISyntaxException e1)
        {
         // this should never happen
            throw new InternalError(e1.toString());
        }

        return myClone;
    }
    
    public SerializableOntModelImpl getInstanceOntModel()
    {
        SerializableOntModelImpl _instanceModel = new SerializableOntModelImpl();

        OntClass _genericType = _instanceModel.createClass(OWLConcepts.Thing());

        if (myTypes.size() <= 0)
            return null;

        String baseInstanceType = (String)myTypes.toArray()[0];
        OntClass instanceType = _instanceModel.createClass(baseInstanceType);
        Individual _instance = _instanceModel.createIndividual(instanceNameURI, instanceType);

        _instanceModel.setOntologyStructureId(instanceNameURI);

        // Add additional types to the instance, if exists
        if (myTypes.size() > 1)
            for (Iterator<String> myTypesIT = myTypes.iterator(); myTypesIT.hasNext();)
            {
                String currentType = myTypesIT.next();

                if (currentType.equals(baseInstanceType))
                    continue; // Skip the first type; already added

                instanceType = _instanceModel.createClass(currentType);
                _instance.addRDFType(instanceType);
            }

        _instance.addRDFType(_genericType);

        // Add properties
        for (Iterator<String> myPropertiesIT = myProperties.keySet().iterator(); myPropertiesIT.hasNext();)
        {
            String currentPropertyName = (String)myPropertiesIT.next();
            OntPropertyDescription currentProperty = (OntPropertyDescription)myProperties.get(currentPropertyName);

            currentProperty.addPropertyToOntModel(_instanceModel, instanceNameURI);
        }

        return _instanceModel;
    }
    
    public String getAsOWL()
    {
        SerializableOntModelImpl instanceModel = getInstanceOntModel();
        
        
        return instanceModel.getAsOWL();
    }

    /**
     * Returns a string representation of this instance.
     */
    public String toString()
    {
        StringBuffer buffer = new StringBuffer();
        buffer.append("\n   Ontology Instance Description - " + getName());
        buffer.append("\n\t types - " + getTypes());
        buffer.append("\n\t properties - " + getProperties());

        return buffer.toString();
    }
}