/**
 * Last Modified by: $Author: lbunch $
 * On: $Date: 2012/04/23 17:02:05 $
 */
package kaos.core.csi;

import com.hp.hpl.jena.ontology.DatatypeProperty;
import com.hp.hpl.jena.ontology.Individual;
import com.hp.hpl.jena.ontology.ObjectProperty;
import com.hp.hpl.jena.ontology.OntClass;
import kaos.ontology.util.SerializableOntModelImpl;
import kaos.ontology.vocabulary.OWLConcepts;
import kaos.ontology.vocabulary.RDFSConcepts;

import java.io.Serializable;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.*;

/**
 * This is an implementation of the OntPropertyDescription interface 
 *
 * $Revision: 1.18 $
 */
public class OntPropertyDescriptionImpl implements OntPropertyDescription, Cloneable , Serializable
{   
    private static final long serialVersionUID = 1944343282020579753L;
    
    private String myName;
    private Set<String> myValues;
    private Map<String, Object> myValueDesc;

    /**
     * Constructor with property name
     *
     * @param name - property name
     */
    public OntPropertyDescriptionImpl(String propertyName) throws URISyntaxException
    {
        super();
        URI testURI = new URI(propertyName);

        if (testURI.getFragment() == null)
            throw new URISyntaxException(propertyName, "No fragment in the name when converted to URI", 1);

        myName = propertyName;
        myValues = new HashSet<String>();
        myValueDesc = new Hashtable<String, Object>();
    }

    /**
     * @see kaos.core.csi.OntPropertyDescription#setValue(java.lang.String)
     */
    public void setValue(String value)
    {
        myValues = new HashSet<String>();
        myValueDesc = new Hashtable<String, Object>();
        
        myValues.add(value);
    }
    
    /**
     * @see kaos.core.csi.OntPropertyDescription#addValue(java.lang.String)
     */
    public void addValue(String value)
    {
        myValues.add(value);
    }

    /**
     * @see kaos.core.csi.OntPropertyDescription#addValue(Set<java.lang.String>)
     */
    public void addMultipleValues(Set<String> values)
    {
        myValues.addAll(values);
    }

    /**
     * @see kaos.core.csi.OntPropertyDescription#setValueDescForValue(java.lang.String, java.lang.Object)
     */
    public void setValueDescForValue(String value, Object desc)
    {
        if ( ! myValues.contains(value))
            addValue(value);

        myValueDesc.put(value, desc);
    }

    /**
     * @see kaos.core.csi.OntPropertyDescription#getName()
     */
    public String getName()
    {
        return myName;
    }

    /**
     * @see kaos.core.csi.OntPropertyDescription#getValues()
     */
    public Set<String> getValues()
    {
        return myValues;
    }
    
    /**
     * @see kaos.core.csi.OntPropertyDescription#getFirstValue()
     */
    public String getFirstValue()
    {
        if (myValues.size() < 1)
            return null;
        
        return (String)myValues.toArray()[0];
    }

    /**
     * @see kaos.core.csi.OntPropertyDescription#getValueDescForValue(java.lang.String)
     */
    public Object getValueDescForValue(String value)
    {
        return myValueDesc.get(value);
    }

    /**
     * @see kaos.core.csi.OntPropertyDescription#getNestedPropertyValues(java.lang.String)
     */
    public Set<String> getNestedPropertyValues(String nestedPropertyName)
    {
        // If this is the nested property, just return the values
        if (this.getName().equals(nestedPropertyName))
            return this.getValues();
        
        // Look for the nestedPropertyName in one of the value descriptions
        for (Object valueDesc : myValueDesc.values())
            if (valueDesc instanceof OntInstanceDescriptionImpl)
            {
                OntPropertyDescription nestedProp = ((OntInstanceDescriptionImpl)valueDesc).getNestedProperty(nestedPropertyName);
                if (nestedProp != null)
                    // Found the property in the current value description
                    return nestedProp.getValues();
            }
        
        // Nested property not found in any value descriptions for this property
        return null;
    }

    /**
     * @see kaos.core.csi.OntPropertyDescription#removeValue(java.lang.String)
     */
    public void removeValue(String value)
    {
        if ( ! myValues.contains(value))
            return;

        myValueDesc.remove(value);
        myValues.remove(value);
    }

    /**
     * Clear the property value.
     *
     */
    public void removeAllValues()
    {
        myValueDesc.clear();
        myValues.clear();
    }
    
    /**
     * @see kaos.core.csi.OntPropertyDescription#removeValueDescForValue(java.lang.String)
     */
    public void removeValueDescForValue(String value)
    {
        myValues.remove(value);
    }

    /**
     * @see kaos.core.csi.OntPropertyDescription#raddPropertyToOntModel(kaos.ontology.util.SerializableOntModelImpl, java.lang.String)
     */
    public void addPropertyToOntModel(SerializableOntModelImpl ontModel, String instanceName)
    {
        if (ontModel == null)
            return;
        
        // Check if property has any value
        if (getValues().size() < 1 )
            return;

        Individual _instance = ontModel.getIndividual(instanceName);
        
        if (_instance ==  null)
            return;
        
        OntClass _genericType = ontModel.createClass(OWLConcepts.Thing());
        
        // Check if the value is not URI or if it is RDFS property
        if ((getName().startsWith(RDFSConcepts.rdfsURL)) || (getFirstValue().indexOf(':') < 0))
        {
            // Datatype property
            DatatypeProperty prop = ontModel.createDatatypeProperty(getName());

            for (Iterator<String> propertyValuesIT = getValues().iterator(); propertyValuesIT.hasNext(); )
                _instance.addProperty(prop, propertyValuesIT.next());
        }
        else
        {
            // Object property
            ObjectProperty prop = ontModel.createObjectProperty(getName());

            for (Iterator<String> propertyValuesIT = getValues().iterator(); propertyValuesIT.hasNext(); )
            {
                String currentValue = (String)propertyValuesIT.next();

                // Check if the value is URI
                if (currentValue.indexOf(':') <= 0)
                    // If the current value is not URI; skip it
                    // TODO: notify about the error
                    continue;

                Individual val = ontModel.createIndividual(currentValue, _genericType);
                _instance.addProperty(prop, val);
                Object currentValueDesc = getValueDescForValue(currentValue);
                
                if (currentValueDesc != null && currentValueDesc instanceof OntInstanceDescription)
                {
                	OntInstanceDescription currentValueOntDesc = (OntInstanceDescription)getValueDescForValue(currentValue);
                	if (ontModel.getOntResource(currentValueOntDesc.getName()) == null)
                	{
	                    SerializableOntModelImpl currentValueDescModel = currentValueOntDesc.getInstanceOntModel();
	                    
	                    if (currentValueDescModel != null)
	                    {
	                        ontModel.add(currentValueDescModel);
	                    }
                	}
                }
            }
        }
    }
    
    public  OntPropertyDescription clone()
    {
        OntPropertyDescription myClone = null;

        try
        {
            myClone = (OntPropertyDescription)super.clone();
        }
        catch (CloneNotSupportedException e)
        {
            // this should never happen
            throw new InternalError(e.toString());
        }

        for (String currentPropertyWithDescription : myValueDesc.keySet())
        {
            Object currentPropertyDescription = myValueDesc.get(currentPropertyWithDescription);

            // TODO: currently supports only cloning of value description of class OntInstanceDescription
            if (currentPropertyDescription instanceof OntInstanceDescription)
                myClone.setValueDescForValue(currentPropertyWithDescription, ((OntInstanceDescription)currentPropertyDescription).clone());
        }

        return myClone;
    }

    public boolean equals (Object obj)
    {
        if (obj instanceof OntPropertyDescriptionImpl)
        {
            OntPropertyDescriptionImpl other = (OntPropertyDescriptionImpl) obj;

            return (other.myName.equals(myName) && other.myValues.equals(myValues) && other.myValueDesc.equals(myValueDesc));
        }

        return false;
    }

    public int hashCode()
    {
        return myName.hashCode() & myValues.hashCode() & myValueDesc.hashCode();
    }

    /**
     * Returns a String representation of this instance.
     */
    public String toString()
    {
       StringBuffer buffer = new StringBuffer();

	   buffer.append("\n Ontology Property Description for - " + getName() + "\n");

       for (String currentValue : getValues())
       {
           buffer.append("  value: " + currentValue);

           if (getValueDescForValue(currentValue) != null)
               buffer.append(" its description - " + getValueDescForValue(currentValue) + "\n");
           else
               buffer.append("\n");
       }

       return buffer.toString();
    }
}