/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2012/03/26 20:37:21 $
 */

package kaos.core.csi.extension;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.agent.service.ServiceProperties;

import kaos.core.csi.CSIFactory;
import kaos.core.csi.ontology.OntologyService;

/**
 * Default implementation of the KAoSExtensionComponent interface
 * $Revision: 1.11 $
 */
public class KAoSExtensionComponentImpl implements KAoSExtensionComponent
{
    protected String                   _componentName;
    private boolean                    _enabledStatus      = true;
    protected Map<String, Set<String>> _associatedOntTypes = new HashMap<String, Set<String>>();

    public KAoSExtensionComponentImpl()
    {
    }

    /**
     * Basic Constructor
     * 
     * @param name - name of the component
     * @param ontTypes - list of ontology concept names characterizing the component functionality
     */
    public KAoSExtensionComponentImpl(String name, List<String> ontTypes)
    {
        _componentName = name;

        for (String currentOntType : ontTypes)
            _associatedOntTypes.put(currentOntType, null);
    }

    /**
     * Simplified Constructor
     * 
     * @param name - name of the component
     * @param ontType - single ontology concept name characterizing the component functionality
     */
    public KAoSExtensionComponentImpl(String name, String ontType)
    {
        _componentName = name;

        _associatedOntTypes = new HashMap<String, Set<String>>();
        _associatedOntTypes.put(ontType, null);
    }

    /**
     * Set the name of this component.
     * 
     * @param componentName String containing the name of this component.
     */
    public void setName(String componentName)
    {
        if (_componentName == null)
            _componentName = componentName;
    }

    /**
     * Returns the associated ontological concepts for the KAoSExtensionComponent object.
     * 
     * @return List containing names of the concepts the KAoSExtensionComponent is associated with.
     */
    public Set<String> getOntologicalAttributes()
    {
        return _associatedOntTypes.keySet();
    }

    /**
     * Return the (unique) name of this KAoSExtensionComponent object.
     * 
     * @return name of this component.
     */
    public String getName()
    {
        return _componentName;
    }

    /**
     * Set 'true/false', indicating the 'enabled/disabled' status of the component.
     * 
     * @param status boolean parameter to set the 'enabled/disabled' status to
     */
    public void setEnabledStatus(boolean status)
    {
        _enabledStatus = status;
    }

    /**
     * Return 'true/false', indicating the 'enabled/disabled' status of the component.
     * 
     * @return a boolean parameter indicating the 'enabled/disabled' status of this component.
     */
    public boolean getEnabledStatus()
    {
        return _enabledStatus;
    }

    /**
     * Add an ontological attribute the component is associated with.
     * 
     * @param ontAttribute String containing the name of the ontological attribute.
     */
    public void addOntologicalAttribute(String ontAttribute)
    {
        if (_associatedOntTypes == null)
            _associatedOntTypes = new HashMap<String, Set<String>>();

        if (!_associatedOntTypes.containsKey(ontAttribute))
            _associatedOntTypes.put(ontAttribute, null);
    }

    public void setSubclassesForAttribute(String ontAttribute, Set<String> subclasses)
    {
        if (_associatedOntTypes == null)
            _associatedOntTypes = new HashMap<String, Set<String>>();

        if (!_associatedOntTypes.containsKey(ontAttribute))
            _associatedOntTypes.put(ontAttribute, subclasses);
        else
        {
            _associatedOntTypes.remove(ontAttribute);
            _associatedOntTypes.put(ontAttribute, subclasses);
        }
    }

    public Set<String> getSubclassesForAttribute(String ontAttribute)
    {
        if (_associatedOntTypes == null)
            return null;

        if (_associatedOntTypes.get(ontAttribute) != null)
            return _associatedOntTypes.get(ontAttribute);
        else
            try
            {
                OntologyService myOntologyService = CSIFactory.getOntologyService();

                if (myOntologyService != null)
                    return myOntologyService.getSubClassesOf(ontAttribute);
            }
            catch (Exception e)
            {
            }

        return null;
    }

    @Override
    public void setProperties(ServiceProperties props)
    {
        // TODO Auto-generated method stub

    }
}