/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/04/02 07:12:11 $
 */

package kaos.core.csi.extension.history;

import kaos.core.csi.ActionInstanceDescription;
import kaos.core.csi.extension.KAoSExtensionComponent;
import kaos.core.csi.extension.KAoSExtensionComponentImpl;
import kaos.core.csi.query.data.OntClassDescription;
import kaos.ontology.vocabulary.ActionConcepts;
import kaos.policy.guard.KAoSGuardPolicyProcessor;
import kaos.time.DurationDescription;

import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

/**
 * This class implements KAoSExtensionComponent interface. It adds a method signature that takes in an event class and returns the # events matching that
 * event class from its set of logged events.
 */
public abstract class EventHistoryComponentImpl extends KAoSExtensionComponentImpl implements EventHistoryComponent, KAoSExtensionComponent
{
    protected List<ActionInstanceDescription> _allLogs;

    public EventHistoryComponentImpl(String name, String ontType)
    {
        super(name, ontType);
        _allLogs = new ArrayList<ActionInstanceDescription>();
    }

    /*
     * This method is used to log an event. Please note that an event will be logged only if this component instance is enabled to log. If it is not
     * enabled, then the event will not be logged even if this method is called. Check the api documentation of KAoSExtensionComponent to enable/disable
     * logging.
     * @param event An ActionInstanceDescription object specifying the event to be logged.
     * @return none.
     */
    public void logEvent(ActionInstanceDescription event)
    {
        if (getEnabledStatus())
            _allLogs.add(event);
    }

    /*
     * This method is used to remove all logged events and clear the list.
     * @param none.
     * @return none.
     */
    public void removeAllEvents()
    {
        _allLogs = new ArrayList<ActionInstanceDescription>();
    }

    /*
     * This method is used to find the list of occurrences of actual events of the type indicated by an event class.
     * @param eventClass A BasicActionDescription object specifying the type of event class.
     * @param historyStartDuration DurationDescription specifying the starting time in which event should be counted
     * @param historyEndDuration DurationDescription specifying the ending time in which event should be counted
     * @return the events in this log that are instances of the specified event class.
     */
    public List<ActionInstanceDescription> findEventsOfType(OntClassDescription eventClass, DurationDescription historyStartDuration,
            DurationDescription historyEndDuration)
    {
        int startDuration;

        if (historyStartDuration != null)
            startDuration = historyStartDuration.getSeconds();
        else
            startDuration = 0;

        int endDuration;

        if (historyEndDuration != null)
            endDuration = historyEndDuration.getSeconds();
        else
            endDuration = 0;

        long currentTime = System.currentTimeMillis();

        // Calculate the time when the validity of event starts
        long earliestTimeOfEventValidity = currentTime - 1000 * endDuration;
        long latestTimeOfEventValidity = currentTime - 1000 * startDuration;

        Vector<ActionInstanceDescription> result = new Vector<ActionInstanceDescription>();
        for (ActionInstanceDescription currentEvent : _allLogs)
        {
            if (endDuration > 0)
            {
                if (currentEvent.getProperty(ActionConcepts.hasBeginTime()) == null)
                    continue;

                String _currentLogBeginTime = currentEvent.getProperty(ActionConcepts.hasBeginTime()).getFirstValue();

                if (_currentLogBeginTime == null)
                    continue;

                long currentLogBeginTime = (new Long(_currentLogBeginTime)).longValue();

                if (currentLogBeginTime <= earliestTimeOfEventValidity || currentLogBeginTime >= latestTimeOfEventValidity)
                    continue;
            }

            if (eventClass.matchInstanceWithClass(currentEvent, null) == KAoSGuardPolicyProcessor._FULL_MATCH_)
                result.add(currentEvent);
        }

        return result;
    }
}