package kaos.core.csi.extension.history.logging;

import kaos.core.csi.ActionInstanceDescription;
import kaos.core.csi.ActionInstanceDescriptionImpl;
import kaos.core.csi.OntPropertyDescription;
import kaos.core.csi.OntPropertyDescriptionImpl;
import kaos.core.csi.extension.history.EventHistoryComponentImpl;
import kaos.database.HistoryDatabase;
import kaos.ontology.vocabulary.ActionConcepts;
import kaos.ontology.vocabulary.PolicyActionConcepts;

import java.net.URISyntaxException;
import java.util.*;

public class HistoryMonitorImpl extends EventHistoryComponentImpl implements HistoryMonitor
{
   private Map<String, List<ActionInstanceDescription>> _actorLogs;
   private Map<String, List<ActionInstanceDescription>> _actionLogs;

   private boolean _dirty = false;

   private boolean USE_DATABASE = false;
   private HistoryDatabase historyDatabase;

   public HistoryMonitorImpl(String name, String ontType)
   {
      super(name, ontType);

      _actionLogs = new Hashtable<String, List<ActionInstanceDescription>>();
      _actorLogs = new Hashtable<String, List<ActionInstanceDescription>>();

      if (USE_DATABASE)
      {
         historyDatabase = new HistoryDatabase();

         // load in history
         ArrayList<ActionInstanceDescription> events = historyDatabase.getAllEvents();
         for (ActionInstanceDescription aid : events)
         {
            System.out.println("found history for " + aid.getActionClassName());
            _allLogs.add(aid);
            addToActionLog(aid);
            addToActorLog(aid);
         }
      }
   }

   /*
    * Log an event
    * @param event The event to be logged.
    */
   public void logEvent(ActionInstanceDescription event)
   {
      if (event.getProperty(ActionConcepts.hasBeginTime()) == null)
         addTimeToEvent(event);

      String _eventBeginTime = event.getProperty(ActionConcepts.hasBeginTime()).getFirstValue();

      if (_eventBeginTime == null)
         addTimeToEvent(event);

      long eventBeginTime = (new Long(_eventBeginTime)).longValue();

      if (eventBeginTime <= 0)
         addTimeToEvent(event);

      addToActionLog(event);

      addToActorLog(event);

      _allLogs.add(event);

      if (USE_DATABASE)
      {
         historyDatabase.addEvent(event);
      }
   }

   private void addToActorLog(ActionInstanceDescription event)
   {
      Set<String> actors = event.getProperty(ActionConcepts.performedBy()).getValues();

      if (actors != null)
      {
         for (String currentActor : actors)
         {
            List<ActionInstanceDescription> log = _actorLogs.get(currentActor);

            if (log == null)
            {
               log = new Vector<ActionInstanceDescription>();
               _actorLogs.put(currentActor, log);
            }

            log.add(event);
         }
      }
      else
      {
         // This is not a 'checkPermission' event, so seek the normal actor
         String actor = event.getActor();

         List<ActionInstanceDescription> log = _actorLogs.get(actor);

         if (log == null)
         {
            log = new Vector<ActionInstanceDescription>();
            _actorLogs.put(actor, log);
         }

         log.add(event);
      }
   }

   private void addToActionLog(ActionInstanceDescription event)
   {
      String eventActionType = event.getActionClassName();

      List<ActionInstanceDescription> log = _actionLogs.get(eventActionType);

      if (log == null)
      {
         log = new Vector<ActionInstanceDescription>();
         _actionLogs.put(eventActionType, log);
      }

      log.add(event);
   }

   private void addTimeToEvent(ActionInstanceDescription event)
   {
      OntPropertyDescription newProperty;
      try
      {
         newProperty = new OntPropertyDescriptionImpl(ActionConcepts.hasBeginTime());
      }
      catch (URISyntaxException e)
      {
         return;
      }

      newProperty.setValue((new Long(System.currentTimeMillis())).toString());
      event.addProperty(newProperty);
   }

   /*
    * Get all the events logged so far.
    * @return A Vector containing aid objects logged so far.
    */
   public List<ActionInstanceDescription> getAllLogs()
   {
      return _allLogs;
   }

   /*
    * Delete all the events logged so far.
    */
   public void deleteAllLogs()
   {
      _actionLogs = new Hashtable<String, List<ActionInstanceDescription>>();
      _actorLogs = new Hashtable<String, List<ActionInstanceDescription>>();
      _allLogs = new Vector<ActionInstanceDescription>();
   }

   public void deleteLogs(List<ActionInstanceDescription> logs)
   {
      _allLogs.removeAll(logs);

      // To indicate that we need to update each event set keyed on actors and actions to reflect this latest deletion
      _dirty = true;
   }

   /*
    * Get all events logged that contain the given action type.
    * @param actionType A String indicating the action type.
    * @return A vector containing aid objects logged for that action.
    */
   public List<ActionInstanceDescription> getLogsforActionType(String actionType)
   {
      List<ActionInstanceDescription> ret = _actionLogs.get(actionType);

      if (_dirty)
      {
         // Update the set to reflect the latest log deletion
         ret.retainAll(_allLogs);
         _actionLogs.put(actionType, ret);
      }

      return ret;
   }

   /*
    * Delete all logged events that contain the given action type.
    * @param actionType A String indicating the action type.
    * @return Vector A Vector containing deleted logs
    */
   public List<ActionInstanceDescription> deleteLogsforActionType(String actionType)
   {
      List<ActionInstanceDescription> logs = _actionLogs.remove(actionType);

      if (logs == null)
         return null;

      _allLogs.removeAll(logs);

      // Now remove those logs from each actor set as well...
      for (String currentActor : _actorLogs.keySet())
      {
         List<ActionInstanceDescription> aLogs = _actorLogs.get(currentActor);

         aLogs.removeAll(logs);
         _actorLogs.put(currentActor, aLogs);
      }

      return logs;
   }

   /*
    * Get all logged events that contain the given actor.
    * @param actorID A String indicating the actor id.
    * @return A vector containing aid objects logged for that actor.
    */
   public List<ActionInstanceDescription> getLogsforActor(String actorID)
   {
      List<ActionInstanceDescription> ret = _actorLogs.get(actorID);

      if (_dirty)
      {
         // Update the set to reflect the latest log deletion
         ret.retainAll(_allLogs);
         _actionLogs.put(actorID, ret);
      }

      return ret;
   }

   /*
    * Delete all logged events that contain the given actor.
    * @param actorID A String indicating the actor id.
    * @return Vector A Vector containing deleted logs
    */
   public List<ActionInstanceDescription> deleteLogsforActor(String actorID)
   {
      List<ActionInstanceDescription> logs = _actorLogs.remove(actorID);

      if (logs == null)
         return null;

      _allLogs.removeAll(logs);

      // Now remove those logs from each action type set as well...
      for (String currentAction : _actionLogs.keySet())
      {
         List<ActionInstanceDescription> aLogs = _actionLogs.get(currentAction);

         aLogs.removeAll(logs);
         _actionLogs.put(currentAction, aLogs);
      }

      return logs;
   }

   /*
    * This method returns the logged events specified by the given parameters. The input parameters are interpreted as follows:
    * @param actionType A String indicating the action type of requested logs. If it is null, 'all action types' would be assumed.
    * @param actorID A String indicating the actor of requested logs. If null, 'all actors' would be assumed.
    * @param startTime A long, indicating the start of time period for which logs are requested. If null, 'the begining of logs' time would be assumed.
    * @param endTime A long, indicating the end of time period for which logs are requested. If null, 'time of the latest log' would be assumed.
    * @param latest A boolean, indicating if only latest logs are requested. If it is set: a) start and end time parameters are ignored. b) If actionType is null, and actorID is
    * null, the latest log for each actor is returned. c) If actorID is null and actionType is not null the latest logs for each actor for the given actionType are returned. d) If
    * actorID is not null, the latest log for that actor is returned. If actionType also is not null, the latest log for that actor and that action is returned. If this boolean is
    * not set, log are retrieved based on preceding parameters.
    */
   public List<ActionInstanceDescription> getSpecifiedLogs(String actionType, String actorID, String startTime, String endTime, Boolean latest)
   {
      List<ActionInstanceDescription> myResult = new Vector<ActionInstanceDescription>();

      if (_allLogs.size() <= 0)

         // Return empty vector of logs
         return myResult;

      // By default we return all logs
      List<ActionInstanceDescription> finalLogs = _allLogs;

      if (!actorID.equals(""))
      {
         // Get logs for the specified actor
         List<ActionInstanceDescription> actorLogs = _actorLogs.get(actorID);

         if (actorLogs == null)
            actorLogs = new Vector<ActionInstanceDescription>();

         finalLogs = actorLogs;
      }

      if (!actionType.equals(""))
      {
         // Get logs only for the specified action types
         List<ActionInstanceDescription> actionLogs = _actionLogs.get(actionType);

         if (actionLogs == null)
            actionLogs = new Vector<ActionInstanceDescription>();

         actionLogs.retainAll(finalLogs);

         finalLogs = actionLogs;
      }

      // Check if only the latest logs should be left
      if (latest.booleanValue())
      {
         if (!actorID.equals(""))
         {
            // Filter the logs and return the latest log for each actor go from end to begin because the latest logs would be towards the end...

            Set<String> actors = new HashSet<String>();

            for (int i = (finalLogs.size() - 1); i >= 0; i--)
            {
               ActionInstanceDescription currentLog = (ActionInstanceDescriptionImpl) finalLogs.get(i);

               if (currentLog.getProperty(ActionConcepts.hasBeginTime()) == null)
                  continue;

               String currentActorIDs = currentLog.getProperty(PolicyActionConcepts.onActor()).getFirstValue();

               if (!actors.contains(currentActorIDs))
               {
                  // We don't have the latest log for this actor yet
                  actors.add(currentActorIDs);

                  myResult.add(currentLog);
               }
            }
         }
         else

            // If we reach here, actor id is non null return the latest log for that actor
            myResult.add(finalLogs.get(finalLogs.size()));

         return myResult;
      }

      // Check if the time window is specified
      if (startTime.equals("") && endTime.equals(""))

         // No time window specified; return the logs that we filtered so far
         return finalLogs;

      long start = 0, end = Long.MAX_VALUE;

      if (!startTime.equals(""))
         start = (new Long(startTime)).longValue();

      if (!endTime.equals(""))
         end = (new Long(endTime)).longValue();

      // And now filter the logs based on their time
      for (ActionInstanceDescription log : finalLogs)
      {
         if (log.getProperty(ActionConcepts.hasBeginTime()) == null)
            continue;

         String _currentLogBeginTime = log.getProperty(ActionConcepts.hasBeginTime()).getFirstValue();

         if (_currentLogBeginTime == null)
            continue;

         long currentLogBeginTime = (new Long(_currentLogBeginTime)).longValue();

         if ((currentLogBeginTime >= start) && (currentLogBeginTime <= end))
            myResult.add(log);
      }

      return myResult;
   }

   /*
    * This method calls 'getSpecifiedLogs' with its input parameters and then deletes the logs returned by that call. For method documentation, see the heading for
    * 'getSpecifiedLogs'.
    * @return Vector A Vector containing deleted logs
    */
   public List<ActionInstanceDescription> deleteSpecifiedLogs(String actionType, String actorID, String startTime, String endTime, Boolean latest)
   {
      List<ActionInstanceDescription> del = getSpecifiedLogs(actionType, actorID, startTime, endTime, latest);

      _allLogs.removeAll(del);

      // To indicate that we need to update each event set keyed on actors and actions to reflect this latest deletion
      _dirty = true;

      return del;
   }
}
