/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/04/02 07:12:11 $
 */
package kaos.core.csi.extension.state;

import kaos.core.csi.OntInstanceDescription;
import kaos.core.csi.OntPropertyDescription;
import kaos.core.csi.extension.KAoSExtensionComponentManager;
import kaos.core.csi.query.data.OntClassDescription;

import java.net.URISyntaxException;
import java.util.Collection;
import java.util.List;
import java.util.Set;

/**
 * This interface defines a manager used to track the properties of a given state for a particular entity. Methods are provided for sensors to update the
 * state and for interested parties to query the state.
 */
public interface StateManager extends KAoSExtensionComponentManager
{
    /******************************************************************************************
     * These methods allow state manager to track state via sensor inputs
     */

    /**
     * This method is used to add a new state to the database (and its properties if included)
     * 
     * @param stateInstance the description of the state instance to add
     */
    public void registerState(OntInstanceDescription stateInstance);

    /**
     * This method is used to add a new state to the database
     * 
     * @param stateId the unique identifier of the instance
     * @param stateClassName the class of the instance
     * @throws java.net.URISyntaxException is URI is invalid
     */
    public void registerState(String stateId, String stateClassName) throws URISyntaxException;

    /**
     * This method is used to remove a state instance
     * 
     * @param stateId the unique identifier of the state instance to remove
     */
    public void deregisterState(String stateId);

    /**
     * This method is used to add or update properties of a state
     * 
     * @param stateId the unique instance id
     * @param propertyName the name of the property
     * @param propertyValue the value of the property
     * @throws Exception if no state with that stateId is registered
     */
    public void updateStateProperty(String stateId, OntPropertyDescription updatedproperty) throws Exception;

    /**
     * This method is used to add or update properties of a state
     * 
     * @param stateId the unique instance id
     * @param properties map of property name -> value
     * @throws Exception if no state with that stateId is registered
     */
    public void updateStateProperties(String stateId, List<OntPropertyDescription> updatedproperties) throws Exception;
    
    /**
     * This method is used to remove a property from a state
     * 
     * @param stateId the unique identifier of the state instance
     * @param propertyName t-remove property property to remove
     * @throws Exception if no state of that stateId is registered
     */
    public void removePropertyFromState(String stateId, String propertyName) throws Exception;

    /*********************************************************************************************************
     * These methods allow others to query for state (pull)
     */

    /**
     * This method will return a list of OntInstanceDescriptions representing known state
    
     * @return a list of OntInstanceDescriptions containing state
     */
    public Collection<OntInstanceDescription> getStates();
    
    /**
     * This method will return a list of OntInstanceDescriptions that contain the state class name
     * 
     * @param stateClassName name of the class of interest
     * @return a list of OntInstanceDescriptions containing the class name of interest
     */
    public List<OntInstanceDescription> getStatesOfClass(String stateClassName);

    /**
     * This method is used to get the OntInstanceDescription for a state instance
     * 
     * @param stateId the unique identifier of the state instance
     * @return the description of the state instance
     */
    public OntInstanceDescription getStateInstance(String stateId);

    /**
     * This method is used to get the property value for a particular state
     * 
     * @param stateId the unique identifier of the state instance
     * @param propertyName the name of the property of interest
     * @return the value of the property
     * @throws Exception if state or property does not exist
     */
    public Set<String> getPropertyValueForState(String stateId, String propertyName) throws Exception;
    
    /**
     * This method check if the state class is satisfied (matched) by any of the actual state in the state manger
     * 
     * @param state the state class to by matched with existing states
     * @return true if matching state found
     */
    public boolean isStateSatisfied(OntClassDescription state);
    
    /**
     * This method is used to get the subjects with a particular state
     * 
     * @param state the state class to by matched with existing states
     * @return the set of subject possessing the given state
     */
    public Set<String> getSubjectWithState(OntClassDescription state);

    /**
     * These methods allow others to receive updates when state changes (push)
     */

    /**
     * register as a listener for all state changes
     * 
     * @param listener interface of recipient of update
     */
    public void registerAsListenerForState(StateUpdateListener listener);

    /**
     * register as a listener for a particular state
     * 
     * @param stateClass String representing state class of interest
     * @param listener interface of recipient of updates
     */
    public void registerAsListenerForStateClass(String stateClass, StateUpdateListener listener);

    /**
     * register as a listener for a particular state
     * 
     * @param stateID String representing instance id of the state of interest
     * @param listener interface of recipient of updates
     */
    public void registerAsListenerForStateInstance(String stateID, StateUpdateListener listener);

    public void registerInterest(OntClassDescription policyConditionDesc);
}
