/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/04/02 07:12:11 $
 */
package kaos.core.csi.extension.state;

import kaos.core.csi.OntInstanceDescription;
import kaos.core.csi.OntInstanceDescriptionImpl;
import kaos.core.csi.OntPropertyDescription;
import kaos.core.csi.extension.KAoSExtensionComponentManagerImpl;
import kaos.core.csi.extension.KAoSExtensionsTypes;
import kaos.core.csi.query.data.OntClassDescription;
import kaos.core.util.Logger;
import kaos.database.StateDatabase;
import kaos.ontology.vocabulary.EntityConcepts;
import kaos.policy.guard.KAoSGuardPolicyProcessor;

import java.io.File;
import java.io.FileInputStream;
import java.net.URISyntaxException;
import java.util.*;

public class StateManagerImpl extends KAoSExtensionComponentManagerImpl implements StateManager
{
    public static final String                         STATE_PROPERTY               = "kaos.state";
    private static StateManagerImpl                    __INSTANCE;
    private static boolean                             __INITIALIZED                = false;

    private boolean                                    USE_DATABASE                 = false;

    private Map<String, Map<String, String>>           mapEntitiesToStates          = new HashMap<String, Map<String, String>>();
    private Map<String, List<StateUpdateListener>>     mapStateInstancesToListeners = new HashMap<String, List<StateUpdateListener>>();
    private HashMap<String, List<StateUpdateListener>> mapStateClassesToListeners   = new HashMap<String, List<StateUpdateListener>>();
    private List<StateUpdateListener>                  stateListeners               = new ArrayList<StateUpdateListener>();
    private final Map<String, OntInstanceDescription>  mapStateIdToDescriptions     = new HashMap<String, OntInstanceDescription>();
    private List<OntClassDescription>                         statesOfInterest             = new ArrayList<OntClassDescription>();
    private final List<StateSensor>                    statesSensors                = new ArrayList<StateSensor>();

    private static Logger                              _logger                      = Logger.getLogger(StateManagerImpl.class);

    public StateManagerImpl()
    {
        super(KAoSExtensionsTypes.STATE_SENSORS);

        // load existing states into memory
        if (USE_DATABASE)
        {
            StateDatabase _stateDatabase = new StateDatabase();
            mapEntitiesToStates = _stateDatabase.getAllStates();
        }
    }

    /**
     * Retrieve an instance of StateManager.
     * 
     * @return a singleton, StateManagerImpl.
     */
    public static StateManagerImpl getInstance()
    {
        if (__INITIALIZED)
            return __INSTANCE;

        __INSTANCE = new StateManagerImpl();
        __INITIALIZED = true;

        return __INSTANCE;
    }

    public void loadConfiguration()
    {
        createStateSensors();
    }

    private void createStateSensors()
    {
        // create state monitors from configuration file
        String configFileName = System.getProperty(STATE_PROPERTY);
        if (configFileName == null)
        {
            if (_logger.isEnabledFor(Logger.INFO))
                _logger.logMessage("State manager started, but no state configuration property found: " + STATE_PROPERTY
                        + " (e.g. kaos.state=.\\config\\guard\\state.properties)", Logger.INFO);

            return;
        }

        File configFile = new File(configFileName);
        if (configFile.isFile())
        {
            Properties properties = new Properties();
            FileInputStream input = null;

            try
            {
                input = new FileInputStream(configFile);
                properties.load(input);

                for (Object key : properties.keySet())
                {
                    // find state to monitor for
                    String stateSensorClassName = (String)key;
                    String ontAndParametersString = properties.getProperty(stateSensorClassName);
                    StringTokenizer st = new StringTokenizer(ontAndParametersString, " \t");
                    String ontString = (String)st.nextToken();
                    ArrayList<String> parameters = new ArrayList<String>();
                    while (st.hasMoreTokens())
                    {
                        parameters.add(st.nextToken());
                    }

                    // // federate property is an url so the parser returns just http and the rest a a value
                    // if (key.equals("http") && (ontAndParametersString.indexOf('=') > 0))
                    // {
                    // // This is a federate property
                    // String wholePropertyWithValue = key + ontAndParametersString;
                    // key = wholePropertyWithValue.substring(0, wholePropertyWithValue.indexOf('=')).trim();
                    // ontAndParametersString = wholePropertyWithValue.substring(wholePropertyWithValue.indexOf('=') + 1).trim();
                    // }

                    System.out.println("creating monitor for " + ontString + " using " + stateSensorClassName + " " + parameters);

                    // instantiate appropriate class
                    Class stateSensorClass = Class.forName(stateSensorClassName);
                    StateSensor stateSensor = (StateSensor)stateSensorClass.newInstance();
                    stateSensor.initialize(parameters);

                    synchronized (statesSensors)
                    {
                        statesSensors.add(stateSensor);

                        // we do this within the sync block so that the statesOfInterest
                        // are not updated while we are doing this (it is also sync'd
                        // on "stateSensors" variable)
                        for (OntClassDescription state : statesOfInterest)
                        {
                            if (stateSensor.getOntologicalAttributes().contains(state.getClassName()))
                            {
                                stateSensor.registerInterest(state);
                            }
                        }
                    }

                    registerComponent(stateSensor);
                }
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
        }
        else
        {
            if (_logger.isEnabledFor(Logger.INFO))
                _logger.logMessage("State manager started, but no state configuration file found: " + configFileName, Logger.INFO);
        }

    }

    public Collection<OntInstanceDescription> getStates()
    {
        return new ArrayList<OntInstanceDescription>(mapStateIdToDescriptions.values());
    }

    public Collection<StateSensor> getStateSensors()
    {
        return this.statesSensors;
    }

    public List<OntInstanceDescription> getStatesOfClass(String stateClassName)
    {
        ArrayList<OntInstanceDescription> states = new ArrayList<OntInstanceDescription>();
        synchronized (mapStateIdToDescriptions)
        {
            for (OntInstanceDescription state : mapStateIdToDescriptions.values())
            {
                for (String type : state.getTypes())
                {
                    if (type.equals(stateClassName))
                    {
                        states.add(state);

                        break;
                    }
                }
            }
        }

        return states;
    }

    public OntInstanceDescription getStateInstance(String stateId)
    {
        return mapStateIdToDescriptions.get(stateId);
    }

    public Set<String> getPropertyValueForState(String stateId, String propertyName) throws Exception
    {
        OntInstanceDescription stateDescription = mapStateIdToDescriptions.get(stateId);
        if (stateDescription == null)
        {
            throw new Exception("No such state registered: " + stateId);
        }

        OntPropertyDescription propertyDescription = stateDescription.getProperty(propertyName);
        if (propertyDescription == null)
        {
            throw new Exception("No such property exists: " + propertyName);
        }

        return propertyDescription.getValues();
    }

    public boolean isStateSatisfied(OntClassDescription stateClass)
    {
        // System.out.println("isStateSatisfied = " + stateClass);
        List<OntInstanceDescription> currentStates = getStatesOfClass(stateClass.getMainSuperClassName());

        // Check if the current state instances match the class
        for (OntInstanceDescription currentState : currentStates)
        {
            int matchType = stateClass.matchInstanceWithClass(currentState, null);

            if (matchType == KAoSGuardPolicyProcessor._FULL_MATCH_)
                return true;
            else if (matchType == KAoSGuardPolicyProcessor._PARTIAL_MATCH_)
                return true; // TODO reconsider this
        }

        // System.out.println("returning false");

        return false; // Have not found state satisfying the state class
    }

    public Set<String> getSubjectWithState(OntClassDescription stateClass)
    {
        // System.out.println("getSubjectWithState = " + stateClass);
        Set<String> subjectsWithState = new HashSet<String>();

        List<OntInstanceDescription> currentStates = getStatesOfClass(stateClass.getMainSuperClassName());

        // Check if the current state instances match the class
        for (OntInstanceDescription currentState : currentStates)
        {
            int matchType = stateClass.matchInstanceWithClass(currentState, null);

            String currentSubject = currentState.getFirstValueForProperty(EntityConcepts.hasSubject());

            if (currentSubject == null)
                continue;

            if (matchType == KAoSGuardPolicyProcessor._FULL_MATCH_)
                subjectsWithState.add(currentSubject);
            else if (matchType == KAoSGuardPolicyProcessor._PARTIAL_MATCH_)
                subjectsWithState.add(currentSubject); // TODO reconsider this
        }

        // System.out.println("returning  = " + subjectsWithState);

        return subjectsWithState;
    }

    public void registerState(OntInstanceDescription stateInstance)
    {
        String stateId = stateInstance.getName();

        OntInstanceDescription description = mapStateIdToDescriptions.get(stateId);
        if (description != null)
        {
            System.out.println(stateId + " already exists, replacing it.");
        }
        else
        {
            // System.out.println("adding new state");
        }

        synchronized (mapStateIdToDescriptions)
        {
            mapStateIdToDescriptions.put(stateId, stateInstance);
        }

        if (USE_DATABASE)
        {
            System.out.println("database support not implemented");

            // long asOfTime = System.currentTimeMillis();
            // _stateDatabase.insertOrUpdateState(entityID, stateID, asOfTime, stateValue);
        }

        // ArrayList<StateUpdateListener> listeners = mapStateInstancesToListeners.get(stateID);
        // if (listeners != null)
        // {
        // for (StateUpdateListener listener : listeners)
        // {
        // listener.stateUpdate(entityID, stateID, stateValue);
        // }
        // }
    }

    public void registerState(String stateId, String stateClassName) throws URISyntaxException
    {
        registerState(new OntInstanceDescriptionImpl(stateId, stateClassName));
    }

    public void deregisterState(String stateId)
    {
        synchronized (mapStateIdToDescriptions)
        {
            mapStateIdToDescriptions.remove(stateId);
        }
    }

    synchronized public void updateStateProperty(String stateId, OntPropertyDescription updateproperty) throws Exception
    {
        String stateClass = updateStatePropertyInternal(stateId, updateproperty);
        if (stateClass != null)
            updateListeners(stateId, stateClass, updateproperty);
    }

    synchronized private String updateStatePropertyInternal(String stateId, OntPropertyDescription updateproperty) throws Exception
    {
        OntInstanceDescription stateDescription = mapStateIdToDescriptions.get(stateId);
        OntPropertyDescription property = null;
        String currentValue = null;

        if (stateDescription == null)
        {
            throw new Exception("No such state registered: " + stateId);
        }

        property = stateDescription.getProperty(updateproperty.getName());

        if (property == null)
        {
            return null;
        }
        else
        {
            currentValue = property.getFirstValue();
        }

        if ((currentValue != null) && currentValue.equals(updateproperty.getFirstValue()))
        {
            // System.out.println("property value has not changed: " + propertyValue + " for " + stateId);
        }
        else
        {
            // remove old value
            stateDescription.removeProperty(updateproperty.getName());

            // add new value
            stateDescription.addProperty(updateproperty.getName(), updateproperty.getFirstValue());

            synchronized (mapStateIdToDescriptions)
            {
                mapStateIdToDescriptions.put(stateId, stateDescription);
            }

            if (USE_DATABASE)
            {
                System.out.println("database support is not implemented yet");

                // long asOfTime = System.currentTimeMillis();
                // _stateDatabase.insertOrUpdateState(entityID, stateID, asOfTime, stateValue);
            }

            String stateClass = (String)stateDescription.getTypes().toArray()[0];
            return stateClass;
        }
        return null;
    }

    synchronized public void updateStateProperties(String stateId, List<OntPropertyDescription> properties) throws Exception
    {
        String stateClass = null;
        for (OntPropertyDescription currentProperty : properties)
        {
            String stateclass = updateStatePropertyInternal(stateId, currentProperty);
            if (stateClass == null)
                stateClass = stateclass;
        }

        if (stateClass != null)
        {
            updateListeners(stateId, stateClass, properties);
        }
    }

    private void updateListeners(String stateId, String stateClass, OntPropertyDescription updatedproperty)
    {
        List<OntPropertyDescription> tempList = new ArrayList<OntPropertyDescription>();
        tempList.add(updatedproperty);

        updateListeners(stateId, stateClass, tempList);
    }

    private void updateListeners(String stateId, String stateClass, List<OntPropertyDescription> properties)
    {
        // update those listening to everything
        for (StateUpdateListener listener : stateListeners)
        {
            listener.stateUpdate(stateId, stateClass, properties);
        }

        // update those listening for a particular class
        List<StateUpdateListener> listeners = mapStateClassesToListeners.get(stateClass);
        if (listeners != null)
        {
            for (StateUpdateListener listener : listeners)
            {
                listener.stateUpdate(stateId, stateClass, properties);
            }
        }

        // update those listening for a particular instance
        listeners = mapStateInstancesToListeners.get(stateId);

        if (listeners != null)
        {
            for (StateUpdateListener listener : listeners)
            {
                listener.stateUpdate(stateId, stateClass, properties);
            }
        }
    }

    public void removePropertyFromState(String stateId, String propertyName) throws Exception
    {
        OntInstanceDescription stateDescription = mapStateIdToDescriptions.get(stateId);

        if (stateDescription == null)
        {
            throw new Exception("No such state registered: " + stateId);
        }

        stateDescription.removeProperty(propertyName);
    }

    public void registerAsListenerForState(StateUpdateListener listener)
    {
        stateListeners.add(listener);
    }

    public void registerAsListenerForStateClass(String stateClass, StateUpdateListener listener)
    {
        System.out.println("registered listener for " + stateClass);
        List<StateUpdateListener> listeners = mapStateClassesToListeners.get(stateClass);
        if (listeners == null)
        {
            listeners = new ArrayList<StateUpdateListener>();
        }

        listeners.add(listener);
        mapStateClassesToListeners.put(stateClass, listeners);
    }

    public void registerAsListenerForStateInstance(String stateID, StateUpdateListener listener)
    {
        System.out.println("registered listener for " + stateID);
        List<StateUpdateListener> listeners = mapStateInstancesToListeners.get(stateID);
        if (listeners == null)
        {
            listeners = new ArrayList<StateUpdateListener>();
        }

        listeners.add(listener);
        mapStateInstancesToListeners.put(stateID, listeners);
    }

    public void registerInterest(OntClassDescription stateOfInterest)
    {
        synchronized (statesSensors)
        {
            statesOfInterest.add(stateOfInterest);

            for (StateSensor sensor : statesSensors)
            {
                if (sensor.getOntologicalAttributes().contains(stateOfInterest.getMainSuperClassName()))
                {
                    sensor.registerInterest(stateOfInterest);
                }
            }
        }
    }

    private Map<String, String> getStatesForEntity(String entityID)
    {
        Map<String, String> mapStatesToValues = mapEntitiesToStates.get(entityID);
        if (mapStatesToValues == null)
        {
            mapStatesToValues = new HashMap<String, String>();
        }

        return mapStatesToValues;
    }
}