/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2011/04/07 17:52:58 $
 */
package kaos.core.csi.extension.state;

import kaos.core.csi.extension.KAoSExtensionComponentManager;

import java.util.HashMap;

/**
 * This interface defines a factory method for obtaining a StateMonitor for the given
 * ontological type, representing a state.  The monitor manager will be used to monitor
 * properties of a given state for a particular entity. Methods are provided for sensors
 * to update the state and for interested parties to query the state.
 */
public interface StateMonitorManager 
	extends KAoSExtensionComponentManager

{
	/**
	 * Return a StateMonitor for the specified ontological type. 
	 *
	 * @param  ontType			String specifying the ontological type, for which a StateMonitor is requested.
	 *     
	 * @return StateMonitor		an instance of the requested StateMonitor, or null, if the manager is not
	 *                          able to find or create the state monitor.
	 */
	public StateSensor getStateMonitor(String ontType);

   /**
    * This method is to be used by state sensors to publish the multiple state updates to an overall state database
    * maintained in the state monitor manager
    *
    * @param updates a mapping of entity IDs to a mapping of stateIDs to state values
    */
   public void updateState(HashMap<String, HashMap> updates);

   /**
    * This method is to be used by state sensors to publish the current state to an overall state database
    * maintained in the state monitor manager
    *
    * @param entityID    String representing the id if the entity
    * @param stateID     String representing the id of the state
    * @param stateValue  String representing the current value of the state
    */
   public void updateState(String entityID, String stateID, String stateValue);

   /**
    * This method is to be used by state sensors to remove the current state from an entity in the
    * overall state database maintained in the state monitor manager
    *
    * @param entityID  The entity to remove state from
    * @param stateID   The particular state to remove
    */
   public void removeStateForEntity(String entityID, String stateID);

   /**
    * This method is to be used by state sensors to remove all state associated with an entity in the
    * overall state database maintained in the state monitor manager
    *
    * @param entityID the entity ID to remove all state from
    */
   public void removeEntity(String entityID);

   /**
    * This method is to be used by state sensors to remove all state associated with an entity in the
    * overall state database maintained in the state monitor manager
    *
    * @param stateID the state ID to be removed from all entities
    */
   public void removeState(String stateID);

   /**
    * This method can be used to query the state database for the value of a state for a particular entity
    *
    * @param entityID  String representing the entities ID
    * @param stateID   String representing the state ID of interest
    * @return  String representing the value of the current state for that specified entity
    */
   public String getStateValueForEntity(String entityID, String stateID);

   /**
    * This method can be used to query the state database for the value of a state for a particular entity
    *
    * @param entityID  String representing the entities ID
    * @return  map of states to their current values for the specified entity
    */
   public HashMap<String, String> getStateValuesForEntity(String entityID);
}
