/*
 * KAoSObservableImpl.java Created on March 4, 2005, 5:52 PM
 */

package kaos.core.csi.observe;

import kaos.core.csi.AgentCapabilityImpl;
import kaos.core.service.directory.KAoSAgentDescription;
import kaos.core.util.Logger;

import javax.agent.Locator;
import javax.agent.service.transport.MessageTransportService;
import java.util.*;

/**
 * @author jlott
 */
public class KAoSObservableImpl implements KAoSObservable
{
    private MessageTransportService _mts;
    private Locator                 _locator;
    Hashtable                       _observers = new Hashtable();
    protected Logger                _logger    = Logger.getLogger(KAoSObservableImpl.class.getName());
    public String                   _name;

    /** Creates a new instance of KAoSObservableImpl */
    public KAoSObservableImpl(MessageTransportService mts, Locator senderLocator, String senderName)
    {
        _mts = mts;
        _locator = senderLocator;
        _name = senderName;
    }

    // @return Hashtable<String,Hashtable<Object,Vector<KAoSObserver>>>
    public Hashtable getObservers()
    {
        return _observers;
    }

    public synchronized void registerObserver(String updateType, Object updateCondition, KAoSObserver observer)
    {
        if (_logger.isEnabledFor(Logger.DEBUG))
            _logger.logMessage("registerObserver: updateType = " + updateType, Logger.DEBUG);

        if (observer instanceof RemoteKAoSObserver)
        {
            RemoteKAoSObserver rObserver = (RemoteKAoSObserver)observer;
            rObserver.setTransportInfo(_mts, _locator, _name);
        }

        if (_logger.isEnabledFor(Logger.DEBUG))
            _logger.logMessage("registerObserver: registering observer for condition" + updateCondition, Logger.DEBUG);

        // Register observer
        if (!(_observers.containsKey(updateType)))
            _observers.put(updateType, new Hashtable());

        Hashtable observersForType = (Hashtable)_observers.get(updateType);

        if (!(observersForType.containsKey(updateCondition)))
            observersForType.put(updateCondition, new Vector());

        Vector observersForCondition = (Vector)observersForType.get(updateCondition);
        observersForCondition.add(observer);

        if (_logger.isEnabledFor(Logger.DEBUG))
            _logger.logMessage("updateObservers: added observer:\n " + observer + "\n for: " + updateCondition + "\n now observer vector: " + observersForCondition, Logger.DEBUG);
    }

    /**
     * Deregister the given KAoSObserver for the given update type from the registry of observers associated with this KAoSObservable.
     * 
     * @param updateType String indicating type of update the KAoSObserver is no longer interested in. For now the DS would handle only KAoSConstants.REGISTRATION_UPDATE.
     * @param updateCondition See the same in the method heading for registerObserver.
     * @param observer the KAoSObserver deregistering.
     */
    public void deregisterObserver(String updateType, Object updateCondition, KAoSObserver observer)
    {
        synchronized (_observers)
        {
            if (_observers.containsKey(updateType))
            {
                Hashtable observersForType = (Hashtable)_observers.get(updateType);
                if (observersForType.containsKey(updateCondition))
                {
                    Vector observersForCondition = (Vector)observersForType.get(updateCondition);
                    if (!observersForCondition.remove(observer))
                    {
                        if (_logger.isEnabledFor(Logger.WARN))
                            _logger.logMessage("deregisterObserver: observer not registered " + observer, Logger.WARN);
                    }
                    if (observersForCondition.isEmpty()) {
                        observersForType.remove(updateCondition);
                    }
                }
                else
                {
                    if (_logger.isEnabledFor(Logger.WARN))
                        _logger.logMessage("deregisterObserver: no observers for condition " + updateCondition, Logger.WARN);
                }
            }
            else
            {
                if (_logger.isEnabledFor(Logger.WARN))
                    _logger.logMessage("deregisterObserver: no observers for type " + updateType, Logger.WARN);
            }
        }
    }

    /**
     * Notify observers interested in the given updateType wiht the given update object.
     * 
     * @param updateType String describing the type of update to perform. For now the DS would handle only KAoSConstants.REGISTRATION_UPDATE.
     * @param updateCondition See the same in the method heading for registerObserver.
     * @param update Object containing the update. For now this is the agent description of the matching agent. Please note that if the observer has registered for 'AGENT_LIST', then
     *            it will initially get a list of all agent descriptions, and afterwards will keep getting individual descriptions as new agents come online. If it has not
     *            registered for AGENT_LIST, then it will descriptions of all matching agents seperately (i.e. not all in one list) and then will keep getting individual
     *            descriptions as new agents come online.
     */
    public void notify(String updateType, Object updateCondition, Object update)
    {
        Thread notifyThread = new NotifyThread2(updateType, updateCondition, update);
        notifyThread.setDaemon(true);
        notifyThread.start();
    }

    class NotifyThread2 extends Thread
    {
        private String updateType;
        private Object updateCondition;
        private Object update;

        public NotifyThread2(String updateType, Object updateCondition, Object update)
        {
            this.updateType = updateType;
            this.updateCondition = updateCondition;
            this.update = update;
        }

        public void run()
        {
            if (_logger.isEnabledFor(Logger.INFO))
                _logger.logMessage(".notify: sending update type: " + updateType + "\n  updateCondition: " + updateCondition + "\n with value: " + update, Logger.INFO);

            synchronized (_observers)
            {
                if (_observers.containsKey(updateType))
                {
                    Hashtable observersForType = (Hashtable)_observers.get(updateType);

                    for (Iterator it = observersForType.keySet().iterator(); it.hasNext();)
                    {
                        Hashtable condition = (Hashtable)it.next();

                        Object matchingUpdate = matchCondition(condition, (Hashtable) updateCondition, update);
                        if (matchingUpdate != null) //(conditionsMatch(condition, (Hashtable)updateCondition))
                        {
                            Vector observers = (Vector)observersForType.get(condition);

                            for (Enumeration observersEnum = observers.elements(); observersEnum.hasMoreElements();)
                            {
                                KAoSObserver observer = (KAoSObserver)observersEnum.nextElement();

                                if (_logger.isEnabledFor(Logger.INFO))
                                    _logger.logMessage(".notify: sending update to : " + observer + " observer condition: " + condition, Logger.INFO);

                                try
                                {
                                    observer.update(_name, updateType, condition, matchingUpdate);

                                    if (_logger.isEnabledFor(Logger.INFO))
                                        _logger.logMessage(".notify:" + _name + "  finished sending update" + " update type = " + updateType + " condition = " + updateCondition,
                                                Logger.INFO);
                                }
                                catch (Exception e)
                                {
                                    e.printStackTrace();
                                }
                            }
                        }
                    }
                }
                else if (_logger.isEnabledFor(Logger.INFO))
                    _logger.logMessage(".notify: there is no observer for updateType: " + updateType, Logger.INFO);
            }
        }
    }

    protected boolean isObserverRegisteredForCondition(String updateType, Object condition)
    {
        if (_observers.containsKey(updateType))
        {
            Hashtable observersForType = (Hashtable)_observers.get(updateType);
            Vector observers = (Vector)observersForType.get(condition);

            if (observers != null && !observers.isEmpty())
            {
                // return true; /**HACK because this does not handle multiple agents**/
                return false;
            }
        }

        return false;
    }

    // @return the update object which matches the requested condition, or
    // null if the actual condition doesn't match the requested condition
    protected Object matchCondition(Hashtable requestedCondition, Hashtable actualCondition, Object update)
    {
        Enumeration keys = requestedCondition.keys();

        while (keys.hasMoreElements())
        {
            String key = (String)keys.nextElement();

            if (actualCondition.containsKey(key))
            {
                // the key exists in the description
                Object actualConditionValue = actualCondition.get(key);
                Object requestedConditionValue = requestedCondition.get(key);

                if (key.equals(KAoSAgentDescription.AGENT_CAPABILITIES))
                {
                    if ((actualConditionValue instanceof List) && (requestedConditionValue instanceof List))
                    {
                        if (!AgentCapabilityImpl.capabilitiesMatch((List)requestedConditionValue, (List)actualConditionValue))
                        {
                            return null;
                        }
                    }
                }
                else if (!requestedConditionValue.equals(actualConditionValue))
                {
                    return null;
                }
            }
            else
            {
                return null;
            }
        }
        return update;
    }
}
