package kaos.core.csi.ontology;

import kaos.core.csi.query.QueryState;
import kaos.core.csi.registration.Property;
import kaos.core.service.directory.ModificationException;

import javax.agent.service.directory.DirectoryFailure;
import javax.agent.service.directory.NotRegisteredException;

import java.io.IOException;
import java.util.List;
import java.util.Set;

/**
 * Defines the operations allowing to load ontology into the repository, methods allowing to modify instance types and values of properties.
 * Additional it has methods allowing to configures filtering out values from specific namespaces from the set of values returned by its methods.
 * Adds methods allowing to query the ontology. The queries do not depend on any specific ontology.
 * 
 * @author KAoS Team
 * @version 1.0 $Revision: 1.22 $
 */
public interface OntologyService extends QueryState
{
    /**
     * Loads ontology in OWL into the repository of the Ontology Service
     * 
     * @param loadedOWL string containing a syntactically correctly OWL
     * @throws OntologyService.ReasoningException when loading of the ontology failed.
     */
    public void loadOntology(String loadedOWL) throws OntologyService.ReasoningException, DirectoryFailure;

    /**
     * Loads ontology in OWL from given URL into the repository of the Ontology Service
     * 
     * @param myURL string containing a syntactically correct URI to an OWL source
     * @param recursiveLoad whether to load the OWL imported by the given URI
     * @return Returns a set to of namespaces loaded during this step
     * @throws OntologyService.ReasoningException when loading of the ontology failed.
     */
    public Set<String> loadOntology(String myURL, boolean recursiveLoad) throws OntologyService.ReasoningException, IOException, DirectoryFailure;

    /**
     * Add new type to the entity
     * 
     * @param entityName - the uri of the entity
     * @param typeName - the type name
     * @param typeDesc - optional ontology description of the type
     * @throws OntologyService.ReasoningException
     */
    public void addOntologicalTypeToEntity(String entityName, String typeName, String typeDesc) throws OntologyService.ReasoningException,
            UnknownConceptException, DirectoryFailure;

    /**
     * Remove a type from the entity
     * 
     * @param entityName - the uri of the entity
     * @param typeName - the type name
     * @throws OntologyService.ReasoningException
     * @throws DirectoryFailure
     */
    public void removeOntologicalTypeFromEntity(String entityName, String typeName) throws OntologyService.ReasoningException, UnknownConceptException,
            DirectoryFailure;

    /**
     * Add a property value to an entity
     * 
     * @param entityName - the uri of the entity
     * @param propertyName - the name of the property
     * @param propertyValue - the value of the property
     * @param propertyValueDesc - optional ontology description of the property value
     * @throws OntologyService.ReasoningException
     */
    public void addPropertyToEntity(String entityName, String propertyName, String propertyValue, String propertyValueDesc)
            throws OntologyService.ReasoningException, UnknownConceptException, DirectoryFailure;

    /**
     * Remove a property value from an entity
     * 
     * @param entityName - the uri of the entity
     * @param propertyName - the name of the property
     * @param propertyValue - the value of the property
     * @throws OntologyService.ReasoningException
     */
    public void removePropertyFromEntity(String entityName, String propertyName, String propertyValue) throws OntologyService.ReasoningException,
            UnknownConceptException, DirectoryFailure;

    /**
     * Modify properties of the specified entity
     * 
     * @param entityGUID - the uri of the entity
     * @param properties - list of properties with new values
     * @throws DirectoryFailure, NotRegisteredException, ModificationException
     */
    public void modifyEntityProperties(String entityGUID, List<Property> properties) throws DirectoryFailure, NotRegisteredException,
            ModificationException;

    /**
     * This method returns a set of urls depicting ontology namespaces concept of which are filtered out when returned by methods of this repository.
     */
    public Set<String> getFilteredNamespaces() throws DirectoryFailure;

    /**
     * This method add a namespace to the set of urls depicting ontology namespaces concept of which are filtered out
     * when returned by methods of this repository.
     * 
     * @param namespace The name of the namespace
     */
    public void addFilteredNamespace(String namespace) throws DirectoryFailure;

    /**
     * This method remove a namespace from the set of urls depicting ontology namespaces concept of which are filtered out
     * when returned by methods of this repository.
     * 
     * @param namespace The name of the namespace
     */
    public void removeFilteredNamespace(String namespace) throws DirectoryFailure;

    public class ReasoningException extends Exception
    {
        private static final long serialVersionUID = 1L;

        /*
         * Constructs a new ReasoningException.
         * @param s the message to be displayed.
         */
        public ReasoningException(String s)
        {
            super(s);
        }
    }
}