/**
 * 
 */
package kaos.core.csi.policy;

import kaos.core.csi.ActionInstanceDescription;
import kaos.core.csi.OntInstanceDescription;
import kaos.core.csi.query.data.ActionClassDescription;

import javax.agent.service.ServiceFailure;
import java.util.List;

/**
 * This interface defines operations to advice the user as to which property values are allowed/forbidden for the given action based on policy.
 */
public interface PolicyAdvice
{
    /*
     * In this case, there are NO properties specified. The function uses the given, partially filled, ActionInstanceDescription object to find those
     * policies that are applicable to this action, and then it selects all the properties from those policies which are not originally contained in the
     * provided ActionInstanceDescription object. It then chooses only those values for the given properties which do not conflict with their higher
     * priority selves, and returns all those properties in the ActionInstanceDescription object.
     * @param actionDesc ActionInstanceDescription object which will be used to find applicable policies.
     * @param pObserver PolicyDecisionObserver, an entity interested in receiving updates whenever policy decisions change.
     * @return List containing multiple ActionClassDescription objects which contain the allowed values for those properties.
     * @throws ServiceFailure when the policy service cannot be reached, or fails.
     */
    public List<ActionClassDescription> getConfiguration(ActionInstanceDescription actionDesc, boolean tight,  
            PolicyDecisionObserver pObserver) throws ServiceFailure;

    /*
     * This function is used when the agent/enforcer has only partial information about an action and would like to determine what range of properties can
     * be allowed by the policy set. The agent/enforcer partially fills an ActionInstanceDescription object and sends it to the method, which finds those
     * policies that are applicable to this action and contain the given property. The method will then select only those values for the given properties
     * that will not conflict with higher priority policies containing the given properties.
     * @param propertyNames List containing the properties for which values are to be found.
     * @param actionDesc ActionInstanceDescription object which will be used to find applicable policies.
     * @param tight boolean, if set to 'true', will result in returning only these values for the missing specified property which would satisfy some
     * policy if used alone.
     * @param pObserver PolicyDecisionObserver, an entity interested in receiving updates whenever policy decisions change.
     * @return List containing multiple ActionClassDescription objects which contain the allowed values for those properties.
     * @throws ServiceFailure when the policy service cannot be reached, or fails.
     */
    public List<ActionClassDescription> getConfiguration(List<String> propertyNames, ActionInstanceDescription actionDesc,
            boolean tight, PolicyDecisionObserver pObserver) throws ServiceFailure;

    /*
     * This method return value forbidden by the policies for a given property (for instance moveTo)
     * @param origPropertyName the property for which to find those values.
     * @param origActionDesc the AID for which to find the values.
     * @param pObserver PolicyDecisionObserver, an entity interested in receiving updates whenever policy decisions change.
     * @return Vector a Vector of OntInstanceDescriptionImpl objects, which are the forbidden values.
     * @throws ServiceFailure when the policy service cannot be reached, or fails.
     */
    public List<OntInstanceDescription> getForbiddenValuesForActionProperty(String origPropertyName, ActionInstanceDescription origActionDesc, 
            PolicyDecisionObserver pObserver) throws ServiceFailure;

    public Integer getPolicyUpdateCount() throws ServiceFailure;
}
