/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/04/13 22:03:56 $
 */
package kaos.core.csi.policy;

import kaos.core.csi.ActionInstanceDescription;
import kaos.core.csi.query.data.ActionClassDescription;
import kaos.core.service.boot.KAoSGuardBootstrapper;
import kaos.core.service.directory.DirectoryService;
import kaos.core.util.Logger;
import kaos.core.util.PolicyMsg;
import kaos.policy.guard.GuardRetriever;
import kaos.policy.guard.KAoSGuard;

import javax.agent.service.ServiceFailure;
import java.util.ArrayList;
import java.util.List;

/**
 * Defines the operations to check permissions, get obligations, analyze policies, etc.
 * 
 * @author KAoS Team $Revision: 1.40 $
 */
public class PolicyCheckingImpl implements PolicyAdvice, PolicyChecking
{
    private static final PolicyCheckingImpl __INSTANCE    = new PolicyCheckingImpl();
    private static boolean                  __INITIALIZED = false;

    private DirectoryService                 _dirProxy     = null;
    private Logger                          _logger       = Logger.getLogger(this);

    /*
     * Retrieve an instance of PolicyChecking. The returned instance has initialized communication with the DS.
     * @param transportName String specifying the name of the transport to use for communication with the KAoS Directory Service.
     * @return a singleton, PolicyChecking.
     */
    public static PolicyChecking getInstance(String transportName) throws Exception
    {
        if (__INITIALIZED)
            return __INSTANCE;

        __INSTANCE.initialize(transportName);
        __INITIALIZED = true;

        return __INSTANCE;
    }

    /*
     * Retrieve an instance of PolicyAdvice. The returned instance has initialized communication with the DS.
     * @param transportName String specifying the name of the transport to use for communication with the KAoS Directory Service.
     * @return a singleton, PolicyAdvice.
     */
    public static PolicyAdvice getPolAdviceInstance(String transportName) throws Exception
    {
        if (__INITIALIZED)
            return __INSTANCE;

        __INSTANCE.initialize(transportName);
        __INITIALIZED = true;

        return __INSTANCE;
    }

    /*
     * Retrieve an instance of PolicyChecking. The returned instance may or may not be able to communicate with the Directory Service. If the communication
     * with the Directory Service is not supported, then all requested operations will be invoked only on the local Guard.
     * @return a singleton, PolicyChecking.
     */
    public static PolicyChecking getInstance() throws Exception
    {
        return __INSTANCE;
    }

    private PolicyCheckingImpl()
    {
    }

    // ----------- kaos.core.csi.policy.PolicyChecking interface implementation - BEGIN

    /*
     * The method check if the given action is permitted according to the current set of policies.
     * @param aid ActionInstanceDescription describing action
     * @param pObserver PolicyDecisionObserver, an entity interested in receiving updates whenever policy decisions change.
     * @throws KAoSSecurityException if the action is not allowed.
     * @throws ServiceFailure if the policy service is not available.
     */
    public void checkPermission(ActionInstanceDescription aid, PolicyDecisionObserver pObserver) throws KAoSSecurityException, ServiceFailure
    {
        KAoSGuard guard = GuardRetriever.getGuard();

        if (guard != null)
        {
            guard.checkPermission(aid, pObserver);
            return;
        }
        else
        if (_dirProxy != null)
        {
            _dirProxy.checkPermission(aid, pObserver);
            return;
        }
        else
            throw new ServiceFailure("Policy service unavailable! Unable to check permission for: " + aid);
    }

    /*
     * The method checks if the given action is permitted according to the current set of policies. However, if the given action is not permitted by a
     * policy, instead of stopping there, it goes on to find all those policies that forbid the given action, and then returns a vector that contains their
     * ids. The vector containing forbidding policies' ids is stored in the KAoSSecurityException.
     * @param aid ActionInstanceDescription to be checked by the Guard in order to allow/disallow an action.
     * @param pObserver PolicyDecisionObserver, an entity interested in receiving updates whenever policy decisions change.
     * @throws ServiceFailure if the policy service is not available, or fails.
     * @throws KAoSSecurityException if the action is not permitted.
     */
    public void checkDeepPermission(ActionInstanceDescription aid, PolicyDecisionObserver pObserver) throws ServiceFailure, KAoSSecurityException
    {
        KAoSGuard guard = GuardRetriever.getGuard();

        if (guard != null)
        {
            guard.checkDeepPermission(aid, pObserver);
            return;
        }
        else
        if (_dirProxy != null)
        {
            _dirProxy.checkDeepPermission(aid, pObserver);
            return;
        }
        else
            throw new ServiceFailure("Policy service unavailable! Unable to check permission for: " + aid);
    }

    /*
     * The method checks if the given action is permitted according to the current set of policies, or if the given action is a trigger for some obligation
     * policy. If the action is not permitted, the method throws an exception, but if the action is a trigger, the method returns a Vector containing all
     * those obligation policies for which the action is a trigger.
     * @param aid ActionInstanceDescription to be checked by the Guard in order to allow/disallow the action; or the trigger to be checked for obligation
     * policies.
     * @param pObserver PolicyDecisionObserver, an entity interested in receiving updates whenever policy decisions change.
     * @return A Vector of obligation ActionInstanceDescriptions. The returned obligations are triggered by the given aid. The actor of the returned
     * obligations is the same as the actor of the given aid.
     * @throws ObligationViolationException some aspect of the returned obligations have been violated
     * @throws ServiceFailure if the policy service is not available, or fails.
     * @throws KAoSSecurityException if the action is not permitted.
     */
    public List<ObligationActionDescription> findPolicyDecision(ActionInstanceDescription aid, PolicyDecisionObserver pObserver)
            throws ObligationViolationException, KAoSSecurityException, ServiceFailure
    {
        KAoSGuard guard = GuardRetriever.getGuard();

        if (guard != null)
            return guard.findPolicyDecision(aid, pObserver);
        else
        if (_dirProxy != null)
            return _dirProxy.findPolicyDecision(aid, pObserver);
        else
            throw new ServiceFailure("Policy service unavailable! Unable to find policy decision for: " + aid);
    }

    /*
     * This function is used when the agent/enforcer has only partial information about an action and would like to determine what range of properties can
     * be allowed by the policy set. The agent/enforcer partially fills an ActionInstanceDescription object and sends it to the method, which finds those
     * policies that are applicable to this action and contain the given property. The method will then select only those values for the given properties
     * that will not conflict with higher priority policies containing the given properties.
     * @param propertyNames Vector containing the properties for which values are to be found.
     * @param actionDesc ActionInstanceDescription object which will be used to find applicable policies.
     * @param tight boolean, if set to 'true', will result in returning only these values for the missing specified property which would satisfy some
     * policy if used alone.
     * @param pObserver PolicyDecisionObserver, an entity interested in receiving updates whenever policy decisions change.
     * @return List The List containing multiple ActionClassDescription objects which contain the allowed values for those properties.
     */
    public List<ActionClassDescription> getConfiguration(List<String> propertyNames, ActionInstanceDescription actionDesc, boolean tight,
            PolicyDecisionObserver pObserver) throws ServiceFailure
    {
        KAoSGuard guard = GuardRetriever.getGuard();

        if (guard != null)
            return guard.getConfiguration(propertyNames, actionDesc, tight, pObserver);
        else
        if (_dirProxy != null)
            return _dirProxy.getConfiguration(propertyNames, actionDesc, tight, pObserver);
        else
            throw new ServiceFailure("Unable to obtain Guard!");
    }

    /*
     * In this case, there are NO properties specified. The function uses the given, partially filled, ActionInstanceDescription object to find those
     * policies that are applicable to this action, and then it selects all the properties from those policies which are not originally contained in the
     * provided ActionInstanceDescription object. It then chooses only those values for the given properties which do not conflict with their higher
     * priority selves, and returns all those properties in the ActionInstanceDescription object.
     * @param actionDesc ActionInstanceDescription object which will be used to find applicable policies.
     * @param pObserver PolicyDecisionObserver, an entity interested in receiving updates whenever policy decisions change.
     * @return Vector The Vector containing multiple ActionInstanceDescription objects which contain the allowed values for those properties.
     */
    public List<ActionClassDescription> getConfiguration(ActionInstanceDescription actionDesc, boolean tight, PolicyDecisionObserver pObserver) 
        throws ServiceFailure
    {
        KAoSGuard guard = GuardRetriever.getGuard();

        if (guard != null)
            return guard.getConfiguration(actionDesc, tight, pObserver);
        else
        if (_dirProxy != null)
            return _dirProxy.getConfiguration(actionDesc, tight, pObserver);
                
        else
            throw new ServiceFailure("Unable to obtain Guard!");
    }

    /*
     * Based on the specified trigger ActionInstanceDescription, select all matching control BasicActionDescriptions, which define the obligations for the
     * trigger. For each control BasicActionDescription create an ActionInstanceDescription. Sort the ActionInstanceDescriptions in descending order by the
     * priority of the obligation policy, in which the BasicActionDescription was defined.
     * @param triggerAID ActionInstanceDescription describes an instance of the trigger condition sent by the trigger condition monitor.
     * @param pObserver PolicyDecisionObserver, an entity interested in receiving updates whenever policy decisions change.
     * @return A Vector of control ActionInstanceDescriptions, whose BasicActionDescriptions are contained in policies matching the parameter triggerAID.
     * The returned control ActionInstanceDescriptions are sorted in descending sequence by their policy priority.
     * @throws ObligationViolationException when specified obligation constraints in a policy applicable to the given trigger are not satisfied.
     * @throws ServiceFailure if the policy service is not available.
     */
    public List<ObligationActionDescription> getObligationsForTriggerCondition(ActionInstanceDescription triggerAID, PolicyDecisionObserver pObserver)
            throws ObligationViolationException, ServiceFailure
    {
        KAoSGuard guard = GuardRetriever.getGuard();

        if (guard != null)
            return guard.getObligationsForTriggerCondition(triggerAID, pObserver);
        else
        if (_dirProxy != null)
            return _dirProxy.getObligationsForTriggerCondition(triggerAID, pObserver);
        else
            throw new ServiceFailure("Policy service unavailable! Unable to retrieve obligations for: " + triggerAID);
    }

    /*
     * At the moment this method would work only for A- policies for 'moveTo' action, and for 'movingTo' property. For other policies, the behavior of
     * this method is undefined. This method would be made more generic in future, for handling nested property values.
     * @param origPropertyName the property for which to find those values.
     * @param origActionDesc the AID for which to find the values.
     * @param pObserver PolicyDecisionObserver, an entity interested in receiving updates whenever policy decisions change.
     * @return Vector a Vector of OntInstanceDescriptionImpl objects,which are the forbidden values.
     */
    public List getForbiddenValuesForActionProperty(String origPropertyName, ActionInstanceDescription origActionDesc, PolicyDecisionObserver pObserver)
            throws ServiceFailure
    {
        KAoSGuard guard = GuardRetriever.getGuard();

        if (guard != null)
            return guard.getForbiddenValuesForActionProperty(origPropertyName, origActionDesc, pObserver);
        else
        if (_dirProxy != null)
            return _dirProxy.getForbiddenValuesForActionProperty(origPropertyName, origActionDesc, pObserver);
        else
            throw new ServiceFailure("Policy service unavailable! Unable to obtain forbidden values for property: " + origPropertyName);
    }

    /*
     * NOTE: this method added temporarily here to support a demo! It will be added to the implementation of PolicyManagement - where it belongs!!
     */
    public void removePolicy(PolicyMsg policyMsg) throws ServiceFailure
    {
        if (policyMsg != null)
        {
            if (_dirProxy != null)
            {
                List<PolicyMsg> policyList = new ArrayList<PolicyMsg>();
                policyList.add(0, policyMsg);

                try
                {
                    _dirProxy.updatePolicies(new ArrayList<PolicyMsg>(), new ArrayList<PolicyMsg>(), policyList);
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                    throw new ServiceFailure();
                }
            }
            else
            {
                throw new ServiceFailure("Policy Service unavailable! Unable to remove policy: " + policyMsg);
            }
        }
    }

    public Integer getPolicyUpdateCount()
    {
        // TODO Auto-generated method stub
        return null;
    }
    
    // ------- kaos.core.csi.policy.PolicyChecking interface implementation - END

    // ------- Private Methods BEGIN

    private void initialize(String transportName) throws Exception
    {
        try
        {
            KAoSGuard guard = GuardRetriever.getGuard();
            _dirProxy = KAoSGuardBootstrapper.getKAoSDirectoryProxyForTransport(transportName);
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new Exception("PolicyCheckingImpl: unable to obtain a reference the DirectoryService for transport: " + transportName + "!");
        }
    }

    protected void logMessage(String msg, int logLevel)
    {
        if (_logger.isEnabledFor(logLevel))
            _logger.logMessage(msg, logLevel);
    }

    protected void logMessage(String msg, Exception e, int logLevel)
    {
        if (_logger.isEnabledFor(logLevel))
            _logger.logMessage(msg, e, logLevel);
    }
}