/**
 * Last Modified by: $Author: lbunch $ On: $Date: 2012/04/23 23:51:28 $
 */

package kaos.core.csi.policy;

import kaos.core.util.PolicyMsg;
import kaos.core.util.PolicySetMsg;
import kaos.policy.management.PolicyPrecedenceDecision;

import javax.agent.service.ServiceFailure;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Defines the operations to manage policies in KAoS including adding, modifying and removing.
 * 
 * @author KAoS Team
 */
public interface PolicyManagement
{
    public void updatePolicies(List<PolicyMsg> addedPolicies, List<PolicyMsg> changedPolicies, List<PolicyMsg> removedPolicies)
            throws PolicyRelationException, ServiceFailure;

    // ----- ADD
    
    /*
     * Add the specified policy express as an OWL string to KAoS DS
     * @param owlPolicyDescription OWL String containing the policy description.
     * @throws PolicyRelationException
     * @throws DirectoryFailure
     */
    public void addPolicy(String owlPolicyDescription) throws PolicyRelationException, ServiceFailure;

    /*
     * Add the policy to KAoS
     * @param policyMsg - message with policy
     * @throws PolicyRelationException
     * @throws ServiceFailure
     */
    public void addPolicy(PolicyMsg policyMsg) throws PolicyRelationException, ServiceFailure;

    public void addPolicies(List<PolicyMsg> policies) throws PolicyRelationException, ServiceFailure;

    // ----- CHANGE
    
    /*
     * Modify the specified policy in KAoS.
     * @param policyID String containing the unique policy id.
     * @param owlPolicyDescription OWL String containing the modified policy description.
     * @throws ServiceFailure if the policy service is not available.
     */
    public void changePolicy(String policyID, String owlPolicyDescription) throws PolicyRelationException,  ServiceFailure;
    
    public void changePolicy(PolicyMsg policy) throws PolicyRelationException, ServiceFailure;

    public void changePolicies(List<PolicyMsg> policies) throws PolicyRelationException, ServiceFailure;

    // ----- REMOVE

    /*
     * Remove the specified policy from KAoS.
     * @param policyID String containing the unique policy id.
     * @throws ServiceFailure if the policy service is not available.
     */
    public void removePolicy(String policyID) throws ServiceFailure;

    public void removePolicy(PolicyMsg policy) throws ServiceFailure;

    public void removePolicies(List<PolicyMsg> policies) throws ServiceFailure;

    // ----- GET
    
    /*
     * Return the policy identified by the given id.
     * @param id String containing the id of the requested policy.
     * @return PolicyMsg representing.
     */
    public PolicyMsg getPolicy(String id) throws ServiceFailure;

    /*
     * Obtain an List over all policies. Each element of the List is an instance of a message class specifying an individual policy fields.
     * @return List of policy message
     */
    public List<PolicyMsg> getPolicies() throws ServiceFailure;

    /*
     * Obtain a map of all original policies in the system, indexed by the id(s) of the entity(s) to which they apply
     * @return Map of Map of policies key = String representing entity id value = Map of policies that are directly applicable to the entity
     */
    public Map<String, Map<String, PolicyMsg>> getPoliciesByEntities() throws ServiceFailure;

    /*
     * Obtain policies applicable to the given actor either individual or class of actors
     * 
     * @param id         the id of actor whose policies are being requested
     * @param inherited  if false only direct polices specifically defined for the actor are returned otherwise also inherited policies are included
     * @return List of policy messages
     */
    public List<PolicyMsg> getPoliciesFor(String id, boolean inherited) throws ServiceFailure;

    /*
     * Obtain policies applicable to the given actor either individual or class of actors
     * 
     * @param actorId    the id of actor whose policies are being requested
     * @param inherited  if false only direct polices specifically defined for the actor are returned otherwise also inherited policies are included
     * @return List of policy ids
     */
    public List<String> getPolicyIDsFor(String actorId, boolean inherited) throws ServiceFailure;
    
    /*
     * Returns policies based on the specified criteria
     * @param actionType - type of the action controlled by the policy
     * @param policyRestrictions - map linking chains of properties to restriction value
     * @return list of found policies
     * @throws ServiceFailure
     */
    public List<PolicyMsg> findPolicies(String actionType, Map<List<String>, String> policyRestrictions) throws ServiceFailure;

    /**
     * 
     * @return the list of all policies in the #PrecedencePolicy OWL class, which is those policies that regulate PolicyPrecedenceAction actions.
     */
    public List<PolicyMsg> getPrecedencePolicies();
    
    /**
     * 
     * @return The top-level PolicySetMsg for the set of Precedence Policies. There should always be one and only one such set.
     */
    public PolicySetMsg getPrecedencePolicySet();
    
    // ----- CHANGE LISTENER
    
    /*
     * Add the listener for policy update
     * @param listener reference to the implementation of PolicyDistributor
     */
    public void addPolicyUpdateListener(PolicyListener listener) throws ServiceFailure;

    /*
     * Remove the listener for policy update
     * @param listener reference to the implementation of PolicyDistributor
     */
    public void removePolicyUpdateListener(PolicyListener listener) throws ServiceFailure;

    // ----- DISTRIBUTED POLICY EXCHANGE
    
    public Set<String> getPeersSubscribedToPolicy (String policyId) throws ServiceFailure;

    public Set<String> getPeersSubscribedToPolicySet (String policySetId) throws ServiceFailure;

    public abstract Boolean requestPermissionToChangePolicy(PolicyMsg policy, String permission, String reason, String requestorId) throws ServiceFailure;

    public void respondToPermissionRequest (String permissionRequestId, Boolean allow) throws ServiceFailure;

    // ----- EXCEPTION
    
    public class PolicyRelationException extends Exception
    {
        private static final long serialVersionUID = 1L;
        
        public PolicyMsg policy;
        public List<String> conflictedPolicies;
        public List<String> overlappedPolicy;
        public List<String> overlapingPolicy;
        public List<PolicyPrecedenceDecision> precedenceDecisions;
        public List<String> uncomittedPolicies;
        
        public PolicyRelationException()
        {
            super();
        }

        public PolicyRelationException(String s)
        {
            super(s);
        }
    }
}