/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2012/03/21 14:22:36 $
 */

package kaos.core.csi.policy.enforcement;

import kaos.core.csi.ActionInstanceDescription;
import kaos.core.csi.ActionInstanceDescriptionImpl;
import kaos.core.csi.CSIFactory;
import kaos.core.csi.extension.KAoSExtensionComponentImpl;
import kaos.core.csi.policy.KAoSSecurityException;
import kaos.core.csi.policy.PolicyChecking;
import kaos.core.util.ExecEnvUtility;
import kaos.core.util.Logger;
import kaos.ontology.vocabulary.ActionConcepts;
import kaos.ontology.vocabulary.GroupConcepts;

import javax.agent.service.ServiceFailure;
import javax.agent.service.ServiceProperties;
import java.net.URISyntaxException;
import java.util.HashMap;

/**
 *
 * This class implements a CommunicationActionEnforcer interface.
 *
 * @author KAoS Team
 * $Revision: 1.14 $
 */
public class CommunicationActionEnforcerImpl extends KAoSExtensionComponentImpl implements CommunicationActionEnforcer
{
    private static final String                 _ENFORCER_ROOT = "CSI-COMM-ENFORCER-FOR-VM-";

    private PolicyChecking                      _policyChecking;

    // The following vars support enforcement for the CommunicationAction policies
	private String                              _enforcerID = null;
	private boolean                             _enforcerEnabled = false; // This enforcer will be enabled
	                                                                      // after registering itself with the EnforcerManager
	                                                                      //(method initializeServices).
	// end enforcement

	private Logger                              _logger = Logger.getLogger(this);

    public CommunicationActionEnforcerImpl()
    {
         initializeVars();
    }

    //////////////////////////////////////////////////////////////
    //
    // Implementation of kaos.policy.enforcement.Enforcer - begin
    //
    //////////////////////////////////////////////////////////////

	/**
     * Return the (unique) name of this KAoSExtensionComponent object.
     *
     * @return  name of this component.
     */
	public String getName()
	{
	    return  _enforcerID;
	}

	/**
     * Set the (unique) name of this KAoSExtensionComponents object.
     *
     * @param  name     String containing the name of this component.
     */
	public void setName(String name)
	{
	    if( _enforcerID == null )
	        _enforcerID = name;
	}

	/**
     * Set 'true/false', indicating the 'enabled/disabled' status of the component.
     *
     * @param	status   boolean parameter to set the 'enabled/disabled' status to
     */
	public void setEnabledStatus(boolean status)
	{
	     _enforcerEnabled = status;
	}

	/**
     * Return 'true/false', indicating the 'enabled/disabled' status of the component.
     *
     * @return	a boolean parameter indicating the 'enabled/disabled' status of
     *			this component.
     */
	public boolean getEnabledStatus()
	{
	     return _enforcerEnabled;
	}

	/**
     * Set the properties of this component, if there are any.
     *
     * @param	props ServiceProperties (optionally) specifying
     * 				  additional ontological properties of the component.
     *
     */
	public void setProperties(ServiceProperties props)
	{
    }

    //////////////////////////////////////////////////////////////
    //
    // Implementation of kaos.policy.enforcement.Enforcer - end
    //
    //////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Implementation of kaos.core.csi.policy.enforcement.CommunicationActionEnforcer - beginning
    //
    /////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Authorize a communication action described by properties expressed by the arguments to this method.
     * Authorization is based on the current set of policies stored in the Guard.
     * All arguments, except for the dsTransportName, contain the values of properties with the same name
     * as the name of the given argument.  Those names are defined in the ontology for CommunicationAction.
     *
     * @param performedBy           String containing the name of the actor performing the CommunicationAction.
     *
     * @param hasDestination        String containing the name of the entity receiving the CommunicationAction.
     *
     * @param carriesMessage        String containing the content of the CommunicationAction.
     *
     * @param hasPartner            String containing the name of the partner in the CommunicationAction.
     *
     * @param dsTransportName       String containing the name of the transport used by KAoS Policy service
     *                              and other KAoS services to communicate with each other.
     *
     * @throws KAoSSecurityException    when the CommunicationAction is forbidden by the current policies.
     *
     * @throws NullPointerException     when performedBy is null.
     *
     * @throws ServiceFailure           when the Policy service cannot be obtained, or fails durin action authorization.
     */

    public void authorizeCommunicationAction(String performedBy,
                                             String hasDestination,
                                             String carriesMessage,
                                             String hasPartner,
                                             String dsTransportName)
    throws KAoSSecurityException, NullPointerException, ServiceFailure
    {
        if ( !  _enforcerEnabled ) {
            try {
                initializeServices(dsTransportName);
            }
            catch(Exception e){
                e.printStackTrace();
                throw new ServiceFailure("Unable to obtain policy service!");
            }
        }

        if ( performedBy == null ){
            if (_logger.isEnabledFor(Logger.WARN))
   			    _logger.logMessage("authorizeCommunicationAction: actorID null, unable to authorize for:\n destination:\n "
   			                       + hasDestination + "\n carriesMessage: " + carriesMessage + "\n hasPartner: "
   			                       + hasPartner, Logger.WARN);
   			throw new NullPointerException();
        }

        HashMap actionProperties = new HashMap();

        if ( hasDestination != null ) {
            actionProperties.put(ActionConcepts.hasDestination(), hasDestination);
        }

        if ( carriesMessage != null ) {
            actionProperties.put(ActionConcepts.carriesMessage(), carriesMessage);
        }

        if ( hasPartner != null ) {
            actionProperties.put(GroupConcepts.hasPartner(), hasPartner);
        }

        try {
        	ActionInstanceDescription aid = new ActionInstanceDescriptionImpl(null,
                                                                              ActionConcepts.CommunicationAction(),
                                                                              performedBy,
                                                                              actionProperties);

            _policyChecking.checkPermission(aid,null);
        }
        catch(KAoSSecurityException kse){
            kse.printStackTrace();
            throw new KAoSSecurityException("\n\n  !!!! SECURITY VIOLATION !!!!\n\n  CommunicationAction NOT ALLOWED BETWEEN: \n\n  SENDER:   "
                                             + performedBy + "\n  RECEIVER: " + hasDestination + "\n\n");
        }
        catch(NullPointerException npe){
            throw npe;
        }
        catch(ServiceFailure sfe){
            throw sfe;
        }
        catch(URISyntaxException urie){
        	urie.printStackTrace();
            throw new ServiceFailure();
        }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////
    //
    // Implementation of kaos.core.csi.policy.enforcement.CommunicationActionEnforcer - end
    //
    /////////////////////////////////////////////////////////////////////////////////////////////

    // PROTECTED METHODS
    protected void initializeVars()
    {
        _enforcerID = _ENFORCER_ROOT + ExecEnvUtility.getVMID();
        addOntologicalAttribute(ActionConcepts.CommunicationAction());
        _enforcerEnabled = false;
    }

    protected void initializeServices(String transportName)
    throws Exception
    {
        CSIFactory.setCurrentTransport(transportName);
        _policyChecking = CSIFactory.getPolicyChecking();
        EnforcerManager polEnforcement = CSIFactory.getPolicyEnforcement();

        if ( polEnforcement.registerComponent(this) ) {
            _enforcerEnabled = true;
            return;
        }
        throw new Exception("Unable to register with EnforcerManager.  Enforcer name: "
                           + getName() + "\n controlled actions: " + getOntologicalAttributes());
    }
 }
