/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/04/03 19:32:26 $
 */

package kaos.core.csi.query;

import kaos.core.csi.query.QueryState.UnknownConceptException;
import kaos.core.csi.query.data.ActionClassDescription;

import javax.agent.service.directory.DirectoryFailure;
import java.util.Set;

/**
 * Defines the operations querying a ontology repository about <a href=http://ontology.coginst.uwf.edu/Actor.daml#Actor>Actors</a> <a
 * href=http://ontology.coginst.uwf.edu/Action.daml#Action>Actions</a> and <a href=http://ontology.coginst.uwf.edu/Entity.daml>Entities</a>.
 * 
 * @author KAoS Team
 * @version 1.0 $Revision: 1.3 $
 */
public interface QueryKAoSOntologyService
{
    /**
     * Get names of the individual Actors known to the ontology repository being an instance of the specified base class.
     * 
     * @param baseActorClass The name of the Actor class which the return elements have to be an instance of If the name is empty uses <a
     *            href=http://ontology.coginst.uwf.edu/Actor.daml#Actor>Actor</a> as a base class
     * @return Set of names of individual Actors
     * @exception UnknownConceptException if baseActorClass is unknown for the ontology repository
     */
    public Set<String> getIndividualActors(String baseActorClass) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get names of the Actors classes known to the ontology repository being a subclass of the specified base class.
     * 
     * @param baseActorClass The name of the Actor class which the return classes have to be a subclass of If the name is empty uses <a
     *            href=http://ontology.coginst.uwf.edu/Actor.daml#Actor>Actor</a> as a base class
     * @return Set of names of Actor subclasses
     * @exception UnknownConceptException if baseActorClass is unknown for the ontology repository
     */
    public Set<String> getClassesOfActors(String baseActorClass) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get names of the individual Places known to the ontology repository being an instance of the specified base class.
     * 
     * @param basePlaceClass The name of the Place class which the return elements have to be an instance of If the name is empty uses <a
     *            href=http://ontology.coginst.uwf.edu/Place.daml#Place>Place</a> as a base class
     * @return Set of names of individual Places
     * @exception UnknownConceptException if basePlaceClass is unknown for the ontology repository
     */
    public Set<String> getIndividualPlaces(String basePlaceClass) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get names of the Places classes known to the ontology repository being a subclass of the specified base class.
     * 
     * @param basePlaceClass The name of the Place class which the return classes have to be a subclass of If the name is empty uses <a
     *            href=http://ontology.coginst.uwf.edu/Place.daml#Place>Place</a> as a base class
     * @return Set of names of Place subclasses
     * @exception UnknownConceptException if basePlaceClass is unknow for the ontology repository
     */
    public Set<String> getClassesOfPlaces(String basePlaceClass) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get names of the individual Groups known to the ontology repository being an instance of the specified base class.
     * 
     * @param baseGroupClass The name of the Group class which the return elements have to be an instance of If the name is empty uses <a
     *            href=http://ontology.coginst.uwf.edu/Actor.daml#Group>Group</a> as a base class
     * @return Set of names of individual Groups
     * @exception UnknownConceptException if baseGroupClass is unknown for the ontology repository
     */
    public Set<String> getIndividualGroups(String baseGroupClass) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get names of the Group classes known to the ontology repository being a subclass of the specified base class.
     * 
     * @param baseGroupClass The name of the Group class which the return elements have to be an instance of If the name is empty uses <a
     *            href=http://ontology.coginst.uwf.edu/Actor.daml#Group>Group</a> as a base class
     * @return Set of names of individual Groups
     * @exception UnknownConceptException if baseGroupClass is unknown for the ontology repository
     */
    public Set<String> getClassesOfGroups(String baseGroupClass) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get names of the individual resources (Actors or Entities) known to the ontology repository being an instance of the specified base class.
     * 
     * @param baseResourceClass The name of the resources (<a href=http://ontology.coginst.uwf.edu/Actor.daml#Actor>Actor</a> or <a
     *            href=http://ontology.coginst.uwf.edu/Entity.daml#Entity>Entity</a>) class which the return classes have to be a subclass of
     * @return Set of names of individual targets (Actors or Entities)
     * @exception UnknownConceptException if baseResourceClass is unknown for the ontology repository
     */
    public Set<String> getIndividualResources(String baseResourceClass) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get names of the resources (Actors or Entities) classes known to the ontology repository being a subclass of the specified base class.
     * 
     * @param baseResourceClass The name of the resource (<a href=http://ontology.coginst.uwf.edu/Actor.daml#Actor>Actor</a> or <a
     *            href=http://ontology.coginst.uwf.edu/Entity.daml#Entity>Entity</a>) class which the return classes have to be a subclass of
     * @return Set of names of resource (Actors or Entities) subclasses
     * @exception UnknownConceptException if baseResourceClass is unknown for the ontology repository
     */
    public Set<String> getClassesOfResources(String baseResourceClass) throws UnknownConceptException, DirectoryFailure;
    
    /**
     * Get descriptions of the Action classes
     * 
     * @param actionClassName The name of the Action class
     * @return ActionClassDescription
     * @exception UnknownConceptException if forActor is unknown for the ontology repository 
     */
    public ActionClassDescription getActionDesc(String actionClassName) throws UnknownConceptException, DirectoryFailure;
    
    /*
     * Get names of the Action classes the specified Actor class or instance can perform based on its class.
     * @param forActor The name of the Actor class or instance
     * @return Set of String
     * @exception UnknownConceptException if forActor is unknown for the ontology repository
     */
    public Set<String> getActionsActorCanPerform(String forActor) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get descriptions of the Action classes the specified Actor class or instance can perform based on its class.
     * 
     * @param forActor The name of the Actor class or instance
     * @return Set of ActionClassDescription
     * @exception UnknownConceptException if forActor is unknown for the ontology repository 
     */
    public Set<ActionClassDescription> getActionsDescActorCanPerform(String forActor) throws UnknownConceptException, DirectoryFailure;
}