/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/04/02 07:12:11 $
 */

package kaos.core.csi.query;

import kaos.core.csi.ontology.OntologyService;
import kaos.core.service.directory.DomainDescription;
import kaos.core.service.directory.KAoSAgentDescription;
import kaos.core.csi.query.QueryState.UnknownConceptException;

import javax.agent.service.directory.DirectoryFailure;
import javax.agent.service.directory.SearchException;
import java.util.List;
import java.util.Map;

/**
 * Defines the operations to query the different aspects of domain and actor registrations
 * 
 * @author KAoS Team $Revision: 1.16 $
 */
public interface QueryRegistration
{
    /**
     * Returns description of the domain identified by the given name
     * 
     * @param name String specifying the name of the domain to lookup
     * @return DomainDescription identified by the given name
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     */
    public DomainDescription getDomain(String name) throws DirectoryFailure;

    /**
     * Returns the id of all domains that have no parent domain
     * 
     * @return List of String identifying by the root domains
     */
    public List<String> getRootDomains() throws DirectoryFailure;
    
    /**
     * Returns description of the actor identified by the given name
     * 
     * @param actorGUID String specifying the GUID of the actor to lookup
     * @return KAoSAgentDescription identified by the given name
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws SearchException if the actor with the given name has not been found in the DS
     */
    public KAoSAgentDescription getActor(String actorGUID) throws DirectoryFailure, SearchException;

    /**
     * Returns description of the actors fulfilling constraints expressed as 
     * the <a href="http://www-ksl.stanford.edu/knowledge-sharing/kif/">KIF</a> statement
     * 
     * @param actorConstraints String specifying the GUID of the actor to lookup An example of a Constraints statement: 
     *            Give me all agents that have as capability 'take picture'  with a cost of less than $1,- with a reliability of 75% 
     *            or better and where the average number of processed messages in the past hour is less than 20.
     * @return KAoSAgentDescription identified by the given name
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws NoSuchActorException - if no actors matching the query properties could be found
     */
    public List getActorsFulfilling(String actorConstraints) throws DirectoryFailure, NoSuchActorException;

    /**
     * Returns descriptions of the actors whose KAoSAgentDescription contains all specified attributes and their values.
     * 
     * @param attributes Mpa containing attribute names and their values to match the attributes in the DS.
     * @return List containing KAoSAgentDescriptions matching the attributes.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws SearchException 
     */
    public List getActorsWithAttributes(Map attributes) throws DirectoryFailure, SearchException;

    /**
     * Return all agent ids stored in the KAoS directory.
     * 
     * @return List of all agent unique IDs in the agent directory, or null, if there aren't any.
     * @throws SearchException if an exception is thrown during the search.
     * @throws DirectoryFailure if the directory cannot be found/is not available.
     */
    public List getAllAgents() throws SearchException, DirectoryFailure;

    /**
     * Get the list of Actors assigned as leaders for the given Group.
     * 
     * @param groupURI - URI for the Group individual for which the members will be returned
     * @return List<KAoSAgentDescription> for each Actor assigned as a member for the given Group
     * @throws DirectoryFailure
     * @throws SearchException 
     */
    public List getGroupMembers(String groupURI) throws DirectoryFailure, SearchException;

    /**
     * Get the list of Actors assigned as leaders for the given Group.
     * 
     * @param groupURI - URI for the Group individual for which the leaders will be returned
     * @return List<KAoSAgentDescription> for each Actor assigned as a leader for the given Group
     * @throws DirectoryFailure
     * @throws SearchException 
     */
    public List getGroupLeaders(String groupURI) throws DirectoryFailure, SearchException;

    /**
     * Get the Groups led by the given Actor
     * 
     * @param actorId - id of the Actor (Agent)
     * @return List<String> URI of each Group led by the given Actor
     * @throws DirectoryFailure
     * @throws UnknownConceptException 
     * @throws OntologyService.ReasoningException 
     */
    public List getGroupsLedBy(String actorId) throws DirectoryFailure, OntologyService.ReasoningException, UnknownConceptException;

    /**
     * Get the vector of subdomains assigned to this domain.
     * 
     * @param domain - Domain name string
     * @return List<KAoSAgentDescription> for each Actor assigned as a leader for the given Group
     * @throws DirectoryFailure
     */
    public List getSubdomains(String domain) throws DirectoryFailure;
    
    /**
     *   NoSuchActorException defines an exception to be thrown by the QueryRegistration 
     *   when no actor matching the query properties is found.
     */   
    public class NoSuchActorException  extends Exception 
    {

        /**
         *   Constructs a new NoSuchActorException with no detailed message.
         */
        public NoSuchActorException () 
        {    
            super(); 
        }

        /**
         *   Constructs a new NoSuchActorException with a detail message.
         *   A detail message is a String that describes the exception.
         */
        public NoSuchActorException (String s) 
        { 
            super(s); 
        }
    }
}
