/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/03/28 20:37:57 $
 */

package kaos.core.csi.query;

import kaos.core.csi.ontology.OntologyService;
import kaos.core.csi.ontology.OntologyService.ReasoningException;
import kaos.core.service.boot.KAoSGuardBootstrapper;
import kaos.core.service.directory.*;
import kaos.core.util.Logger;
import kaos.core.util.NamespaceValidator;
import kaos.core.csi.query.QueryState.UnknownConceptException;
import kaos.ontology.vocabulary.GroupConcepts;
import kaos.ontology.vocabulary.KAoSSystemConcepts;

import javax.agent.service.directory.AgentDescription;
import javax.agent.service.directory.DirectoryFailure;
import javax.agent.service.directory.SearchException;
import java.util.*;

/**
 * Defines the operations to query the different aspects of domain and actor registrations
 * 
 * @author KAoS Team $Revision: 1.27 $
 */
public class QueryRegistrationImpl implements QueryRegistration
{
    private static final QueryRegistrationImpl __INSTANCE    = new QueryRegistrationImpl();
    private static boolean                     __INITIALIZED = false;

    private DirectoryService                   _dirProxy = null;
    private Logger                             _logger       = Logger.getLogger(this);

    /**
     * Retrieve an instance of Registration.
     * 
     * @param transportName String specifying the name of the transport to use for communication with the KAoS Directory Service.
     * @return a singleton, Registration.
     */
    public static QueryRegistration getInstance(String transportName) throws Exception
    {
        if (__INITIALIZED)
            return __INSTANCE;

        __INSTANCE.initialize(transportName);
        __INITIALIZED = true;
        return __INSTANCE;
    }

    private QueryRegistrationImpl()
    {
    }

    /*****************************************************************************************
     * kaos.core.csi.query.QueryRegistration interface implementation - begin
     *****************************************************************************************/
    /**
     * Returns description of the domain identified by the given name
     * 
     * @param name String specifying the name of the domain to lookup
     * @return DomainDescription identified by the given name
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws DirectoryFailure if the domain with the given name has not been registered in the DS
     */
    public DomainDescription getDomain(String name) throws DirectoryFailure
    {
        name = NamespaceValidator.validateURI(name);
        DomainDescription desc = _dirProxy.getDomainDescription(name);
        
        if (desc == null)
            throw new DirectoryFailure();
        
        return desc;
    }

    
    public List<String> getRootDomains() throws DirectoryFailure
    {
    	return _dirProxy.getRootDomains();
    }
    /**
     * Returns description of the actor identified by the given name
     * 
     * @param actorGUID String specifying the GUID of the actor to lookup
     * @return KAoSAgentDescription identified by the given name
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws SearchException if the actor with the given name has not been registered in the DS
     */
    public KAoSAgentDescription getActor(String actorGUID) throws DirectoryFailure, SearchException
    {
        actorGUID = NamespaceValidator.validateURI(actorGUID);
        KAoSAgentDescription desc = null;
        desc = _dirProxy.getAgentDescriptionForAgentID(actorGUID);
        
        desc.addDirProxy(_dirProxy);
        
        return desc;
    }

    /**
     * Return all agent ids stored in the KAoS directory.
     * 
     * @return List of all agent unique IDs in the agent directory, or null, if there aren't any.
     * @throws SearchException if an exception is thrown during the search.
     * @throws DirectoryFailure if the directory cannot be found/is not available.
     */
    public List getAllAgents() throws SearchException, DirectoryFailure
    {
        List agentDescList = _dirProxy.getAllAgentsInDirectory();
        
        if (agentDescList == null)
            return null;
        
        Vector agentIDs = new Vector(agentDescList.size());
        
        for (Iterator agentDescListIt = agentDescList.iterator(); agentDescListIt.hasNext();)
        {
            KAoSAgentDescription agentDesc = (KAoSAgentDescription)agentDescListIt.next();

            Map ontTypes = agentDesc.getEntityOntologicalTypes();
            if (ontTypes.containsKey(KAoSSystemConcepts.KAoSDirectoryService()))
                continue;

            String agentID = agentDesc.getEntityNameAsString();
            
            if (agentID != null)
                agentIDs.add(agentID);
        }
        
        if (_logger.isEnabledFor(Logger.INFO))
            logMessage(".getAllAgents: returning \n" + agentIDs, Logger.INFO);
        
        return agentIDs;
    }

    /**
     * Returns description of the actors fulfilling constraints expressed as the 
     * <a href="http://www-ksl.stanford.edu/knowledge-sharing/kif/">KIF</a> statement
     * 
     * @param actorConstraints String specifying the GUID of the actor to lookup An example of a Constraints statement: Give me all agents that have as capability 'take picture'
     *            with a cost of less than $1,- with a reliability of 75% or better and where the average number of processed messages in the past hour is less than 20.
     * @return KAoSAgentDescription identified by the given name
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws NoSuchActorException - if no actors matching the query properties could be found
     */
    public List getActorsFulfilling(String actorConstraints) throws DirectoryFailure, NoSuchActorException
    {
        return new ArrayList();
    }

    /**
     * Returns descriptions of the actors whose KAoSAgentDescription contains all specified attributes and their values.
     * 
     * @param attributes Map containing attribute names and their values to match the attributes in the DS.
     * @return List containing KAoSAgentDescriptions matching the attributes.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws QueryFailure if query failed
     */
    public List getActorsWithAttributes(Map attributes) throws DirectoryFailure, SearchException
    {
        List descList = new ArrayList();
        DefaultKAoSAgentDescription searchDesc = new DefaultKAoSAgentDescription();
        searchDesc.set(KAoSEntityDescription.SEARCH_PARAMS, attributes);

        AgentDescription[] results = null;

        try
        {
            results = _dirProxy.search(searchDesc);
        }
        catch (DirectoryFailure df)
        {
            throw df;
        }

        if ((results == null) || results.length == 0)
            return descList;

        for (int i = 0; i < results.length; i++)
            if (results[i] instanceof KAoSAgentDescription)
                descList.add((KAoSAgentDescription)results[i]);

        if (_logger.isEnabledFor(Logger.INFO))
            logMessage("getActorsWithAttributes: returning List : " + descList, Logger.INFO);

        return descList;
    }

    /**
     * Get the list of KAoSAgentDescriptions for the Actors assigned as members for the given Group.
     * 
     * @param groupURI - URI for the Group individual for which the members will be returned
     * @return List<KAoSAgentDescription> for each Actor assigned as a member for the given Group
     * @throws DirectoryFailure
     * @throws SearchException 
     * @throws QueryFailure
     */
    public List getGroupMembers(String groupURI) throws DirectoryFailure, SearchException
    {
        Hashtable attribs = new Hashtable(1);
        Vector teamNames = new Vector();
        
        groupURI = NamespaceValidator.validateURI(groupURI);
        teamNames.add(groupURI);
        attribs.put(KAoSEntityDescription.ENTITY_DOMAIN_NAMES, teamNames);
        
        return getActorsWithAttributes(attribs);
    }

    /**
     * Get the list of KAoSAgentDescriptions for the Actors assigned as leaders for the given Group.
     * 
     * @param groupURI - URI for the Group individual for which the leaders will be returned
     * @return List<KAoSAgentDescription> for each Actor assigned as a leader for the given Group
     * @throws DirectoryFailure
     * @throws SearchException 
     */
    public List getGroupLeaders(String groupURI) throws DirectoryFailure, SearchException
    {
        groupURI = NamespaceValidator.validateURI(groupURI);
        Hashtable attribs = new Hashtable();
        
        attribs.put(GroupConcepts.isLeaderOf(), groupURI);
        
        return getActorsWithAttributes(attribs);
    }

    /**
     * Get the URI's for all Groups led by the given Actor
     * 
     * @param actorId - id of the Actor (Agent)
     * @return List<String> URI of each Group led by the given Actor
     * @throws DirectoryFailure
     * @throws UnknownConceptException 
     * @throws OntologyService.ReasoningException 
     */
    public List getGroupsLedBy(String actorId) throws DirectoryFailure, OntologyService.ReasoningException, UnknownConceptException
    {
        actorId = NamespaceValidator.validateURI(actorId);
        Set s = _dirProxy.getValuesForProperty(GroupConcepts.isLeaderOf(), actorId);
        ArrayList l = new ArrayList(s);
        
        return l;
    }

    /**
     * Get the vector of subdomains assigned to this domain.
     * 
     * @param domain - Domain name string
     * @return Vector<KAoSAgentDescription> for each Actor assigned as a leader for the given Group
     * @throws DirectoryFailure
     */
    public List<String> getSubdomains(String domain) throws DirectoryFailure
    {
        return _dirProxy.getSubdomains(domain);
    }

    /*****************************************************************************************
     * kaos.core.csi.query.QueryRegistration interface implementation - end
     *****************************************************************************************/
   
    private void initialize(String transportName) throws Exception
    {
        try
        {
            _dirProxy = KAoSGuardBootstrapper.getKAoSDirectoryProxyForTransport(transportName);
        }
        catch (Exception e)
        {
            // e.printStackTrace();
            throw new Exception("    \n\n    *** QueryStateImpl: unable to obtain a reference the DirectoryService for transport: " + transportName + "!");
        }
    }

    protected void logMessage(String msg, int logLevel)
    {
        _logger.logMessage(msg, logLevel);
    }

    protected void logMessage(String msg, Exception e, int logLevel)
    {
        _logger.logMessage(msg, e, logLevel);
    }
}
