/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2012/04/05 05:16:55 $
 */
 
package kaos.core.csi.query;

import kaos.core.csi.OntInstanceDescription;
import kaos.core.csi.ontology.OntologyService;
import kaos.core.csi.ontology.OntologyService.ReasoningException;
import kaos.core.csi.query.data.OntClassDescription;

import javax.agent.service.directory.DirectoryFailure;

import java.util.List;
import java.util.Set;

/**
 * Defines the operations to query the different aspect of state stored in the Directory Service 
 * 
 * @author KAoS Team
 * $Revision: 1.12 $
 */
public interface QueryState
{   
    /**
     * This method returns a set of urls depicting ontologies loaded into the Ontology Service
     */
    public Set<String> getListOfLoadedOntologies() throws DirectoryFailure;

    /**
     * This method returns a set of urls depicting ontologies loaded into the repository as well namespaces used internally by KAoS.
     * @throws DirectoryFailure 
     */
    public Set<String> getAllNamespaces() throws DirectoryFailure;
    
    /**
     * Test if triple is true according to the ontologies in the ontology repository.
     * 
     * @param subject of the triple
     * @param predicate of the triple
     * @param object of the triple
     * @return true if the triple is true and false in the opposite case
     * @throws OntologyService.ReasoningException if reasoning failed.
     */
    public boolean testTrue(String subject, String predicate, String object) 
        throws OntologyService.ReasoningException, UnknownConceptException, DirectoryFailure;

    /**
     * Get values of the selected variable satisfying the <a href="http://www-ksl.stanford.edu/knowledge-sharing/kif/">KIF</a> 
     * query with <b>possibly many variable</b>.
     * 
     * @param variableName the name of the selected variable
     * @param multiplyVariablesQuery the query
     * @return a set with values satisfying the query
     * @throws OntologyService.ReasoningException if reasoning failed.
     */
    public Set<String> askQuery(String variableName, String multiplyVariablesQuery) throws OntologyService.ReasoningException, DirectoryFailure;

    /**
     * Get a single value of the variable satisfying 
     * the <a href="http://www-ksl.stanford.edu/knowledge-sharing/kif/">KIF</a> query: <b>(property subject ?val)</b>.
     * 
     * @param property the name of the property
     * @param subject the name of the subject
     * @return a single value for the property on this subject
     * @throws OntologyService.ReasoningException if reasoning failed.
     */
    public String getUniqueValueForProperty(String property, String subject) 
        throws OntologyService.ReasoningException, UnknownConceptException, DirectoryFailure;

    /**
     * Get values of the variable satisfying 
     * the <a href="http://www-ksl.stanford.edu/knowledge-sharing/kif/">KIF</a> query: <b>(property subject ?val)</b>.
     * 
     * @param property the name of the property
     * @param subjectStr the name of the subject
     * @return a set with values for the property on this subject
     * @throws OntologyService.ReasoningException if reasoning failed.
     */
    public Set<String> getValuesForProperty(String property, String subjectStr) 
        throws OntologyService.ReasoningException, UnknownConceptException, DirectoryFailure;

    /**
     * Get values of the variable satisfying 
     * the <a href="http://www-ksl.stanford.edu/knowledge-sharing/kif/">KIF</a> query: <b>(property ?subject value)</b>.
     * 
     * @param property the name of the property
     * @param value the value
     * @return a set with name of subjects which has the specified value for the given property
     * @throws OntologyService.ReasoningException if reasoning failed.
     */
    public Set<String> getResourcesWithValueForProperty(String property, String value) 
        throws OntologyService.ReasoningException, UnknownConceptException, DirectoryFailure;

    /**
     * Get properties applicable to the given class
     * 
     * @param className the name of the class
     * @return vector of properties names
     * @throws OntologyService.ReasoningException if reasoning failed.
     */
    public List<String> getPropertiesApplicableTo(String className) throws OntologyService.ReasoningException, UnknownConceptException, DirectoryFailure;

    /**
     * Get the name of the class defining the range of the specified property for the given class.
     * 
     * @param className the name of the class
     * @param propertyName the name of the property
     * @return the name of the class defining the range
     * @throws OntologyService.ReasoningException if reasoning failed.
     */
    public String getRangeOnPropertyForClass(String className, String propertyName) 
        throws OntologyService.ReasoningException, UnknownConceptException, DirectoryFailure;

    /*
     * Get description of the class including its ranges of properties
     * 
     * @param className the name of the class 
     * @return OntClassInfo containing the description
     * @throws UnknownConceptException, DirectoryFailure
     */
    public OntClassDescription getClassDesc(String className) throws UnknownConceptException, DirectoryFailure, ReasoningException;
    
    /**
     * Finds the Ontology description of the provided individual
     * 
     * @param individualName the name of the individual
     * @return OntInstanceDescription
     * @throws UnknownConceptException if concept unknown
     */
    public OntInstanceDescription getIndividualDesc(String individualName) throws UnknownConceptException, DirectoryFailure;
    
    /**
     * Finds the ultimate Ontology type of the provided individual, from the ontologies explicitly loaded into repository
     * 
     * @param individualName the name of the individual
     * @return the vector of names of the Ontology classes being the most narrow ones for the given individual
     * @throws UnknownConceptException if concept unknown
     */
    public Set<String> getOntologyTypesOfIndividual(String individualName) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get set of instances of the given concept; if the concept is an instance itself the set contains just this instance.
     * 
     * @param conceptName The name of the concept in the Jena format
     * @return Set of instance ids in the Jena format
     * @exception UnknownConceptException if forActor is unknown for the ontology repository
     */
    public Set<String> getInstancesOf(String conceptName) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get set of superclasses of the given class.
     * 
     * @param className The name of the class in Jena format
     * @return Set of superclasses names in Jena format
     * @exception UnknownConceptException if className is unknown for the ontology repository
     */
    public Set<String> getSuperClassesOf(String className) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get set of subclasses of the given class.
     * 
     * @param className The name of the class in Jena format
     * @return Set of subclasses names in Jena format
     * @exception UnknownConceptException if className is unknown for the ontology repository
     */
    public Set<String> getSubClassesOf(String className) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get set of immediate subclasses of the given class (direct children of the class).
     * 
     * @param className The name of the class in Jena format
     * @return Set of subclasses names in Jena format
     * @throws UnknownConceptException
     * @exception UnknownConceptException if className is unknown for the ontology repository
     */
    public Set<String> getImmediateSubClassesOf(String className) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get set of super properties of the given property.
     * 
     * @param propertyName The name of the property in Jena format
     * @return Set of super properties names in Jena format
     * @exception Exception if propertyName is unknown for the ontology repository
     */
    public Set<String> getSuperPropertiesOf(String propertyName) throws UnknownConceptException, DirectoryFailure;

    /**
     * Get set of subproperties of the given property.
     * 
     * @param propertyName The name of the property in Jena format
     * @return Set of subproperties names in Jena format
     * @exception Exception if propertyName is unknown for the ontology repository
     */
    public Set<String> getSubPropertiesOf(String propertyName) throws UnknownConceptException, DirectoryFailure;

    /**
     * Test if two provided classes are disjoint
     * 
     * @param firstClass the name of the first class
     * @param secondClass the name of the second class
     * @return true if the classes are not disjoint
     * @throws UnknownConceptException 
     */
    public boolean areClassesDisjoint(String firstClass, String secondClass) 
        throws OntologyService.ReasoningException, DirectoryFailure, UnknownConceptException;
    
    /**
     * Get set of classes defined in the given namespace.
     * 
     * @param namespaceName The name of the namespace in the Jena format
     * @return Set of classes names in the Jena format
     * @exception UnknownNamespaceException if namespaceName is unknown for the ontology repository
     */
    public Set<String> getClassesDefInNamespace(String namespaceName) throws UnknownNamespaceException, DirectoryFailure;

    /**
     * Get set of properties defined in the given namespace.
     * 
     * @param namespaceName The name of the namespace in the Jena format
     * @return Set of properties names in the Jena format
     * @exception UnknownNamespaceException if namespaceName is unknown for the ontology repository
     */
    public Set<String> getPropertiesDefInNamespace(String namespaceName) throws UnknownNamespaceException, DirectoryFailure;

    /**
     * Get set of instances defined in the given namespace.
     * 
     * @param namespaceName The name of the namespace in the Jena format
     * @return Set of instances names in the Jena format
     * @exception UnknownNamespaceException if namespaceName is unknown for the ontology repository
     */
    public Set<String> getInstancesDefInNamespace(String namespaceName) throws UnknownNamespaceException, DirectoryFailure;

    /**
     * Get set of namespaces imported by the given namespace.
     * 
     * @param namespaceName The name of the namespace in the Jena format
     * @return Set of namespaces names in the Jena format
     * @exception UnknownNamespaceException if namespaceName is unknown for the ontology repository
     */
    public Set<String> getNamspacesImportedByNamespace(String namespaceName) throws UnknownNamespaceException, DirectoryFailure;

    /**
     * Get mapping to the urls of ontology concept for the given concept name. 
     * Currently, it tries to match the concept name with the local name of the ontology definitions url.
     * 
     * @param conceptName The name of the namespace in the Jena format
     * @return Set of ontology definition url potentially matching the search concept
     */
    public Set<String> getOntologyDefinitionForConcept(String conceptName) throws UnknownNamespaceException, DirectoryFailure;
    
    /**
     *   UnknownNamespaceException defines an exception to be thrown when a provided namespace is not know for the ontology repository 
     **/
    public class UnknownNamespaceException extends Exception 
    {
        private static final long serialVersionUID = 2182972964812555580L;
        public String unknownNamespace = null;

        /**
         *   Constructs a new UnknownNamespaceException with information about the unknown namespace.
         */
        public UnknownNamespaceException (String _unknownNamespace) 
        {    
            super("The unknown namespace was: " + _unknownNamespace); 
            
            unknownNamespace = _unknownNamespace;
        }

        /**
         *   Constructs a new UnknownNamespaceException with a detail message.
         *   A detail message is a String that describes the exception.
         */
        public UnknownNamespaceException (String s, String _unknownNamespace) 
        {     
            super(s + "; the unknown namespace was: " + _unknownNamespace); 
            
            unknownNamespace = _unknownNamespace;
        }
    }

    /**
     * UnknownConceptException defines an exception to be thrown when a provided name of the concept is not know for the ontology repository
     **/
    public class UnknownConceptException extends Exception
    {
        private static final long serialVersionUID = 8227045952445163547L;

        public String             unknownConcept   = null;

        /**
         * Constructs a new UnknownConceptException with information about the unknown concept.
         */
        public UnknownConceptException(String _unknownConcept)
        {
            super("The unknown concept was: " + _unknownConcept);

            unknownConcept = _unknownConcept;
        }

        /**
         * Constructs a new UnknownConceptException with a detail message.
         * A detail message is a String that describes the exception.
         */
        public UnknownConceptException(String s, String _unknownConcept)
        {
            super(s + "; the unknown concept was: " + _unknownConcept);

            unknownConcept = _unknownConcept;
        }
    }
}