/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/04/14 05:50:40 $
 */
package kaos.core.csi.query.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;

import kaos.core.csi.CSIFactory;
import kaos.core.csi.OntInstanceDescription;
import kaos.core.csi.OntPropertyDescription;
import kaos.core.csi.extension.classification.ClassifierException;
import kaos.core.csi.extension.classification.InstanceClassifier;
import kaos.ontology.classifier.GenericOntClassifier;
import kaos.ontology.util.KAoSOntClassBuilder;
import kaos.policy.guard.KAoSGuardPolicyProcessor;
import kaos.policy.util.SerializableOntDescription;
import kaos.xmlschema.classifier.XMLSchemaClassifier;

/*
 * This class represents a precomputed ontology class
 */
public class OntClassDescription implements SerializableOntDescription
{
    private static final long                  serialVersionUID        = 1L;

    private   String                                         _className              = null;
    private   Set<String>                                    _classTypes             = null;
    private   Set<String>                                    _classSubclasses        = null;
    protected Map<String, OntClassPropertyDescription>       _classProperties        = null;
    private   Map<String, List<OntClassPropertyDescription>> _classRangeToProperties = null;
    private   Map<String, OntClassDescription>               _classSuperclasses      = null;

    public OntClassDescription(String className)
    {
        _className = className;
        _classTypes = new HashSet<String>();
        _classSubclasses = new HashSet<String>();
        _classProperties = new Hashtable<String, OntClassPropertyDescription>();
        _classRangeToProperties = new Hashtable<String, List<OntClassPropertyDescription>>();
        _classSuperclasses = new Hashtable<String, OntClassDescription>();
    }

    public String getClassName()
    {
        return _className;
    }

    // ------------- Class types and their subtypes

    public void setMainSuperClassName(String classType)
    {
        _classTypes.add(classType);
    }

    public String getMainSuperClassName()
    {
        if (_classTypes.isEmpty())
            return null;

        return (String)_classTypes.toArray()[0];
    }

    public void setClassTypes(List<String> types)
    {
        _classTypes.addAll(types);
    }

    public Set<String> getClassTypes()
    {
        return _classTypes;
    }

    /*
     * Update the subclasses of this class.
     * @param className - the class whose subclasses are updated
     * @param subclasses - the set of updated subclasses
     */
    public void updateClassSubclasses(String className, Set<String> subclasses)
    {
        if (!_className.equals(className))
            return;

        _classSubclasses = subclasses;
    }

    public Set<String> getSubclasses()
    {
        return _classSubclasses;
    }

    public void setAnnotationFor(String annotationType, String annotationed, String annotationValue)
    {

    }

    public String getAnnotationFor(String annotationType, String annotationed)
    {
        return null;
    }

    // --- Properties

    public void addProperty(OntClassPropertyDescription currentProperty)
    {
        String name = currentProperty.getPropertyName();
        _classProperties.put(name, currentProperty);

        String className = currentProperty.getParentRangeName();

        if (className != null)
        {
            if (_classRangeToProperties.get(className) == null)
                _classRangeToProperties.put(className, new ArrayList<OntClassPropertyDescription>());

            _classRangeToProperties.get(className).add(currentProperty);
        }
    }

    /*
     * Remove the property with the given name from the set of this instance's properties. Only looks at properties of this instance and not properties of
     * instances nested within this one
     * @param propertyName - URI for the name of the property to remove
     */
    public void removeProperty(String propertyName)
    {
        OntClassPropertyDescription prop = (OntClassPropertyDescription)_classProperties.remove(propertyName);

        if (prop == null)
            return;

        String className = prop.getParentRangeName();

        if (className != null)
            if (_classRangeToProperties.get(className) != null)
                _classRangeToProperties.get(className).remove(prop);
    }

    public OntClassPropertyDescription getProperty(String propertyName)
    {
        return (OntClassPropertyDescription)_classProperties.get(propertyName);
    }

    /*
     * Get all of the properties for this instance. Excluding nested properties
     * @return
     */
    public Collection<OntClassPropertyDescription> getAllProperties()
    {
        return _classProperties.values();
    }

    /*
     * This function returns the number of properties held in this description.
     */
    public int numberOfProperties()
    {
        return _classProperties.keySet().size();
    }

    public Set<String> getPropertyNames()
    {
        return _classProperties.keySet();
    }

    public boolean isPropertyPresent(String tRoleName)
    {
        return _classProperties.keySet().contains(tRoleName);
    }

    public Set<String> getInstancesForProperty(String tRoleName)
    {
        OntClassPropertyDescription targetInfo = (OntClassPropertyDescription)_classProperties.get(tRoleName);

        if (targetInfo != null)
            return targetInfo.getAllInstances();

        return null;
    }

    /*
     * This method returns all the properties and their subproperties contained in this class.
     */
    public List<String> getPropretiesOfInterest()
    {
        List<String> ret = new ArrayList<String>();

        for (String currentProperty : _classProperties.keySet())
        {
            OntClassPropertyDescription prop = getProperty(currentProperty);
            ret.add(currentProperty);

            if (prop.getAllSubpropertyNames() != null)
                ret.addAll(prop.getAllSubpropertyNames());
        }

        return ret;
    }

    /*
     * This method returns all the class ranges of properties in this class.
     */
    public List<String> getClassRangesOfInterest()
    {
        List<String> ret = new ArrayList<String>();

        for (String currentTarget : _classProperties.keySet())
        {
            OntClassPropertyDescription prop = getProperty(currentTarget);
            String className = prop.getParentRangeName();

            if (className != null)
                ret.add(className);
        }

        return ret;
    }

    /*
     * This method returns all the subclasses of this class, along with the class name.
     */
    public List<String> getActionSubclassesOfInterest()
    {
        List<String> ret = new ArrayList<String>();

        ret.addAll(_classTypes);

        return ret;
    }

    // --- Caches update

    /*
     * Update the cached subproperties of the given property
     * @param propName the given property
     * @param subProps names of its subproperties
     */
    public void updateSubpropertyNamesForProperty(String propName, Collection<String> subProps)
    {
        OntClassPropertyDescription prop = (OntClassPropertyDescription)_classProperties.get(propName);

        if (prop != null)
            prop.updateSubpropertyNames(propName, subProps);
    }

    /*
     * Update the class instances of a property contained in this instance
     */
    public void updateInstancesForClass(String classRangeName, Collection<String> inst)
    {
        // Get the corresponding property from the class name
        List<OntClassPropertyDescription> propertiesWithThisRange = _classRangeToProperties.get(classRangeName);

        if (propertiesWithThisRange != null)
            for (OntClassPropertyDescription currentProperty : propertiesWithThisRange)
                currentProperty.updateInstances(inst);
    }

    // ---- Get as Text

    public String toString()
    {
        StringBuffer buffer = new StringBuffer();
        buffer.append("\n Class Name: ");
        buffer.append(this.getClassName());
        buffer.append("\n Class Types: [" + _classTypes.size() + "]");
        if (_classTypes.size() > 0)
            buffer.append("\t first type = " + _classTypes.toArray()[0]);
        buffer.append("\n with properties: \n");

        for (String propertyName : _classProperties.keySet())
        {
            buffer.append(propertyName + " ");
            buffer.append(_classProperties.get(propertyName));
            buffer.append("\n");
        }

        return buffer.toString();
    }

    // -------------------

    public int matchInstanceWithClass(OntInstanceDescription testedInstance, Set<String> missingPropertiesInInstance)
    {
        boolean partial = false; // Indicates if some of the expected properties having restriction (allValues) in this class are missing in the instance
        if (missingPropertiesInInstance == null)
            missingPropertiesInInstance = new HashSet<String>();

        // Match types
        String matchedActionInstanceBaseType = testedInstance.getFirstType(); // TODO: consider other types of the instance if present
        if ( ! this.getClassTypes().contains(matchedActionInstanceBaseType)) 
        {
            if (this.getSubclasses() == null)
                return KAoSGuardPolicyProcessor._NO_MATCH_;
            
            if ( ! this.getSubclasses().contains(matchedActionInstanceBaseType))
                return KAoSGuardPolicyProcessor._NO_MATCH_;
        }

        // Match properties
        for (OntClassPropertyDescription currentClassProperty : getAllProperties())
        {
            OntPropertyDescription instanceProperty = testedInstance.getProperty(currentClassProperty.getPropertyName());
            
            // Find if tested instance has the current property of the class
            if (instanceProperty == null)
                // See if the action instance has one of class property's sub-property
                for (String currentSubproperty : currentClassProperty.getAllSubpropertyNames())
                {
                    instanceProperty = testedInstance.getProperty(currentSubproperty);

                    if (instanceProperty != null)
                        break;
                }
            
            // Check if the current property is found in the instance and if it has value(s) assigned
            if ((instanceProperty == null) || (instanceProperty.getValues().isEmpty()))
            {
                // If the type of the property is the someValues then lack of value(s) is no satisfying it (even if complement)
                if (currentClassProperty.getRangeRestrictionType().equals(KAoSOntClassBuilder._someValuesFromRestriction))
                    return KAoSGuardPolicyProcessor._NO_MATCH_;
                else
                {
                    // TODO: This is a current trick assuming that if user specified the allValues restrictions then it also meant requirement (someValues)
                    // we assume that cardinality  at least on - check this
                    partial = true; // Even that no restriction is violated still we lack presumed information to check the allValues restriction
                    missingPropertiesInInstance.add(currentClassProperty.getPropertyName());

                    continue; // Go to the next class property
                }
            }

            // Test if the instance property values satisfy restrictions on the class property
            boolean testPropertyPresenceInAIDResult = matchInstanceProperty(currentClassProperty, instanceProperty);

            if (testPropertyPresenceInAIDResult)
                continue; // Found match thus continue to check the next class property
            else
                return KAoSGuardPolicyProcessor._NO_MATCH_;
        }

        // Finished with proceeding properties
        if (partial)
            return KAoSGuardPolicyProcessor._PARTIAL_MATCH_;
        else
            return KAoSGuardPolicyProcessor._FULL_MATCH_;
    }

    /*
     * @return true if instance property satisfy the class property range
     */
    private boolean matchInstanceProperty(OntClassPropertyDescription classProperty, OntPropertyDescription instanceProperty)
    {
        if ((classProperty == null) || (instanceProperty == null))
            return false;

        Set<String> classPropertyCachedInstances = classProperty.getAllInstances();

        int countOfMatchedInstanceValues = 0;

        // Check each value of the instance property
        for (String currentValueOfInstanceProperty : instanceProperty.getValues())
            if (classProperty.isDatatypeProperty()) // The property is datatype
            {
                InstanceClassifier classifier = XMLSchemaClassifier.getInstance();
                String currentOntPropertyRange = classProperty.getParentRangeName();
                Object currentValueOfInstancePropertyDesc = instanceProperty.getValueDescForValue(currentValueOfInstanceProperty);

                // Use the XML Schema classifier
                try
                {
                    List<String> mySchema = classProperty.getSchema();
                    
                    if (classifier.classify(currentOntPropertyRange, mySchema, currentValueOfInstanceProperty, currentValueOfInstancePropertyDesc))
                        countOfMatchedInstanceValues++; // It is a match
                }
                catch (ClassifierException e)
                {
                    // TODO: print the exception
                    return false;
                }
            }
            else
            {
                Object currentValueOfInstancePropertyDesc = instanceProperty.getValueDescForValue(currentValueOfInstanceProperty);
                
                // Check if the current value is in the set of cached values for this property range in the class
                if (classPropertyCachedInstances.contains(currentValueOfInstanceProperty))
                    countOfMatchedInstanceValues++;
                else if (currentValueOfInstancePropertyDesc != null) // There is a description for the current value of the instance property
                {
                    // Check if a classifier can be used
                    InstanceClassifier classifier = null;
                    String currentOntPropertyRange = classProperty.getParentRangeName();
                    
                    // Check if the property range has a specialized classifier associated with it
                    if (CSIFactory.getClassifierManagement() != null)
                        classifier = CSIFactory.getClassifierManagement().getClassifier(currentOntPropertyRange);

                    // If no specialized classifier then use a generic one
                    if (classifier == null)
                        classifier = GenericOntClassifier.getInstance();

                    // Use the found classifier
                    try
                    {
                        // TODO: use class information about the range when available
                        if (classifier.classify(currentOntPropertyRange, classProperty.getRangeClassDesc(), currentValueOfInstanceProperty, 
                                currentValueOfInstancePropertyDesc))
                            countOfMatchedInstanceValues++; // It is a match
                    }
                    catch (ClassifierException e)
                    {
                        // TODO: print the exception
                        return false;
                    }
                }
            }

        // - Decide on the final result after testing all property values

        if (  !classProperty.isComplement()) // Range in the class definition of the property is NOT complement
        {
            if (countOfMatchedInstanceValues == instanceProperty.getValues().size()) // All values satisfy range
                return true;
            else if (countOfMatchedInstanceValues == 0) // None of the values range
                return false;
        }
        else // Range in the class definition of the property is complement
        {
            if (countOfMatchedInstanceValues == 0) // All values are outside of range so there are in the complement of them
                return true;
            else if (countOfMatchedInstanceValues == instanceProperty.getValues().size()) // All value are inside the range
                return false;
        }

        // Only some (no zero) values satisfy the range
        if (classProperty.getRangeRestrictionType().equals(KAoSOntClassBuilder._someValuesFromRestriction))
            return true; // Values satisfy the restriction, some are in the range of the class property
        else
            return false; // This is the allValues restriction so it is not satisfy
    }
}