/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2012/04/14 05:50:40 $
 */

package kaos.core.csi.query.data;

import kaos.core.csi.OntInstanceDescription;
import kaos.core.csi.extension.state.StateManager;
import kaos.core.csi.extension.state.StateManagerImpl;
import kaos.xmlschema.util.XMLSchemaBuilder;
import kaos.xmlschema.util.XMLSchemaBuilderImpl;

import java.io.Serializable;
import java.util.*;

import org.relaxng.datatype.DatatypeException;

/*  
 * This class represents an information about the ontological property associated with a certain class. 
 */
public class OntClassPropertyDescription implements Serializable
{
    private static final long   serialVersionUID    = 3704104206562318307L;

    public static final String MAIN_PROPERTY       = "MainProperty";
    public static final String SECONDARY_PROPERTY  = "SecondaryProperty";

    private String              _propertyName       = null;
    private String              _propertyRangeName  = null;
    private OntClassDescription _rangeClass         = null;
    private String              _parentRangeName    = null;
    private boolean             _isComplement       = false;
    private String              _restrictionType    = null;

    // My cached instances
    private Set<String>         _instances;

    // My cached instance description
    private Map<String, Object> _instanceDescriptions;
    
    // My cached subproperties
    private Set<String>         _subProperties;                            // to store the names of subproperties of this property

    // If the property is datatype property then information about its XML schema
    private boolean             _isDataContext     = false;               // to indicate if it is a data property
    private List<String>        _xmlSchema;

    private boolean  _isFunctional = false;
    protected String _relationToClass = MAIN_PROPERTY;

    public OntClassPropertyDescription(String propertyName)
    {
        _propertyName = propertyName;
        _instances = new HashSet<String>();
        _instanceDescriptions = new Hashtable<String, Object>();
        _subProperties = new HashSet<String>();
    }

    public OntClassPropertyDescription(OntClassPropertyDescription ontPropertyInfo)
    {
        this(ontPropertyInfo._propertyName);
        
        _propertyRangeName = ontPropertyInfo._propertyRangeName;
        _rangeClass = ontPropertyInfo._rangeClass;
        _parentRangeName = ontPropertyInfo._parentRangeName;
        _isComplement = ontPropertyInfo._isComplement;
        _restrictionType = ontPropertyInfo._restrictionType;
        _instances = ontPropertyInfo._instances;
        _instanceDescriptions = ontPropertyInfo._instanceDescriptions;
        _subProperties = ontPropertyInfo._subProperties;
        _isDataContext = ontPropertyInfo._isDataContext;
        _xmlSchema = ontPropertyInfo._xmlSchema;
        _isFunctional = ontPropertyInfo._isFunctional;
        _relationToClass = ontPropertyInfo._relationToClass;
    }

    public String getPropertyName()
    {
        return _propertyName;
    }

    public void setRangeName(String name)
    {
        _propertyRangeName = name;
    }

    public String getRangeName()
    {
        return _propertyRangeName;
    }
    
    public OntClassDescription getRangeClassDesc()
    {
        return _rangeClass;
    }

    public void setRangeClassDesc(OntClassDescription rangeClass)
    {
        _rangeClass = rangeClass;
    }

    public void setParentRangeName(String name)
    {
        _parentRangeName = name;
    }

    public String getParentRangeName()
    {
        return _parentRangeName;
    }

    public void setComplement(boolean b)
    {
        _isComplement = b;
    }

    public boolean isComplement()
    {
        return _isComplement;
    }

    public void setDatatypeProperty(boolean b)
    {
        _isDataContext = b;
    }

    public boolean isDatatypeProperty()
    {
        return _isDataContext;
    }
    
    /**
     * Sets functional property information
     */
    public void setFunctionalProperty()
    {
        _isFunctional = true;
    }

    /**
     * Check if functional property information
     */
    public boolean isFunctionalProperty()
    {
        return _isFunctional;
    }

    /**
     * Sets Relation to Action property information
     */
    public void setRelationToActionProperty(String myValue)
    {
        if ((myValue.equals(MAIN_PROPERTY)) || (myValue.equals(SECONDARY_PROPERTY)))
            _relationToClass = myValue;
    }

    /**
     * Gets  Relation to Action property information
     */
    public String getRelationToActionProperty()
    {
        return _relationToClass;
    }

    public boolean isInstancePresent(Object inst)
    {
        return _instances.contains(inst);
    }

    public void addInstance(String inst)
    {
        if (inst instanceof String)
            if (inst.equals(""))
                return;

        _instances.add(inst);
    }

    /*
     * this function adds more than one instances to this property
     */
    public void addMultipleInstances(Collection<String> inst)
    {
        _instances.addAll(inst);
    }

    // this function replaces the instances of this property with those given.
    public void setMultipleInstances(Collection<String> inst)
    {
        _instances = new HashSet<String>(inst);
    }
    
    /*
     * remove all the old instances and add the new ones
     */
    public void updateInstances(Collection<String> inst)
    {
        _instances = new HashSet<String>();
        _instances.addAll(inst);
    }

    public void addInstanceDescription(String instID, Object desc)
    {
        _instanceDescriptions.put(instID, desc);
    }
    
    public Set<String> getAllInstances()
    {
        Set<String> myResult = _instances;
        
        // Find if the instance depends on state
        if (_rangeClass != null)
        {  
            // TODO: do this recursively
            OntClassPropertyDescription hasItsState = null;// _rangeClass.getProperty(SituationConcepts.hasItsState());
            
            if (hasItsState != null)
            {
                OntClassDescription hasItsStateRange = hasItsState.getRangeClassDesc();
                
                if (hasItsStateRange != null)
                {
                    StateManager stateManager = StateManagerImpl.getInstance();
                    
                    return stateManager.getSubjectWithState(hasItsStateRange);
                }      
            } 
        }
        
        return myResult;
    }

    public void addSubpropertyNames(Collection<String> propNames)
    {
        _subProperties.addAll(propNames);
    }

    public void updateSubpropertyNames(String mainProp, Collection<String> propNames)
    {
        if ((_propertyName.equals(mainProp)) || (_subProperties.contains(mainProp)))
        {
            _subProperties.clear();
            _subProperties.addAll(propNames);
        }
    }

    public Set<String> getAllSubpropertyNames()
    {
        return _subProperties;
    }

    public boolean isSubpropertyPresent(String propName)
    {
        return _subProperties.contains(propName);
    }

    /*
     * @param instID - id of the property
     * 
     * @return OntInstanceDescription description IF it is a OntInstanceDescription, otherwise null.
     */
    public OntInstanceDescription getInstanceDescription(String instID)
    {
        if (instID == null)
        {
            // Return the first instance description which is OntInstanceDescription
            for (Object desc : _instanceDescriptions.values())
                if (desc instanceof OntInstanceDescription)
                    return (OntInstanceDescription)desc;

            return null;
        }
        
        Object desc = _instanceDescriptions.get(instID);

        if (desc instanceof OntInstanceDescription)
            return (OntInstanceDescription)desc;

        return null;
    }

    public void removeInstance(Object inst)
    {
        _instances.remove(inst);
    }

    public void removeInstanceDescription(String instID)
    {
        _instanceDescriptions.remove(instID);
    }

    /*
     * This function removes all those instances from this property that are indicated in the input parameter. 
     * All those instances not indicated by the input parameter remain in this property.
     */
    public void removeMultipleInstances(Collection<String> inst)
    {
        _instances.removeAll(inst);
    }

    public void removeAllInstances()
    {
        _instances = new HashSet<String>();
    }

    public void removeAllInstanceDescriptions()
    {
        _instanceDescriptions = new Hashtable<String, Object>();
    }

    /*
     * To set and get the restriction type of this property, i.e. is it a 'toClass' restriction or a 'hasClass' restriction etc.
     */
    public void setRangeRestrictionType(String type)
    {
        _restrictionType = type;
    }

    public String getRangeRestrictionType()
    {
        return _restrictionType;
    }

    public void setSchema(List<String> objectDef)
    {
        if ((objectDef == null) || (objectDef.isEmpty()))
            return;
          
        _xmlSchema = objectDef;
    }

    public boolean isSchemaSet()
    {
        if (_xmlSchema != null)
            return true;

        return false;
    }

    public List<String> getSchema()
    {
        return _xmlSchema;
    }
    
    public XMLSchemaBuilder getSchemaParser()
    {
        try
        {
            return new XMLSchemaBuilderImpl(_xmlSchema);
        }
        catch (DatatypeException e)
        {
            e.printStackTrace();
        }
        
        return null;
    }

    public String toString()
    {
        String s = "\n  Name: " + _propertyName + "\n";
        s += "  Class: " + _propertyRangeName + "\n";
        s += "  Parent: " + _parentRangeName + "\n";
        s += "  Restriction Type: " + _restrictionType + "\n";
        s += "  isComplement " + _isComplement + "\n";
        s += "  instances: " + _instances + "\n";
        s += "  subproperty names: " + _subProperties + "\n";

       if(isSchemaSet())
       {
          s += "  schema: " + getSchema() + "\n";
       }

        return s;
    }
}