/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2012/03/28 20:37:56 $
 */

package kaos.core.csi.registration;

import kaos.core.csi.AgentCapability;
import kaos.core.service.directory.KAoSAgentDescription;
import kaos.core.service.directory.ModificationException;
import kaos.policy.guard.GuardInstantiationException;

import javax.agent.JasBean;
import javax.agent.service.directory.AgentDescription;
import javax.agent.service.directory.AlreadyRegisteredException;
import javax.agent.service.directory.DirectoryFailure;
import javax.agent.service.directory.NotRegisteredException;
import java.util.List;

/**
 * Defines the operations to register and deregister entities and entity hosting environments in the KAoS Directory Service and operations to modify the registration information.
 * 
 * @author KAoS Team
 * $Revision: 1.20 $
 */
public interface DirectoryRegistration
{
    /**
     * Register the entity described by the given AgentDescription in the KAoS Directory Service.
     * 
     * @param entityDescription AgentDescription is the JAS AgentDescription, whose subclasses describe the various types of entities to be registered in KAoS: agents, guards,
     *          domains, containers. Those entities are subjects to policies or potential subjects to policies, or the entities are using KAoS for messaging. 
     * @param guarded boolean indicating the guarded/unguarded status for agents and containers. If the entity is to be guarded, the implementation of this method will instantiate a
     *          guard, or check if a guard already exists and proceed with the registration via the guard.
     * @param replace boolean indicating whether to replace existing AgentDescription.
     * 
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws AlreadyRegisteredException if the registering entity has been already registered in the DS (if the replace = false and there is already registration for the specified
     *          AgentDescription).
     */
    public void registerEntity(AgentDescription entityDescription, boolean guarded, boolean replace) 
        throws DirectoryFailure, AlreadyRegisteredException, GuardInstantiationException;

    /**
     * Deregister the entity described by the given AgentDescription from the KAoS Directory Service.
     * 
     * @param entityDescription AgentDescription of the entity to be deregistered.
     * 
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws NotRegisteredException if the entity to be deregistered has not been registered in the DS.
     */
    public void deregisterEntity(AgentDescription entityDescription) throws DirectoryFailure, NotRegisteredException;

    /**
     * Register the entity described as simpler method registerEntity but also register properties of this agent
     * 
     * @param properties List of Property objects further describing the registering entity.
     */
    public void registerEntityWithProperties(AgentDescription entityDescription, List<Property> properties, boolean guarded, boolean replace) 
        throws DirectoryFailure, AlreadyRegisteredException, GuardInstantiationException;

    /**
     * Modify properties of the entity identified by the given entityGUID.
     * 
     * @param entityGUID String containing the entity GUID.
     * 
     * @param properties List of Properties to be modified. Each Property contains the name, value, value description, modification type (add/remove/set, etc), modification result.
     *          Refer to the Property and ModificationResult interfaces for detailed description.
     * 
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws NotRegisteredException if the entity whose properties are to be modified has not been registered in the DS. 
     * @throws ModificationException if one or more property update failed.
     */
    public void modifyEntityProperties(String entityGUID, List<Property> properties) 
      throws DirectoryFailure, NotRegisteredException, ModificationException;

    /**
     * Create an instance of the KAoSAgentDescription using the given parameters.
     * 
     * @param guid String containing the globally unique identifier for the agent, or null. If null is passed, then the Agent Naming Service will create the identifier.
     * @param nickname String containing agent's nickname.
     * @param domainNames Vector containing the names of the domains the agent will become a member of.
     * @param transportName String specifying the name of the transport the agent wants to use, or null. If null is specified, then the Locators will be empty.
     */
    public KAoSAgentDescription createKAoSAgentDescription(String guid, String nickname, List<String> domainNames, String transportName, 
            JasBean agentInitInfo);

    /**
     * Create an instance of the KAoSAgentDescription as simpler method createKAoSAgentDescription but additionally specifies agent capabilities
     * 
     * @param agentCapability List of AgentCapabilities describing agent
     */
    public KAoSAgentDescription createKAoSAgentDescription(String guid, String nickname, List<String> domainNames, List<AgentCapability> agentCapabilities, String transportName,
            JasBean agentInitInfo);

    /**
     * Register domain with the given name in the KAoS Directory Service.
     * 
     * @param domainName String containing the name of domain to register.
     * 
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws AlreadyRegisteredException if the registering domain has been already registered in the DS
     */
    public void registerDomain(String domainName) throws DirectoryFailure, AlreadyRegisteredException;

    /**
     * Register domain with the given name in the KAoS Directory Service.
     * 
     * @param domainName String containing the name of domain to register.
     * @param modality String containing the default policy authorization result if no policies apply [PolicyConcepts.PosAuthorizationPolicy()=laissez-faire or
     *          PolicyConcepts.NegAuthorizationPolicy()=tyrannical].
     * 
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws AlreadyRegisteredException if the registering domain has been already registered in the DS
     */
    public void registerDomain(String domainName, String modality) throws DirectoryFailure, AlreadyRegisteredException;
}
