package kaos.core.csi.registration;

import com.hp.hpl.jena.ontology.Individual;
import com.hp.hpl.jena.rdf.model.Resource;

import kaos.core.csi.ontology.OntologyService;
import kaos.core.csi.ontology.OntologyService.ReasoningException;
import kaos.core.util.NamespaceValidator;
import kaos.core.csi.query.QueryState.UnknownConceptException;
import kaos.ontology.util.SerializableOntModelImpl;
import kaos.ontology.vocabulary.OWLConcepts;
import kaos.ontology.vocabulary.RDFConcepts;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;

import javax.agent.service.directory.DirectoryFailure;

/**
 * @author lbunch
 *
 * Utility class for manipulating kaos.core.csi.Property objects
 */
public class PropertyUtil
{
    /**
     * Create list of OWL individuals identified by entityGUID with the given property values assigned
     * 
     * @param entityGUID globally unique identifier String for an ontology individual
     * @param properties List of properties (name-value pairs)
     * 
     * @return List of OWL strings containing the individual with the given id and properties
     * @throws UnknownConceptException 
     */
    public static List<String> getPropertiesAsOWL(String entityGUID, List<Property> properties, OntologyService ontRepository) 
        throws OntologyService.ReasoningException, UnknownConceptException
    {
        if (entityGUID == null || entityGUID.length() == 0 || properties == null || properties.size() == 0)
            return null;

        List<String> myResult = new ArrayList<String>();

        for (Property prop : properties)
            myResult.add(getPropertyAsOWL(entityGUID, prop, ontRepository));

        return myResult;
    }

    /**
     * Create an OWL individual identified by entityGUID with the given property values assigned
     * 
     * @param entityGUID globally unique identifier String for an ontology individual
     * @param myProperty Property (name-value pair)
     * 
     * @return String containing the OWL definition of individual with the given id and property
     * @throws UnknownConceptException 
     */
    public static String getPropertyAsOWL(String entityGUID, Property myProperty, OntologyService ontRepository) 
        throws OntologyService.ReasoningException, UnknownConceptException
    {
        if (entityGUID == null || entityGUID.length() == 0 || myProperty == null || myProperty.getName() == null || myProperty.getValue() == null)
            return null;

        return getTripleAsOWL(entityGUID, myProperty.getName(), myProperty.getValue().toString(), ontRepository);
    }

    /**
     * Create an OWL individual identified by the given triple
     * 
     * @param entityGUID globally unique identifier String for an ontology individual
     * @param myProperty property name as String
     * @param myValue property value name as String
     * 
     * @return String containing the OWL definition of individual with the given id, property and its value
     * @throws UnknownConceptException 
     */
    public static String getTripleAsOWL(String entityGUID, String myProperty, String myValue, OntologyService ontRepository) 
        throws OntologyService.ReasoningException, UnknownConceptException
    {
        if (entityGUID == null || entityGUID.length() == 0 || myProperty == null || myValue == null)
            return null;

        SerializableOntModelImpl om = new SerializableOntModelImpl();
        Resource thing = om.createClass(OWLConcepts.Thing());
        Individual indiv = om.createIndividual(NamespaceValidator.validateURI(entityGUID), thing);
        com.hp.hpl.jena.rdf.model.Property p = om.createProperty(myProperty);

        try
        {
            if (ontRepository.testTrue(myProperty, RDFConcepts._type_, OWLConcepts.DatatypeProperty()))
                indiv.addProperty(p, myValue);
            else
            {
                Resource r = om.createResource(NamespaceValidator.validateURI(myValue));
                indiv.addProperty(p, r);
            }
        }
        catch (DirectoryFailure e)
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        StringWriter sw = new StringWriter();
        om.write(sw, "RDF/XML-ABBREV");

        return sw.toString();
    }
}