/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2012/03/28 20:37:56 $
 */
package kaos.core.csi.registration;

import kaos.core.csi.AgentCapability;
import kaos.core.csi.CSIFactory;
import kaos.core.csi.policy.enforcement.EnforcerInstantiationException;
import kaos.core.csi.transport.TransportSupport;
import kaos.core.service.KAoSServiceRoot;
import kaos.core.service.boot.KAoSGuardBootstrapper;
import kaos.core.service.boot.SystemBooter;
import kaos.core.service.directory.*;
import kaos.core.service.naming.KAoSAgentNamingService;
import kaos.core.util.KAoSConfigurationParameters;
import kaos.core.util.Logger;
import kaos.core.util.NamespaceValidator;
import kaos.core.csi.query.QueryState.UnknownConceptException;
import kaos.ontology.vocabulary.DefaultNamespaces;
import kaos.ontology.vocabulary.GroupConcepts;
import kaos.ontology.vocabulary.PlaceConcepts;
import kaos.ontology.vocabulary.PolicyConcepts;
import kaos.policy.guard.GuardInstantiationException;
import kaos.policy.guard.GuardRetriever;
import kaos.policy.guard.KAoSGuard;
import kaos.spatial.Block3D;
import kaos.spatial.Location3D;
import kaos.spatial.Orientation3D;
import kaos.spatial.SpatialReasoner;
import kaos.spatial.vocabulary.PhysicalAddressConcepts;
import ri.acr.AcrAgentName;

import javax.agent.AgentName;
import javax.agent.JasBean;
import javax.agent.Locator;
import javax.agent.service.ServiceProperties;
import javax.agent.service.directory.AgentDescription;
import javax.agent.service.directory.AlreadyRegisteredException;
import javax.agent.service.directory.DirectoryFailure;
import javax.agent.service.directory.NotRegisteredException;
import javax.agent.service.naming.AgentNamingService;
import java.util.*;

/**
 * This class implements the Registration interface, which defines the operations to register and deregister entities and entity hosting environments
 * in the KAoS Directory Service and operations to modify the registration information. It also defines operations to
 * register entities in KAoS Domains and to maintain domain hierarchy.
 * $Revision: 1.50 $
 */
public class RegistrationImpl implements Registration
{
    private static final RegistrationImpl  __INSTANCE    = new RegistrationImpl();
    private static boolean                 __INITIALIZED = false;

    private DirectoryService               _dirProxy;
    private String                         _transportName;
    private Logger                         _logger       = Logger.getLogger(this);

    /*
     * Retrieve an instance of Registration.
     * @param transportName String specifying the name of the transport to use for communication with the KAoS Directory Service.
     * @return a singleton, Registration.
     */
    public static Registration getInstance(String transportName) throws Exception
    {
        if (__INITIALIZED)
            return __INSTANCE;

        __INSTANCE.initialize(transportName);
        __INITIALIZED = true;

        return __INSTANCE;
    }

    /*
     * Retrieve an instance of Registration.
     * @return a singleton, Registration.
     */
    public static Registration getInstance() throws Exception
    {
        if (__INITIALIZED)
            return __INSTANCE;

        String transportName = System.getProperty(KAoSConfigurationParameters.DIRECTORY_TRANSPORT);

        if (transportName == null)
            throw new Exception("System property: directory.transport is null, unable to connect to the Directory Service!");

        return getInstance(transportName);
    }

    private RegistrationImpl()
    {
    }

    /*
     * kaos.core.csi.registration.DirectoryRegistration interface implementation - begin
     */

    /*
     * Register domain with the given name in the KAoS Directory Service.
     * @param domainName String containing the name of domain to register.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws AlreadyRegisteredException if the registering domain has been already registered in the DS
     */
    public void registerDomain(String domainName, String modality) throws DirectoryFailure, AlreadyRegisteredException
    {
        DefaultDomainDescription desc = new DefaultDomainDescription();
        desc.setDomainName(domainName);
        desc.setModality(modality);
        try
        {
            registerEntity(desc, false, false);
        }
        catch (DirectoryFailure df)
        {
            throw df;
        }
        catch (AlreadyRegisteredException are)
        {
            throw are;
        }
        catch (GuardInstantiationException gie)
        {
            // ignore since this should not happen
        }
    }

    /*
     * Register domain with the given name in the KAoS Directory Service.
     * @param domainName String containing the name of domain to register.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws AlreadyRegisteredException if the registering domain has been already registered in the DS
     */
    public void registerDomain(String domainName) throws DirectoryFailure, AlreadyRegisteredException
    {
        registerDomain(domainName, PolicyConcepts.PosAuthorizationPolicy());
    }

    /*
     * Register the entity described by the given AgentDescription in the KAoS Directory Service.
     * @param entityDescription AgentDescription is the JAS AgentDescription, whose subclasses describe the various types of entities to be registered
     * in KAoS: agents, guards,domains, containers. Those entities are subjects to policies or potential subjects to policies, or the entities are using
     * KAoS for messaging.
     * @param guarded boolean indicating the guarded/unguarded status for agents and containers. If the entity is to be guarded, the implementation of
     * this method will instantiate a guard, or check if a guard already exists and proceed with the registration via the guard.
     * @param replace boolean indicating whether to replace existing AgentDescription.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws AlreadyRegisteredException if the registering entity has been already registered in the DS (if the replace = false and there is
     * already registration for the specified AgentDescription).
     */
    public void registerEntity(AgentDescription entityDescription, boolean guarded, boolean replace) throws DirectoryFailure, AlreadyRegisteredException,
            GuardInstantiationException
    {
        entityDescription.set(KAoSDirectoryService.REGISTRATION_REPLACE_FLAG, new Boolean(replace));

        // Check, if the agent is guarded.
        if (guarded)
        {
            KAoSGuard guard = GuardRetriever.getGuard();
            KAoSAgentDescription desc = null;

            if (entityDescription instanceof KAoSAgentDescription)
                desc = (KAoSAgentDescription)entityDescription;
            else
                throw new IllegalArgumentException("AgentDescription not instance of KAoSAgentDescription - Guard unable to register entity!");

            if (guard == null)
            {
            	_logger.logMessage("Registering guarded entity, but Guard is not present.", Logger.WARN);
                return;
            }
            
            try
            {
                guard.newAgent(desc, null);
            }
            catch (AlreadyRegisteredException are)
            {
                throw are;
            }
            catch (DirectoryFailure df)
            {
                throw df;
            }
            catch (EnforcerInstantiationException eie)
            {
                // TO DO: check how to handle this exception. Is it ever thrown?
                eie.printStackTrace();
            }
        }
        else
            _dirProxy.register(entityDescription);
    }

    /*
     * Register the entity described by the given AgentDescription and the spatial properties in the KAoS Directory ServiceAND the SpatialReasoner.
     * @param entityDescription AgentDescription is the JAS AgentDescription, whose subclasses describe the various types of entities to be registered in
     * KAoS: agents, guards,
     * domains, containers. Those entities are subjects to policies or potential subjects to policies, or the entities are using KAoS for messaging.
     * @param objectLocation Location3D containing the X,Y,Z coordinates of the registering entity.
     * @param objectOrientation Orientation3D containing the X,Y,Z radius of the registering entity.
     * @param objectOccupiedSpace Block3D containing corner locations determining the occupied space of the registering entity.
     * @param guarded boolean indicating the guarded/unguarded status for agents and containers. If the entity is to be guarded, the implementation of this
     * method will instantiate a
     * guard, or check if a guard already exists and proceed with the registration via the guard.
     * @param replace boolean indicating whether to replace existing AgentDescription.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws AlreadyRegisteredException if the registering entity has been already registered in the DS (if the replace = false and there is already
     * registration for the specified
     * AgentDescription).
     */
    public void registerEntityWithSpatialProperties(AgentDescription entityDescription, Location3D objectLocation, Orientation3D objectOrientation,
            Block3D objectOccupiedSpace, boolean guarded, boolean replace) throws DirectoryFailure, AlreadyRegisteredException,
            GuardInstantiationException
    {
        Property locationProp = null;
        Property orientationProp = null;
        Property blockProp = null;
        ArrayList aList = null;

        if (objectLocation != null || objectOrientation != null || objectOccupiedSpace != null)
            aList = new ArrayList();

        if (objectLocation != null)
        {
            locationProp = new PropertyImpl();
            locationProp.setName(PhysicalAddressConcepts.hasLocation());
            locationProp.setValue(objectLocation.getMyId());
            locationProp.setPropertyObject(objectLocation);
            aList.add(locationProp);
        }

        if (objectOrientation != null)
        {
            orientationProp = new PropertyImpl();
            orientationProp.setName(PhysicalAddressConcepts.hasOrientation());
            orientationProp.setValue(objectOrientation.getMyId());
            orientationProp.setPropertyObject(objectOrientation);
            aList.add(orientationProp);
        }

        if (objectOccupiedSpace != null)
        {
            blockProp = new PropertyImpl();
            blockProp.setName(PlaceConcepts.isEncompassedBy());
            blockProp.setValue(objectOccupiedSpace.getMyId());
            blockProp.setPropertyObject(objectOccupiedSpace);
            aList.add(blockProp);
        }

        try
        {
            if (aList != null)
                registerEntityWithProperties(entityDescription, aList, guarded, replace);
            else
            {
                registerEntity(entityDescription, guarded, replace);
                return;
            }
        }
        catch (AlreadyRegisteredException ex)
        {
            if (aList == null)
                return;
        }

        DefaultKAoSEntityDescription kDesc = (DefaultKAoSEntityDescription)entityDescription;
        Set names = new HashSet();

        for (String currentType : kDesc.getEntityOntologicalTypeNames())
            names.add(currentType);

        registerObjectWithSpatialProperties(kDesc.getEntityNameAsString(), names, objectLocation, objectOrientation, objectOccupiedSpace);
    }

    /*
     * Register object identified by the given objectID and its spatial properties in the SpatialReasoner.
     * Important: Note that the object is not registered in the KAoS DirectoryService, only in the SpatialReasoner.
     * If you want to register a given object in the KAoS Directory service, use method registerEntityWithSpatialProperties(), which will do both,
     * register in the KAoS DS and the SpatialReasoner.
     * @param objectID String uniquely identifying registering object in the SpatialReasoner.
     * @param objectOntologyTypes Set containing Strings describing the ontological types of the registering object in the SpatialReasoner.
     * @param objectLocation Location3D containing the X,Y,Z coordinates of the registering object.
     * @param objectOrientation Orientation3D containing the X,Y,Z radius of the registering object.
     * @param objectOccupiedSpace Block3D containing corner locations determining the occupied space of the registering object.
     * @throws DirectoryFailure if the connection to the SpatialReasoner cannot be established.
     * @throws AlreadyRegisteredException if the registering object has been already registered in the SpatialReasoner.
     */
    public void registerObjectWithSpatialProperties(String objectID, Set objectOntologyTypes, Location3D objectLocation, Orientation3D objectOrientation,
            Block3D objectOccupiedSpace) throws DirectoryFailure, AlreadyRegisteredException
    {
        SpatialReasoner mySpatialReasoner = null;

        try
        {
            CSIFactory.setCurrentTransport(_transportName);
            mySpatialReasoner = CSIFactory.getSpatialReasoner();
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new DirectoryFailure("\n  *** Unable to obtain SpatialReasoner for transport: " + _transportName + "\n  *** Unable to register object: "
                    + objectID);
        }

        mySpatialReasoner.registerSpatialObject(objectID, objectOntologyTypes, objectLocation, objectOrientation, objectOccupiedSpace);
    }

    /*
     * Register object identified by the given objectID and its spatial properties in the SpatialReasoner.
     * Important: Note that the object is not registered in the KAoS DirectoryService, only in the SpatialReasoner.
     * If you want to register a given object in the KAoS Directory service, use method registerEntityWithSpatialProperties(), which will do both,
     * register in the KAoS DS and the SpatialReasoner.
     * @param objectID String uniquely identifying registering object in the SpatialReasoner.
     * @param objectOntologyType Set containing Strings describing the ontological types of the registering object in the SpatialReasoner.
     * @param objectOrientation Orientation3D containing the X,Y,Z radius of the registering object.
     * @param boundingPoints An ordered List of points used to define the objects boundary. Location should then be defined as the average of the points.
     * Occupied Space should be the max and min of the points along each axis.
     * @throws DirectoryFailure if the connection to the SpatialReasoner cannot be established.
     * @throws AlreadyRegisteredException if the registering object has been already registered in the SpatialReasoner.
     */
    public void registerObjectWithSpatialProperties(String objectID, Set objectOntologyTypes, Orientation3D objectOrientation,
            Vector<Location3D> boundingPoints) throws DirectoryFailure, AlreadyRegisteredException
    {
        SpatialReasoner mySpatialReasoner = null;

        try
        {
            CSIFactory.setCurrentTransport(_transportName);
            mySpatialReasoner = CSIFactory.getSpatialReasoner();
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new DirectoryFailure("\n  *** Unable to obtain SpatialReasoner for transport: " + _transportName + "\n  *** Unable to register object: "
                    + objectID);
        }

        mySpatialReasoner.registerSpatialObject(objectID, objectOntologyTypes, objectOrientation, boundingPoints);
    }

    /*
     * Deregister object identified by the given objectID and its spatial properties in the SpatialReasoner.
     * Important: Note that the object is not deregistered from the KAoS DirectoryService, only from the SpatialReasoner.
     * @param objectID String uniquely identifying registering object in the SpatialReasoner.
     * @throws DirectoryFailure if the connection to the SpatialReasoner cannot be established.
     * @throws NotRegisteredException if the registering object has not been registered in the SpatialReasoner.
     */
    public void deregisterObjectWithSpatialProperties(String objectID) throws DirectoryFailure, NotRegisteredException
    {
        SpatialReasoner mySpatialReasoner = null;

        try
        {
            CSIFactory.setCurrentTransport(_transportName);
            mySpatialReasoner = CSIFactory.getSpatialReasoner();
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new DirectoryFailure("\n  *** Unable to obtain SpatialReasoner for transport: " + _transportName + "\n  *** Unable to register object: "
                    + objectID);
        }

        mySpatialReasoner.deregisterSpatialObject(objectID);
    }

    /*
     * Register the entity described by the given AgentDescription and the List of Properties in the KAoS Directory Service.
     * @param entityDescription AgentDescription is the JAS AgentDescription, whose subclasses describe the various types of entities
     * to be registered in KAoS: agents, guards, domains, containers. Those entities are subjects to policies or potential subjects
     * to policies, or the entities are using KAoS for messaging.
     * @param properties List of Property objects further describing the registering entity.
     * @param guarded boolean indicating the guarded/unguarded status for agents and containers. If the entity is to be guarded,
     * the implementation of this method will instantiate a guard, or check if a guard already exists and proceed with the registration via the guard.
     * @param replace boolean indicating whether to replace existing AgentDescription.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws AlreadyRegisteredException if the registering entity has been already registered in the DS
     * (if the replace = false and there is already registration for the specified
     * AgentDescription).
     */
    public void registerEntityWithProperties(AgentDescription entityDescription, List properties, boolean guarded, boolean replace)
            throws DirectoryFailure, AlreadyRegisteredException, GuardInstantiationException
    {
        for (Iterator propertiesIt = properties.iterator(); propertiesIt.hasNext();)
        {
            Object propObj = propertiesIt.next();

            if (propObj instanceof Property)
            {
                Property prop = (Property)propObj;
                entityDescription.set(prop.getName(), prop);
            }
        }

        registerEntity(entityDescription, guarded, replace);
    }

    /*
     * Register the entity hosting environment described by the given EntityHostEnvDescription in the KAoS Directory Service.
     * @param eheDescription EntityHostEnvDescription describes the hosting environment for agents and guards.
     * @param guarded boolean indicating the guarded/unguarded status for the environment. If the hosted environemnt is to be guarded, the implementation
     * of this method will
     * instantiate a guard, or check if a guard already exists and proceed with the registration via the guard.
     * @param replace boolean indicating whether to replace existing EntityHostEnvDescription.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws AlreadyRegisteredException if the registering hosting environment has been already registered in the DS (if the replace = false and there is
     * already registration for
     * the specified EntityHostEnvDescription).
     */
    public void registerEntityHostingEnv(EntityHostEnvDescription eheDescription, boolean guarded, boolean replace) throws DirectoryFailure,
            AlreadyRegisteredException, GuardInstantiationException
    {
        System.out.println("    \n\n    TO BE IMPLEMENTED!");
    }

    /*
     * Modify properties of the entity identified by the given entityGUID.
     * @param entityGUID String containing the entity GUID.
     * @param properties List of Properties to be modified. Each Property contains the name, value, value description,
     * modification type (add/remove/set, etc), modification result. Refer to the Property and ModificationResult interfaces for detailed description.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws NotRegisteredException if the entity whose properties are to be modified has not been registered in the DS.
     * @throws ModificationException if one or more property update failed.
     */
    public void modifyEntityProperties(String entityGUID, List properties) throws DirectoryFailure, NotRegisteredException, ModificationException
    {
        if (properties == null)
            throw new ModificationException("Properties to modify are null! Requested modification failed");

        if (properties.isEmpty())
            return;

        _dirProxy.modifyEntityProperties(entityGUID, properties);
    }

    /*
     * Modify given spatial properties of the entity identified by the given entityGUID in the KAoS Directory ServiceAND the SpatialReasoner.
     * @param entityGUID String containing the entity GUID.
     * @param objectLocation modified Location3D containing the X,Y,Z coordinates of the entity.
     * @param objectOrientation modified Orientation3D containing the X,Y,Z radius of the entity.
     * @param objectOccupiedSpace modified Block3D containing corner locations determining the occupied space of the entity.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws NotRegisteredException if the entity whose properties are to be modified has not been registered in the DS.
     * @throws ModificationException if one or more property update failed.
     */
    public void modifyEntitySpatialProperties(String entityGUID, Location3D objectLocation, Orientation3D objectOrientation, Block3D objectOccupiedSpace)
            throws DirectoryFailure, NotRegisteredException, ModificationException
    {
        SpatialReasoner mySpatialReasoner = null;

        try
        {
            CSIFactory.setCurrentTransport(_transportName);
            mySpatialReasoner = CSIFactory.getSpatialReasoner();
        }
        catch (Exception e)
        {
            e.printStackTrace();

            throw new DirectoryFailure("Unable to obtain SpatialReasoner for transport: " + _transportName + " Modification of properties for: "
                    + entityGUID + " fails!");
        }

        Property locationProp = null;
        Property orientationProp = null;
        Property blockProp = null;
        ArrayList aList = null;

        if (objectLocation != null || objectOrientation != null || objectOccupiedSpace != null)
            aList = new ArrayList();

        if (objectLocation != null)
        {
            if (objectLocation.getModificationType() == null)
            {
                if (_logger.isEnabledFor(Logger.WARN))
                    logMessage(".modifyEntitySpatialProperties: NO MODIFICATION TYPE SPECIFIED FOR LOCATION: " + objectLocation + "\n  FOR ENTITY: "
                            + entityGUID + " NO UPDATE WILL BE MADE!", Logger.WARN);
            }
            else
            {
                locationProp = new PropertyImpl();
                locationProp.setName(PhysicalAddressConcepts.hasLocation());
                locationProp.setValue(objectLocation.getMyId());
                locationProp.setPropertyObject(objectLocation);
                ModificationTypeImpl result = new ModificationTypeImpl();
                result.setModificationType(objectLocation.getModificationType());
                locationProp.setModificationType(result);
                aList.add(locationProp);
            }
        }

        if (objectOrientation != null)
        {
            if (objectOrientation.getModificationType() == null)
            {
                if (_logger.isEnabledFor(Logger.WARN))
                    logMessage(".modifyEntitySpatialProperties: NO MODIFICATION TYPE SPECIFIED FOR ORIENTATION: " + objectOrientation + "\n  FOR ENTITY: "
                            + entityGUID + " NO UPDATE WILL BE MADE!", Logger.WARN);
            }
            else
            {
                orientationProp = new PropertyImpl();
                orientationProp.setName(PhysicalAddressConcepts.hasOrientation());
                orientationProp.setValue(objectOrientation.getMyId());
                orientationProp.setPropertyObject(objectOrientation);
                ModificationTypeImpl result = new ModificationTypeImpl();
                result.setModificationType(objectOrientation.getModificationType());
                orientationProp.setModificationType(result);
                aList.add(orientationProp);
            }
        }

        if (objectOccupiedSpace != null)
        {
            if (objectOccupiedSpace.getModificationType() == null)
            {
                if (_logger.isEnabledFor(Logger.WARN))
                    logMessage(".modifyEntitySpatialProperties: NO MODIFICATION TYPE SPECIFIED FOR BLOCK: " + objectOccupiedSpace + "\n  FOR ENTITY: "
                            + entityGUID + " NO UPDATE WILL BE MADE!", Logger.WARN);
            }
            else
            {
                blockProp = new PropertyImpl();
                blockProp.setName(PlaceConcepts.isEncompassedBy());
                blockProp.setValue(objectOccupiedSpace.getMyId());
                blockProp.setPropertyObject(objectOccupiedSpace);
                ModificationTypeImpl result = new ModificationTypeImpl();
                result.setModificationType(objectOccupiedSpace.getModificationType());
                blockProp.setModificationType(result);
                aList.add(blockProp);
            }
        }

        if (aList != null && (aList.size() > 0))
        {
            modifyEntityProperties(entityGUID, aList);

            modifyObjectSpatialProperties(entityGUID, objectLocation, objectOrientation, objectOccupiedSpace);
        }
    }

    /*
     * Modify spatial properties of the object identified by the given objectID in the SpatialReasoner.
     * Important: Note that the properties arenot modified in the KAoS DirectoryService, only in the SpatialReasoner.
     * If you want to modify properties of a given object in the KAoS Directory service, use method modifyEntitySpatialProperties(),
     * which will do both, modify in the KAoS DS and the SpatialReasoner.
     * @param objectID String uniquely identifying object, whose spatial properties will be modified in the SpatialReasoner.
     * @param objectLocation modified Location3D containing the X,Y,Z coordinates of the entity.
     * @param objectOrientation modified Orientation3D containing the X,Y,Z radius of the entity.
     * @param objectOccupiedSpace modified Block3D containing corner locations determining the occupied space of the entity.
     * @throws DirectoryFailure if the connection to the SpatialReasoner cannot be established.
     * @throws NotRegisteredException if the object, whose properties are to be modified has not been registered in the SpatialReasoner.
     */
    public void modifyObjectSpatialProperties(String objectID, Location3D objectLocation, Orientation3D objectOrientation, Block3D objectOccupiedSpace)
            throws DirectoryFailure, NotRegisteredException
    {
        SpatialReasoner mySpatialReasoner = null;

        try
        {
            CSIFactory.setCurrentTransport(_transportName);
            mySpatialReasoner = CSIFactory.getSpatialReasoner();
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new DirectoryFailure("\n  Unable to obtain SpatialReasoner for transport: " + _transportName
                    + "\n   Unable to update properties for object: " + objectID);
        }

        mySpatialReasoner.updateSpatialObject(objectID, objectLocation, objectOrientation, objectOccupiedSpace);
    }

    /*
     * Deregister the entity described by the given AgentDescription from the KAoS Directory Service.
     * @param entityDescription AgentDescription of the entity to be deregistered.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws NotRegisteredException if the entity to be deregistered has not been registered in the DS.
     */
    public void deregisterEntity(AgentDescription entityDescription) throws DirectoryFailure, NotRegisteredException
    {
        _dirProxy.deregister(entityDescription);

        try
        {
            AcrAgentName myAgentName = (AcrAgentName)entityDescription.getAgentName();

            deregisterObjectWithSpatialProperties(myAgentName.getName());
        }
        catch (Exception ex)
        {
            // Ignore if any exception
        }
    }

    /*
     * Deregister the entity hosting environment described by the given EntityHostEnvDescription from the KAoS Directory Service.
     * @param eheDescription EntityHostEnvDescription of the entity hosting environment to be deregistered.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws NotRegisteredException if the entity to be deregistered has not been registered in the DS.
     */
    public void deregisterEntityHostingEnv(EntityHostEnvDescription eheDescription) throws DirectoryFailure, NotRegisteredException
    {
        System.out.println("    \n\n    TO BE IMPLEMENTED!");
    }

    public AgentConfiguration getAgentConfiguration()
    {
        return getAgentConfiguration(null);
    }

    public AgentConfiguration getAgentConfiguration(String configFile)
    {
        ServiceProperties props = null;
        if (configFile == null)
        {
            props = SystemBooter.getAgentConfigProps();
        }
        else
        {
            props = SystemBooter.getAgentConfigProps(configFile);
        }

        if (props == null)
        {
            return null;
        }

        return new AgentConfiguration(props);
    }

    /*
     * Create an instance of the KAoSAgentDescription using the given parameters.
     * @param guid String containing the globally unique identifier for the agent, or null. If null is passed,
     * then the Agent Naming Service will create the identifier.
     * @param nickname String containing agent's nickname.
     * @param domainNames Vector containing the names of the domains the agent will become a member of.
     * @param transportName String specifying the name of the transport the agent wants to use, or null.
     * If null is specified, then the Locators will be empty.
     */
    public KAoSAgentDescription createKAoSAgentDescription(String guid, String nickname, List<String> domainNames, String transportName,
            JasBean agentInitInfo)
    {
        KAoSAgentDescription desc = new DefaultKAoSAgentDescription();
        AgentName name = null;

        try
        {
            name = acquireAgentName(guid);
        }
        catch (Exception e)
        {
            logMessage(".getKAoSAgentDescription: Unable to obtain AgentName for guid:  " + guid + " nickname: " + nickname
                    + "\n   Requested KAoSAgentDescription for cannot be created!", e, Logger.ERROR);

            return null;
        }

        if (name != null)
            desc.setAgentName(name);
        else
        {
            logMessage(".getKAoSAgentDescription: unable to obtain AgentName for agent guid: " + guid + " nickname: " + nickname, Logger.ERROR);

            return null;
        }

        if (nickname != null)
            desc.setAgentNickname(nickname);

        if (transportName != null)
        {
            KAoSServiceRoot sr = null;

            try
            {
                sr = KAoSGuardBootstrapper.getKAoSServiceRootForTransport(transportName);
            }
            catch (Exception e)
            {
                logMessage("\n  *** Unable to obtain KAoSServiceRoot for transport:  " + transportName + "\n   Requested KAoSAgentDescription for "
                        + nickname + " cannot be created!", e, Logger.ERROR);

                return null;
            }

            Locator loc = createLocator(nickname, desc.getEntityNameAsString(), agentInitInfo, transportName);

            if (loc == null)
                return null;

            sr.setLocatorForDescription(desc, loc);
        }

        if (domainNames != null)
            for (String domainName : domainNames)
                desc.addDomainName(domainName);

        if (_logger.isEnabledFor(Logger.INFO))
            logMessage(".getKAoSAgentDescription: returning KAoSAgentDescription: " + desc + "\n", Logger.INFO);

        return desc;
    }

    /*
     * Create an instance of the KAoSAgentDescription using the given parameters.
     * @param guid String containing the globally unique identifier for the agent, or null. If null is passed, then the Agent Naming Service will create
     * the identifier.
     * @param nickname String containing agent's nickname.
     * @param domainNames Vector containing the names of the domains the agent will become a member of.
     * @param agentCapabilities list with AgentCapability
     * @param transportName String specifying the name of the transport the agent wants to use, or null. If null is specified, then the Locators will be
     * empty.
     */
    public KAoSAgentDescription createKAoSAgentDescription(String guid, String nickname, List<String> domainNames,
            List<AgentCapability> agentCapabilities, String transportName, JasBean agentInitInfo)
    {
        KAoSAgentDescription desc = createKAoSAgentDescription(guid, nickname, domainNames, transportName, agentInitInfo);

        if (desc == null)
            return desc;

        for (AgentCapability agentCapability : agentCapabilities)
            desc.addCapability(agentCapability);

        return desc;
    }

    /*
     * kaos.core.csi.registration.DirectoryRegistration interface implementation - end
     */

    /*
     * kaos.core.csi.registration.DomainRegistration interface implementation - begin
     */
    /*
     * Register the entity described by the given KAoSEntityDescription in the specified domain.
     * @param domain DomainDescription describing the domain, in which to register the given entity.
     * @param entityDesc KAoSEntityDescription describes the entity to be registered in the given domain.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws AlreadyRegisteredException if the registering entity has been already registered in the the given domain.
     */
    public KAoSEntityDescription registerEntityInDomain(DomainDescription domain, KAoSEntityDescription entityDesc) throws DirectoryFailure
    {
        KAoSGuard guard = GuardRetriever.getGuard();

        try
        {
            guard.addDomain(domain);
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }

        try
        {
            entityDesc.addDomainName(domain.getDomainName());

            if (entityDesc instanceof KAoSAgentDescription)
                guard.modifyAgent((KAoSAgentDescription)entityDesc, null);
        }
        catch (DirectoryFailure df)
        {
            entityDesc.removeDomainName(domain.getDomainName());

            throw (df);
        }
        catch (EnforcerInstantiationException e)
        {
            e.printStackTrace();
        }
        catch (UnknownConceptException e)
        {
            e.printStackTrace();
        }

        return entityDesc;
    }

    /*
     * Deregister the entity described by the given KAoSEntityDescription from the specified domain.
     * @param domain DomainDescription describing the domain, from which to deregister the given entity.
     * @param entityDesc KAoSEntityDescription describes the entity to be deregistered from the given domain.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws NotRegisteredException if the deregistering entity has not been registered in the the given domain.
     */
    public KAoSEntityDescription deregisterEntityFromDomain(DomainDescription domain, KAoSEntityDescription entityDesc) throws DirectoryFailure,
            NotRegisteredException
    {
        try
        {
            entityDesc.removeDomainName(domain.getDomainName());

            if (entityDesc instanceof KAoSAgentDescription)
            {
                KAoSGuard guard = GuardRetriever.getGuard();
                guard.modifyAgent((KAoSAgentDescription)entityDesc, null);
            }
        }
        catch (DirectoryFailure df)
        {
            entityDesc.addDomainName(domain.getDomainName());
            throw (df);
        }
        catch (EnforcerInstantiationException e)
        {
            e.printStackTrace();
        }
        catch (UnknownConceptException e)
        {
            e.printStackTrace();
        }

        return entityDesc;
    }

    // --- kaos.core.csi.registration.DomainRegistration interface implementation - end

    // --- kaos.core.csi.registration.DomainHierarchyConfiguration interface implementation - begin

    /*
     * Set as a the parent the given parentDomain for the specified domain.
     * @param domain DomainDescription describes the domain, whose parent is to be updated.
     * @param parentDomain DomainDescription describes the parent domain.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws HasParentException if the specified domain has already a parent.
     */
    public void setParentForDomain(DomainDescription domain, DomainDescription parentDomain) throws DirectoryFailure, HasParentException
    {
        System.out.println("    \n\n    TO BE IMPLEMENTED!");
    }

    /*
     * Add the childDomain as a child of the specified domain.
     * @param domain DomainDescription describes the domain, to which the childDomain will be added.
     * @param childDomain DomainDescription describes the child domain.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws HasParentException if the specified child domain has already a parent.
     */
    public void addChildToDomain(DomainDescription domain, DomainDescription childDomain) throws DirectoryFailure, HasParentException
    {
        System.out.println("    \n\n    TO BE IMPLEMENTED!");
    }

    /*
     * Add the childDomain as a child of the specified domain.
     * @param domain DomainDescription describes the domain, from which to remove the child domain.
     * @param childDomain DomainDescription describes the child domain to be removed.
     * @throws DirectoryFailure if the connection to the DS cannot be established.
     * @throws NotRegisteredException if the specified domain does not have the specified child domain as a child.
     */
    public void removeChildFromDomain(DomainDescription domain, DomainDescription childDomain) throws DirectoryFailure, NotRegisteredException
    {
        System.out.println("    \n\n    TO BE IMPLEMENTED!");
    }

    /*
     * Adds the specified actor as a leader of the given group. Sets values in the ontology service for both the Group.hasLeader and
     * Actor.isLeaderOf properties.
     * @param actorId - unique identifier of the Actor (Agent) individual to be set as the leader
     * @param teamURI - unique URI of the Group for which to set the leader
     * @return
     * @throws NotRegisteredException
     * @throws DirectoryFailure
     * @throws ModificationException
     */
    public ModificationType addGroupLeader(String actorId, String groupURI) throws NotRegisteredException, DirectoryFailure, ModificationException
    {
        actorId = NamespaceValidator.validateURI(actorId);
        groupURI = NamespaceValidator.validateURI(groupURI);

        ModificationTypeImpl modResult = new ModificationTypeImpl();
        modResult.setModificationType(ModificationType.ADD);

        // TODO: Assert that the team has the actor as its leader
        kaos.core.csi.registration.Property teamHasLeaderProp = new kaos.core.csi.registration.PropertyImpl(GroupConcepts.hasLeader(), "team has leader",
                modResult, 0, actorId);
        Vector propList = new Vector();
        propList.add(teamHasLeaderProp);

        // TODO: Since JTP is not honoring the inverseOf attribute also assert that the actor is the leader of the team
        kaos.core.csi.registration.Property actorIsLeaderOfProp = new kaos.core.csi.registration.PropertyImpl(GroupConcepts.isLeaderOf(),
                "actor leads team", modResult, 0, groupURI);
        propList.add(actorIsLeaderOfProp);
        modifyEntityProperties(actorId, propList);

        return modResult;
    }

    /*
     * Remove the specified actor as a leader of the given group. Clears values in the ontology service for both the
     * Group.hasLeader and Actor.isLeaderOf properties.
     * @param actorId - unique identifier of the Actor (Agent) individual to be removed as the leader
     * @param teamURI - unique URI of the Group from which to remove the leader
     * @return
     * @throws NotRegisteredException
     * @throws DirectoryFailure
     * @throws ModificationException
     */
    public ModificationType removeGroupLeader(String actorId, String groupURI) throws NotRegisteredException, DirectoryFailure, ModificationException
    {
        actorId = NamespaceValidator.validateURI(actorId);
        groupURI = NamespaceValidator.validateURI(groupURI);

        ModificationTypeImpl modResult = new ModificationTypeImpl();
        modResult.setModificationType(ModificationType.DELETE);

        // TODO: Assert that the team has the actor as its leader
        kaos.core.csi.registration.Property teamHasLeaderProp = new kaos.core.csi.registration.PropertyImpl(GroupConcepts.hasLeader(), "team has leader",
                modResult, 0, actorId);
        Vector propList = new Vector();
        propList.add(teamHasLeaderProp);
        modifyEntityProperties(groupURI, propList);

        // TODO: Since JTP is not honoring the inverseOf attribute also assert that the actor is the leader of the team
        kaos.core.csi.registration.Property actorIsLeaderOfProp = new kaos.core.csi.registration.PropertyImpl(GroupConcepts.isLeaderOf(),
                "actor leads team", modResult, 0, groupURI);
        propList.remove(0);
        propList.add(actorIsLeaderOfProp);
        modifyEntityProperties(actorId, propList);

        return modResult;
    }

    // --- kaos.core.csi.registration.DomainHierarchyConfiguration interface implementation - end

    public void addEventListener(DirectoryEventListener listener) throws DirectoryFailure
    {
        _dirProxy.addEventListener(listener);
    }

    public void addRemoteEventListener(Locator l) throws DirectoryFailure
    {
        _dirProxy.addRemoteEventListener(l);
    }

    public void removeEventListener(DirectoryEventListener listener) throws DirectoryFailure
    {
        _dirProxy.removeEventListener(listener);
    }

    public void removeRemoteEventListener(Locator l) throws DirectoryFailure
    {
        _dirProxy.removeRemoteEventListener(l);
    }

    protected void logMessage(String msg, int logLevel)
    {
        _logger.logMessage(msg, logLevel);
    }

    protected void logMessage(String msg, Exception e, int logLevel)
    {
        _logger.logMessage(msg, e, logLevel);
    }

    // --- PRIVATE METHODS BEGIN

    private void initialize(String transportName) throws Exception
    {
        try
        {
            _dirProxy = KAoSGuardBootstrapper.getKAoSDirectoryProxyForTransport(transportName);
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new Exception("    \n\n    *** RegistrationImpl: unable to obtain a reference the DirectoryService for transport: " + transportName
                    + "!");
        }

        _transportName = transportName;
    }

    private AgentName acquireAgentName(String guid) throws Exception
    {
        AgentName name = null;

        if (guid != null)
        {
            if (guid.indexOf("#") == -1)
                guid = NamespaceValidator.validateURI(guid, DefaultNamespaces.defaultKAoSActorNamespace);

            name = new AcrAgentName(guid);
        }
        else
        {
            AgentNamingService ans = new KAoSAgentNamingService();
            name = ans.createAgentName();
        }

        return name;
    }

    private Locator createLocator(String nickname, String id, JasBean agentInitInfo, String transportName)
    {
        String agentTransportName = null;

        if (nickname != null)
            agentTransportName = nickname;
        else
            agentTransportName = id;

        TransportSupport transport = CSIFactory.getTransportSupport();
        return transport.createLocator(agentTransportName, agentInitInfo, transportName);

    }
}