/**
 * Last Modified by: $Author: mvignati $
 * On: $Date: 2012/01/26 19:29:58 $
 */

package kaos.core.csi.registration;

import kaos.core.service.directory.ModificationException;
import kaos.policy.guard.GuardInstantiationException;
import kaos.spatial.Block3D;
import kaos.spatial.Location3D;
import kaos.spatial.Orientation3D;

import javax.agent.service.directory.AgentDescription;
import javax.agent.service.directory.AlreadyRegisteredException;
import javax.agent.service.directory.DirectoryFailure;
import javax.agent.service.directory.NotRegisteredException;
import java.util.Set;
import java.util.Vector;

/**
 * Defines the operations to register and deregister entities with spatial properties
 * 
 * @author KAoS Team
 * $Revision: 1.2 $
 */
public interface SpatialRegistration
{
    /** 
     * Register the entity as the simpler method registerEntity but also its spatial properties in the KAoS Directory Service *AND* the SpatialReasoner. 
     *
     * @param objectLocation     Location3D containing the X,Y,Z coordinates of the registering entity.
     *
     * @param objectOrientation  Orientation3D containing the X,Y,Z radius of the registering entity.
     *
     * @param objectOccupiedSpace    Block3D containing corner locations determining the occupied space of the registering entity.
     */
    public void registerEntityWithSpatialProperties(AgentDescription entityDescription, Location3D objectLocation, Orientation3D objectOrientation, 
            Block3D objectOccupiedSpace, boolean guarded, boolean replace) throws DirectoryFailure, AlreadyRegisteredException, GuardInstantiationException;

    /** 
     * Register object identified by the given objectID and its spatial properties in the SpatialReasoner.
     * Important:  Note that the object is *not* registered in the KAoS DirectoryService, only in the SpatialReasoner.
     *             If you want to register a given object in the KAoS Directory service, use method registerEntityWithSpatialProperties(),
     *             which will do both, register in the KAoS DS and the SpatialReasoner.
     *
     * @param objectID               String uniquely identifying registering object in the SpatialReasoner.
     *
     * @param objectOntologyType     Set containing Strings describing the ontological types of the registering object in the SpatialReasoner.
     *
     * @param objectLocation         Location3D containing the X,Y,Z coordinates of the registering object.
     *
     * @param objectOrientation      Orientation3D containing the X,Y,Z radius of the registering object.
     *
     * @param objectOccupiedSpace    Block3D containing corner locations determining the occupied space of the registering object.
     *
     * @throws                   DirectoryFailure if the connection to the SpatialReasoner cannot be established.
     *
     * @throws                   AlreadyRegisteredException if the registering object has been already registered in the SpatialReasoner. 
     */
    public void registerObjectWithSpatialProperties(String objectID, Set objectOntologyType, Location3D objectLocation,  Orientation3D objectOrientation, 
            Block3D objectOccupiedSpace) throws DirectoryFailure, AlreadyRegisteredException;

    /** 
     * Register object identified by the given objectID and its spatial properties in the SpatialReasoner.
     * Important:  Note that the object is *not* registered in the KAoS DirectoryService, only in the SpatialReasoner.
     *             If you want to register a given object in the KAoS Directory service, use method registerEntityWithSpatialProperties(),
     *             which will do both, register in the KAoS DS and the SpatialReasoner.
     *
     * @param objectID               String uniquely identifying registering object in the SpatialReasoner.
     *
     * @param objectOntologyType     Set containing Strings describing the ontological types of the registering object in the SpatialReasoner.
     *
     * @param objectOrientation      Orientation3D containing the X,Y,Z radius of the registering object.
     *
     * @param boundingPoints    	 An ordered List of points used to define the objects boundary. Location should then be defined as the average of the points.
     * 							     Occupied Space should be the max and min of the points along each axis. 
     *
     * @throws                       DirectoryFailure if the connection to the SpatialReasoner cannot be established.
     *
     * @throws                       AlreadyRegisteredException if the registering object has been already registered in the SpatialReasoner. 
     */
    public void registerObjectWithSpatialProperties(String objectID, Set objectOntologyTypes, Orientation3D objectOrientation, Vector<Location3D> boundingPoints)
        throws DirectoryFailure, AlreadyRegisteredException;

    /**
     * Deregister object identified by the given objectID and its spatial properties in the SpatialReasoner.
     * Important:  Note that the object is *not* deregistered from the KAoS DirectoryService, only from the SpatialReasoner.
     * 
     * @param objectID           String uniquely identifying registering object in the SpatialReasoner.
     *
     * @throws                   DirectoryFailure if the connection to the SpatialReasoner cannot be established.
     *
     * @throws                   NotRegisteredException if the registering object has not been registered in the SpatialReasoner.
     */
    public void deregisterObjectWithSpatialProperties(String objectID) throws DirectoryFailure, NotRegisteredException;

    /** 
     * Modify given spatial properties of the entity identified by the given entityGUID in the KAoS Directory Service *AND* the SpatialReasoner.
     *
     * @param entityGUID            String containing the entity GUID.
     *
     * @param objectLocation        modified Location3D containing the X,Y,Z coordinates of the entity.
     *
     * @param objectOrientation     modified Orientation3D containing the X,Y,Z radius of the entity.
     *
     * @param objectOccupiedSpace   modified Block3D containing corner locations determining the occupied space of the entity.
     *
     * @throws                      DirectoryFailure if the connection to the DS cannot be established.
     *
     * @throws                      NotRegisteredException if the entity whose properties are to be modified has not been registered in the DS.
     *
     * @throws                      ModificationException if one or more property update failed.  
     */
    public void modifyEntitySpatialProperties(String entityGUID, Location3D objectLocation, Orientation3D objectOrientation, Block3D objectOccupiedSpace)
        throws DirectoryFailure, NotRegisteredException, ModificationException;

    /** 
     * Modify spatial properties of the object identified by the given objectID in the SpatialReasoner.
     * Important:  Note that the properties are *not* modified in the KAoS DirectoryService, only in the SpatialReasoner.
     *             If you want to modify properties of a given object in the KAoS Directory service, use method modifyEntitySpatialProperties(),
     *             which will do both, modify in the KAoS DS and the SpatialReasoner.
     *
     * @param objectID               String uniquely identifying object, whose spatial properties will be modified in the SpatialReasoner.
     *
     * @param objectLocation         modified Location3D containing the X,Y,Z coordinates of the entity.
     *
     * @param objectOrientation      modified Orientation3D containing the X,Y,Z radius of the entity.
     *
     * @param objectOccupiedSpace    modified Block3D containing corner locations determining the occupied space of the entity.
     *
     * @throws                       DirectoryFailure if the connection to the SpatialReasoner cannot be established.
     *
     * @throws                       NotRegisteredException if the object, whose properties are to be modified has not been registered in the SpatialReasoner.
     */
    public void modifyObjectSpatialProperties(String objectID, Location3D objectLocation, Orientation3D objectOrientation, Block3D objectOccupiedSpace)
        throws DirectoryFailure, NotRegisteredException;
}
