/**
 * Last Modified by: $Author: auszok $ On: $Date: 2012/02/08 17:37:58 $
 */

package kaos.core.csi.request;

import kaos.core.csi.ActionInstanceDescription;
import kaos.core.csi.ActionInstanceDescriptionImpl;
import kaos.core.csi.CSIFactory;
import kaos.core.csi.OntPropertyDescriptionImpl;
import kaos.core.csi.policy.KAoSSecurityException;
import kaos.core.csi.policy.PolicyChecking;
import kaos.core.service.ontology.request.ActionStatusCallback;
import kaos.core.service.ontology.request.RemoteActionStatusCallback;
import kaos.core.service.util.MethodCallRequestHandler;
import kaos.core.util.Logger;
import kaos.policy.builder.KAoSPolicyBuilderUtil;

import javax.agent.service.ServiceFailure;
import java.lang.reflect.Method;
import java.net.URISyntaxException;
import java.util.*;

/**
 * This class implements the operations to receive and process requests for actions. The operations to receive/process requests are defined in the
 * HashMapRequestReceiver and
 * AIDRequestReceiver. In the case of HashMapRequestReceiver, all requests are received in the form of a String containing the request name, a Hashtable,
 * containing request
 * properties and their values, and in some cases, a callback. In the case of AIDRequestReceiver, requests are received in the form of
 * ActionInstanceDescription. Before a request is
 * executed, a permission is checked to make sure that the request (action) is allowed. Requests are executed using the reflection mechanism. Request name
 * is the name of the method
 * to be invoked; properties and callbakc (if present) are the method's arguments.
 */
public abstract class ActionMethodRequestHandler extends MethodCallRequestHandler implements HashMapRequestReceiver, AIDRequestReceiver
{
    protected PolicyChecking _policyService;

    // ////////////////////////////////////////////////////////////////////////////
    // kaos.core.csi.request.HashMapRequestReceiver implementation - begin
    // ////////////////////////////////////////////////////////////////////////////

    /**
     * This method accepts a request described by the given String and HashMap of properties. Feedback from the action execution is sent to the requestor
     * via the given
     * ActionStatusCallback.
     * 
     * @param actionName is a String representing the name of the action as defined in the associated ontology.
     * @param properties is an HashMap of ontological Strings representing the desired properties of the action and the associated values.
     * @param statusCallback ActionStatusCallback to be called to pass back to the action requestor information about progression of the action execution
     *            and the final result.
     * @throws ActionFailureException containing the reason for failure. See ontology to determine specific reasons available for a given action
     **/
    public void processHashMapRequest(String actionName, HashMap properties, ActionStatusCallback statusCallback) throws ActionFailureException
    {
        if ((actionName == null) || (properties == null) || (statusCallback == null))
            throw new ActionFailureException("Required argument is null: actionName = " + actionName + " properties = " + properties + " statusCallback: "
                    + statusCallback);

        try
        {
            checkPermissionForAction(actionName, properties);
        }
        catch (ActionFailureException ae)
        {
            if (statusCallback != null)
            {
                if (statusCallback instanceof RemoteActionStatusCallback)
                {
                    RemoteActionStatusCallback rCallback = (RemoteActionStatusCallback)statusCallback;
                    rCallback.setTransportInfo(myMts, myLocator, myNickName);

                }
                statusCallback.onAborted(new GregorianCalendar(), ae);
            }
            return;
        }

        try
        {
            executeHashMapBasedAction(actionName, properties, statusCallback);
        }
        catch (Exception e)
        {
            if (e instanceof ActionFailureException)
            {
                throw (ActionFailureException)e;
            }
            else
            {
                throw new ActionFailureException(e.getMessage());
            }
        }
    }

    /**
     * This method accepts a request described by the given String containing the action name. Feedback from the action execution is sent to the requestor
     * via the given
     * ActionStatusCallback.
     * 
     * @param actionName is a String representing the name of the action as defined in the associated ontology.
     * @param statusCallback ActionStatusCallback to be called to pass back to the action requestor information about progression of the action execution
     *            and the final result.
     * @throws ActionFailureException containing the reason for failure. See ontology to determine specific reasons available for a given action
     **/
    public void processHashMapRequest(String actionName, ActionStatusCallback statusCallback) throws ActionFailureException
    {
        if ((actionName == null) || (statusCallback == null))
            throw new ActionFailureException("Required argument is null: actionName = " + actionName + " statusCallback: " + statusCallback);

        try
        {
            checkPermissionForAction(actionName, new HashMap());
        }
        catch (ActionFailureException ae)
        {
            if (statusCallback != null)
            {
                if (statusCallback instanceof RemoteActionStatusCallback)
                {
                    RemoteActionStatusCallback rCallback = (RemoteActionStatusCallback)statusCallback;
                    rCallback.setTransportInfo(myMts, myLocator, myNickName);

                }
                statusCallback.onAborted(new GregorianCalendar(), ae);
            }
            return;
        }

        try
        {
            executeHashMapBasedAction(actionName, null, statusCallback);
        }
        catch (Exception e)
        {
            if (e instanceof ActionFailureException)
            {
                throw (ActionFailureException)e;
            }
            else
            {
                throw new ActionFailureException(e.getMessage());
            }
        }
    }

    /**
     * This method accepts a request described by the given String and HashMap of properties. The result is returned synchronously.
     * 
     * @param actionName is a String representing the name of the action as defined in the associated ontology.
     * @param properties is an HashMap of ontological Strings representing the desired properties of the action and the associated values.
     * @return Object containing the final result of the action.
     * @throws ActionFailureException containing the reason for failure. See ontology to determine specific reasons available for a given action
     **/
    public Object processHashMapRequest(String actionName, HashMap properties) throws ActionFailureException
    {
        if ((actionName == null) || (properties == null))
            throw new ActionFailureException("Required argument is null: actionName = " + actionName + " properties: " + properties);

        checkPermissionForAction(actionName, properties);

        Object result = null;
        try
        {
            result = executeHashMapBasedAction(actionName, properties, null);
        }
        catch (Exception e)
        {
            if (e instanceof ActionFailureException)
            {
                throw (ActionFailureException)e;
            }
            else
            {
                throw new ActionFailureException(e.getMessage());
            }
        }
        return result;
    }

    /**
     * This method accepts a request described by the given String containing action name. The result is returned synchronously.
     * 
     * @param actionName is a String representing the name of the action as defined in the associated ontology.
     * @return Object containing the final result of the action.
     * @throws ActionFailureException containing the reason for failure. See ontology to determine specific reasons available for a given action
     **/
    public Object processHashMapRequest(String actionName) throws ActionFailureException
    {
        if (actionName == null)
            throw new ActionFailureException("Required argument is null: actionName = " + actionName);

        checkPermissionForAction(actionName, new HashMap());

        Object result = null;
        try
        {
            result = executeHashMapBasedAction(actionName, null, null);
        }
        catch (Exception e)
        {
            if (e instanceof ActionFailureException)
            {
                throw (ActionFailureException)e;
            }
            else
            {
                throw new ActionFailureException(e.getMessage());
            }
        }
        return result;
    }

    // ////////////////////////////////////////////////////////////////////////////
    // kaos.core.csi.request.HashMapRequestReceiver implementation - end
    // ////////////////////////////////////////////////////////////////////////////

    // ////////////////////////////////////////////////////////////////////////////
    // kaos.core.csi.request.AIDRequestReceiver implementation - begin
    // ////////////////////////////////////////////////////////////////////////////

    /**
     * This method accepts a request described by an ActionInstanceDescription to perform an action. Feedback from the action execution is sent to the
     * requestor via the given
     * ActionStatusCallback.
     * 
     * @param aid is an ActionInstanceDescription that describes the action. It includes the action name and all action properties and their values.
     * @param statusCallback ActionStatusCallback to be called to pass back to the action requestor information about progression of the action execution
     *            and the final result.
     * @throws ActionFailureException containing the reason for failure. See ontology to determine specific reasons available for a given action
     **/
    public void processAIDRequest(ActionInstanceDescription aid, ActionStatusCallback statusCallback) throws ActionFailureException
    {
        if ((aid == null) || (statusCallback == null))
            throw new ActionFailureException("Required argument is null: aid = " + aid + " statusCallback: " + statusCallback);

        String actionName = getActionMethodNameFromAID(aid);

        if (actionName == null)
            throw new ActionFailureException("Action name is null, unable to continue processing of: " + aid);

        HashMap properties = getPropertiesFromAID(aid);

        if ((properties == null) || (properties.size() == 0))
        {
            processHashMapRequest(actionName, statusCallback);
        }
        else
        {
            processHashMapRequest(actionName, properties, statusCallback);
        }

    }

    /**
     * This method accepts a request described in an ActionInstanceDescription to perform an action. The result is returned synchronously.
     * 
     * @param aid is an ActionInstanceDescription that describes the action. It includes the action name and all action properties and their values.
     * @return Object containing the final result of the action.
     * @throws ActionFailureException containing the reason for failure. See ontology to determine specific reasons available for a given action
     **/
    public Object processAIDRequest(ActionInstanceDescription aid) throws ActionFailureException
    {
        if ((aid == null))
            throw new ActionFailureException("Required argument is null: aid = " + aid);

        String actionName = getActionMethodNameFromAID(aid);

        if (actionName == null)
            throw new ActionFailureException("Action name is null, unable to continue processing of: " + aid);

        HashMap properties = getPropertiesFromAID(aid);

        if ((properties == null) || (properties.size() == 0))
        {
            return processHashMapRequest(actionName);
        }
        else
        {
            return processHashMapRequest(actionName, properties);
        }
    }

    // ////////////////////////////////////////////////////////////////////////////
    // kaos.core.csi.request.AIDRequestReceiver implementation - end
    // ////////////////////////////////////////////////////////////////////////////

    /**
    */
    protected void bindToPolicyService(String transportName) throws Exception
    {
        CSIFactory.setCurrentTransport(transportName);
        _policyService = CSIFactory.getPolicyChecking();
    }

    /**
     */
    protected Object executeHashMapBasedAction(String actionName, HashMap properties, ActionStatusCallback statusCallback) throws Exception
    {
        Object result = null;

        if (_logger.isEnabledFor(Logger.INFO))
            _logger.logMessage("executeAction:  action name: " + actionName, Logger.INFO);

        String methodName = KAoSPolicyBuilderUtil.getShortName(actionName);
        if (_logger.isEnabledFor(Logger.INFO))
            _logger.logMessage("executeAction: Method requested: " + methodName + "\n   action properties: " + properties + "\n   statusCallback: "
                    + statusCallback, Logger.INFO);

        if (methodName == null)
            throw new Exception("Action method name null! Unable to proceed!");

        try
        {
            if ((properties == null) && (statusCallback == null))
            {
                // invoke method with no args
                result = this.getClass().getMethod(methodName, (Class[])null).invoke(this, (Object[])null);
            }
            else
            {
                if (statusCallback != null)
                {
                    if (statusCallback instanceof RemoteActionStatusCallback)
                    {
                        RemoteActionStatusCallback rCallback = (RemoteActionStatusCallback)statusCallback;
                        rCallback.setTransportInfo(myMts, myLocator, myNickName);

                    }
                }

                List argObjectList = new ArrayList();
                if ((properties != null) && (statusCallback != null))
                {
                    argObjectList.add(properties);
                    argObjectList.add(statusCallback);
                }
                else if ((properties != null))
                {
                    argObjectList.add(properties);
                }
                else if (statusCallback != null)
                {
                    argObjectList.add(statusCallback);
                }

                Object argObjects[] = argObjectList.toArray(new Object[] {});
                // we have to check this the "hard way" because Class.getMethod(name, args[])
                // does not work if the args are subclasses of the parameter classes

                // find the method matching the method name requested
                Method methods[] = this.getClass().getMethods();
                boolean foundMethod = false;
                for (int i = 0; i < methods.length; i++)
                {
                    if (methods[i].getName().equals(methodName))
                    {
                        // check to see if the parameter classes match the arguments
                        Class parameters[] = methods[i].getParameterTypes();
                        boolean classesMatch = false;
                        if (parameters.length == argObjects.length)
                        {
                            classesMatch = true;
                            for (int j = 0; j < parameters.length; j++)
                            {
                                if (!parameters[j].isInstance(argObjects[j]))
                                {
                                    classesMatch = false;
                                    break;
                                }
                            }
                        }
                        // if we found the right method, execute it
                        if (classesMatch)
                        {
                            foundMethod = true;
                            result = methods[i].invoke(this, argObjects);
                            break;
                        }
                    }
                }
                // if we didn't find the method, throw an exception
                if (!foundMethod)
                {
                    throw new NoSuchMethodException(methodName);
                }
            }
        }
        catch (Exception xcp)
        {
            if (_logger.isEnabledFor(Logger.ERROR))
                _logger.logMessage("", xcp, Logger.ERROR);
            throw xcp;

        }
        return result;
    }

    protected void checkPermissionForAction(String actionName, HashMap properties) throws ActionFailureException
    {
        if (_policyService == null)
            return;

        try
        {
            ActionInstanceDescription aid = new ActionInstanceDescriptionImpl(null, myAgentDescription.getEntityNameAsString(), actionName,
                    properties);
            _policyService.checkPermission(aid, null);
        }
        catch (KAoSSecurityException kse)
        {
            kse.printStackTrace();
            throw new ActionFailureException("Action: " + actionName + " NOT PERMITTED! Check Permission failed for properties: " + properties);
        }
        catch (NullPointerException npe)
        {
            npe.printStackTrace();
            throw new ActionFailureException("Action: " + actionName
                    + " NOT PERMITTED! Check Permission failed due to NullPointerException. Specified roperties: " + properties);
        }
        catch (ServiceFailure sfe)
        {
            sfe.printStackTrace();
            throw new ActionFailureException("Action: " + actionName + " NOT PERMITTED due to security service failure! Specified action properties: "
                    + properties);
        }
        catch (URISyntaxException urie)
        {
            urie.printStackTrace();
            throw new ActionFailureException("Action: " + actionName + " NOT PERMITTED due to security service failure! Specified action properties: "
                    + properties);
        }

    }

    private String getActionMethodNameFromAID(ActionInstanceDescription aid)
    {
        String actionName = aid.getActionClassName();
        if (_logger.isEnabledFor(Logger.DEBUG))
            _logger.logMessage("getActionMethodNameFromAID: got long name: " + actionName, Logger.DEBUG);
        String methodName = KAoSPolicyBuilderUtil.getShortName(actionName);
        if (_logger.isEnabledFor(Logger.DEBUG))
            _logger.logMessage("getActionMethodNameFromAID: got short name: " + methodName, Logger.DEBUG);
        return methodName;
    }

    public HashMap getPropertiesFromAID(ActionInstanceDescription aid)
    {
        // get arguments
        Set targets = aid.getProperties();
        Iterator targetIt = targets.iterator();
        HashMap properties = new HashMap();
        while (targetIt.hasNext())
        {
            OntPropertyDescriptionImpl tid = (OntPropertyDescriptionImpl)targetIt.next();
            Set tidValues = tid.getValues();
            Object value = tid.getValueDescForValue((String)tidValues.toArray()[1]);
            if (value == null)
            {
                value = (String)tidValues.toArray()[1];
            }
            else
            {
                /*
                 * try{ } catch(Exception e){ e.printStackTrace(); }
                 */
            }

            properties.put(tid.getName(), value);
        }

        return properties;
    }
}