/**
 * Last Modified by: $Author: mvignati $
 * On: $Date: 2012/01/26 19:30:02 $
 */

package kaos.core.csi.request;

import kaos.core.service.directory.KAoSEntityDescription;

import javax.agent.service.transport.NotLocatableException;
import javax.agent.service.transport.TransportFailure;

/**
 * Defines the operations to send a request expressed by an OWL String. 
 * 
 * @author KAoS Team
 */
public interface OWLRequestManager 
{
    /**
	 * This method sends a request to an entity identified by the given receiverID to perform an action described  
	 * by the given String, which is in the OWL format.  Feedback from the action execution is sent to the
	 * given feedbackReceiver.
	 * 
	 * @param		feedbackReceiver	    is the FeedbackReceiver requesting the action.  
	 *							            All feedback from the action execution will be sent to it.
	 * 
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 * 
	 * @param	    owlActionDescription    is an OWL String that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		String		            containing the unique id of the requested action. This id may be used
	 *                                      by the FeedbackReceiver to track the status of action execution.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *				                        
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 * 
	 **/
    public String sendOWLRequest(FeedbackReceiver feedbackReceiver, 
							     String receiverID, 
                                 String owlActionDescription)
		throws ActionFailureException, NotLocatableException, TransportFailure;
		
	/**
	 * This method sends a request to an entity identified by the given receiverDesc to perform an action described  
	 * by the given String, which is in the OWL format.  Feedback from the action execution is sent to the
	 * given feedbackReceiver.
	 * 
	 * @param		feedbackReceiver        is the FeedbackReceiver requesting the action.  
	 *							            All feedback from the action execution will be sent to it.
	 * 
	 * @param	    receiverDesc	        is a KAosEntityDescription of the entity, which is requested
	 *								        to perform the action.
	 * 
	 * @param	    owlActionDescription    is an OWL String that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		String		            containing the unique id of the requested action. This id may be used
	 *                                      by the FeedbackReceiver to track the status of the action execution.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 *				                        
	 **/
    public String sendOWLRequest(FeedbackReceiver feedbackReceiver,
								 KAoSEntityDescription receiverDesc,
								 String owlActionDescription)
		throws ActionFailureException, NotLocatableException, TransportFailure;

		
	/**
	 * This method sends a request described by an OWL String to perform an action by an entity identified 
	 * by the receiverID.
	 * The result is returned synchronously to the caller.
	 *
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 *
	 * @param	    owlActionDescription    is an OWL String that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		Object		            containing the final result of the action.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 
	 **/
    public Object sendOWLRequest(String receiverID, 
                                 String owlActionDescription)
        throws ActionFailureException, NotLocatableException, TransportFailure;
        
    /**
	 * This method sends a request described by an OWL String to perform an action by an entity identified 
	 * by the KAoSEntityDescription.
	 * The result is returned synchronously to the caller.
	 *
	 * @param	    receiverDesc	        is the KAoSEntityDescription that identifies the entity, which is  
	 *							            requested to perform the action.  
	 *
	 * @param	    owlActionDescription    is an OWL String that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		Object		            containing the final result of the action.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 
	 **/
    public Object sendOWLRequest(KAoSEntityDescription receiverDesc, 
                                 String owlActionDescription)
        throws ActionFailureException, NotLocatableException, TransportFailure;


}
