package kaos.core.csi.request;

import kaos.core.csi.ActionInstanceDescription;
import kaos.core.csi.CSIFactory;
import kaos.core.csi.query.QueryRegistration;
import kaos.core.csi.transport.TransportSupport;
import kaos.core.service.directory.KAoSAgentDescription;
import kaos.core.service.directory.KAoSEntityDescription;
import kaos.core.service.ontology.request.RemoteActionStatusCallback;
import kaos.core.service.util.RemoteActionCallbackImpl;
import kaos.core.util.*;
import ri.acr.FipaAclPerformative;

import javax.agent.Locator;
import javax.agent.TransportMessage;
import javax.agent.service.directory.DirectoryFailure;
import javax.agent.service.directory.SearchException;
import javax.agent.service.transport.MessageListener;
import javax.agent.service.transport.NotLocatableException;
import javax.agent.service.transport.TransportFailure;
import java.lang.reflect.Method;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Vector;

/**
 * This class implements the RequestManager interface, which defines sending 
 * requests to entities implementing RequestReceiver interfaces.
 */
public class RequestManagerImpl
    implements RequestManager, MessageListener, FeedbackReceiver 
               
{
    
    private static final RequestManagerImpl     __INSTANCE = new RequestManagerImpl() ;
    private static boolean                      __INITIALIZED = false;
    private static final String                 _NICKNAME_ROOT = "CSI-Request-";
    
    protected Locator                           _myLocator;
    protected String                            _nickname;
    protected QueryRegistration                 _query;
    protected TransportSupport                  _transport;
    protected String                            _transportName;	
    protected Hashtable                         _actionRequestors = new Hashtable();
    protected Hashtable                         _waiters;
    
	private Logger		                        _logger = Logger.getLogger(this);
    
    /**
     * Retrieve a reference to the RequestManager.
     *
	 * @param       transportName   is a String containing  the name of the transport to use by the RequestManager.
	 *                                      
     * @return a reference to the RequestManager
     */
    public static RequestManager getInstance(String transportName)
        throws Exception
    {
        if ( __INITIALIZED ) return __INSTANCE;
            TransportSupport transport = CSIFactory.getTransportSupport();
            if ( (transportName == null) || ( ! transport.isTransportSupported(transportName) )  )
                throw new Exception("RequestManager: transport name is invalid: " + transportName 
                                    + "! Please specify valid transport! Valid transport names: "
                                    + transport.getSupportedTransportsNames());
            __INSTANCE.initialize(transportName, _NICKNAME_ROOT + ExecEnvUtility.getVMID());
            __INITIALIZED = true;
            return __INSTANCE;
    }
    
    private RequestManagerImpl()
    {
    }

    /*****************************************************************************************
       kaos.core.csi.request.HashMapRequestManager interface implementation - begin
    *****************************************************************************************/
	/**
	 * This method sends a request to an entity identified by the given receiverID to perform an action described  
	 * by the given String and a HashMap containing the action properties.  Feedback from the action execution is sent to the
	 * given feedbackReceiver.
	 * 
	 * @param		feedbackReceiver	    is the FeedbackReceiver requesting the action.  
	 *							            All feedback from the action execution will be sent to it.
	 * 
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 * 
     * @param       actionName		        is a String representing the name of the action as defined in 
	 *                                      the associated ontology.
	 *
	 * @param		properties	            is a HashMap of ontological Strings representing 
	 *							            the desired properties of the action and the 
	 *							            associated values.
	 *
	 * @return		String		            containing the unique id of the requested action. This id may be used
	 *                                      by the FeedbackReceiver to track the status of action execution.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *				                        
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 * 
	 **/
    public String sendHashMapRequest(FeedbackReceiver feedbackReceiver, 
							         String receiverID, 
                                     String actionName,
								     HashMap properties)
		throws ActionFailureException, NotLocatableException, TransportFailure
	{
	    if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("sendHashMapRequest: received receiverID " + receiverID + "\n   action name: " 
		                     + actionName + "\n   action properties: " + properties, Logger.INFO);  
		validateFeedbackReceiver(feedbackReceiver);
				
        KAoSAgentDescription receiverDesc = getAgentDescription(receiverID);
        
        if ( receiverDesc == null )
            throw new NotLocatableException("Unable to locate the request receiver: " + receiverID);
		
        return sendHashMapRequest(feedbackReceiver,
							      receiverDesc,
                                  actionName,
                                  properties);
	
	}

    /**
	 * This method sends a request to an entity identified by the given receiverID to perform an action described  
	 * by the given String.  Feedback from the action execution is sent to the
	 * given feedbackReceiver.
	 * 
	 * @param		feedbackReceiver	    is the FeedbackReceiver requesting the action.  
	 *							            All feedback from the action execution will be sent to it.
	 * 
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 * 
     * @param       actionName		        is a String representing the name of the action as defined in 
	 *                                      the associated ontology.
	 *
	 * @return		String		            containing the unique id of the requested action. This id may be used
	 *                                      by the FeedbackReceiver to track the status of action execution.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *				                        
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 * 
	 **/
    public String sendHashMapRequest(FeedbackReceiver feedbackReceiver, 
							         String receiverID, 
                                     String actionName)
		throws ActionFailureException, NotLocatableException, TransportFailure
	{
	    if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("sendHashMapRequest: received receiverID " + receiverID + "\n   action name: " 
		                     + actionName, Logger.INFO);  
		validateFeedbackReceiver(feedbackReceiver);
				
        KAoSAgentDescription receiverDesc = getAgentDescription(receiverID);
        
        return sendHashMapRequest(feedbackReceiver,
							      receiverDesc,
                                  actionName);
	
	}
	
	/**
	 * This method sends a request to an entity identified by the given receiverDesc to perform an action described  
	 * by the given String and a HashMap containing the action properties.  Feedback from the action execution is sent to the
	 * given feedbackReceiver.
	 * 
	 * @param		feedbackReceiver        is the FeedbackReceiver requesting the action.  
	 *							            All feedback from the action execution will be sent to it.
	 * 
	 * @param	    receiverDesc	        is a KAosEntityDescription of the entity, which is requested
	 *								        to perform the action.
	 * 
	 * @param       actionName		        is a String representing the name of the action as defined in 
	 *                                      the associated ontology.
	 *
	 * @param		properties	            is a HashMap of ontological Strings representing 
	 *							            the desired properties of the action and the 
	 *							            associated values.
	 * 
	 * @return		String		            containing the unique id of the requested action. This id may be used
	 *                                      by the FeedbackReceiver to track the status of the action execution.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 *				                        
	 **/
    public String sendHashMapRequest(FeedbackReceiver feedbackReceiver,
								     KAoSEntityDescription receiverDesc,
								     String actionName,
								     HashMap properties)
		throws ActionFailureException, NotLocatableException, TransportFailure
	{
	    if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("requestAction: received receiver description: \n" + receiverDesc
		                     + "\n  action name: " + actionName + "\n  properties: " + properties, Logger.INFO);
		                 
	    validateFeedbackReceiver(feedbackReceiver);
	    String requestID = sendHashMapBasedReqMsg(receiverDesc,
						                          actionName,
						                          properties,
						                          false);
		_actionRequestors.put(requestID,feedbackReceiver);
        return requestID;
	}
	
	/**
	 * This method sends a request to an entity identified by the given receiverDesc to perform an action described  
	 * by the given String.  Feedback from the action execution is sent to the
	 * given feedbackReceiver.
	 * 
	 * @param		feedbackReceiver        is the FeedbackReceiver requesting the action.  
	 *							            All feedback from the action execution will be sent to it.
	 * 
	 * @param	    receiverDesc	        is a KAosEntityDescription of the entity, which is requested
	 *								        to perform the action.
	 * 
	 * @param       actionName		        is a String representing the name of the action as defined in 
	 *                                      the associated ontology.
	 *
	 * @return		String		            containing the unique id of the requested action. This id may be used
	 *                                      by the FeedbackReceiver to track the status of the action execution.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 *				                        
	 **/
    public String sendHashMapRequest(FeedbackReceiver feedbackReceiver,
								     KAoSEntityDescription receiverDesc,
								     String actionName)
		throws ActionFailureException, NotLocatableException, TransportFailure
	{
	    if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("requestAction: received receiver description: \n" + receiverDesc
		                     + "\n  action name: " + actionName, Logger.INFO);
		                 
	    validateFeedbackReceiver(feedbackReceiver);
	    String requestID = sendHashMapBasedReqMsg(receiverDesc,
						                          actionName,
						                          false);
		_actionRequestors.put(requestID,feedbackReceiver);
        return requestID;
	}
	
    /**
	 * This method sends a request described by the given String and a HashMap containing the action properties 
	 * to perform an action by an entity identified by the receiverID.
	 * 
	 * The result is returned synchronously to the caller.
	 *
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 *
     * @param       actionName		        is a String representing the name of the action as defined in 
	 *                                      the associated ontology.
	 *
	 * @param		properties	            is a HashMap of ontological Strings representing 
	 *							            the desired properties of the action and the 
	 *							            associated values.
	 * 
	 * @return		Object		            containing the final result of the action.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 
	 **/
    public Object sendHashMapRequest(String receiverID, 
                                     String actionName,
								     HashMap properties)
        throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendHashMapRequest: received receiverID " + receiverID + "\n   action name: " 
		                     + actionName + "\n   action properties: " + properties, Logger.INFO);  				
		// Get receiver descritption		
        KAoSAgentDescription receiverDesc = getAgentDescription(receiverID);
        return sendHashMapRequest (receiverDesc,
							       actionName,
							       properties);
    }
    
    /**
	 * This method sends a request described by the given String containing the action name
	 * to perform the action by an entity identified by the receiverID.
	 * 
	 * The result is returned synchronously to the caller.
	 *
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 *
     * @param       actionName		        is a String representing the name of the action as defined in 
	 *                                      the associated ontology.
	 *
	 * @return		Object		            containing the final result of the action.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 **/
    public Object sendHashMapRequest(String receiverID, 
                                     String actionName)
        throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendHashMapRequest: received receiverID " + receiverID + "\n   action name: " 
		                     + actionName, Logger.INFO);  
		                     
        KAoSAgentDescription receiverDesc = getAgentDescription(receiverID);
        return sendHashMapRequest (receiverDesc,
							       actionName);
    }
    
    /**
	 * This method sends a request described by a String and a HashMap containing the action properties to perform 
	 * by an entity identified by the KAoSEntityDescription.
	 * 
	 * The result is returned synchronously to the caller.
	 *
	 * @param	    receiverDesc	        is the KAoSEntityDescription that identifies the entity, which is  
	 *							            requested to perform the action.  
	 *
	  * @param      actionName		        is a String representing the name of the action as defined in 
	 *                                      the associated ontology.
	 *
	 * @param		properties	            is a HashMap of ontological Strings representing 
	 *							            the desired properties of the action and the 
	 *							            associated values.
	 *
	 * @return		Object		            containing the final result of the action.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 
	 **/
    public Object sendHashMapRequest(KAoSEntityDescription receiverDesc, 
                                     String actionName,
								     HashMap properties)
        throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendHashMapRequest: received receiverDesc " + receiverDesc + "\n   action name: " 
		                      + actionName + "\n   action properties: " + properties, Logger.INFO); 
		                  		                  
        String requestID = sendHashMapBasedReqMsg(receiverDesc,
						                          actionName,
						                          properties,
						                          true);
		return getResult(requestID);
						                     
    }
    
    /**
	 * This method sends a request described by a String and a HashMap containing the action properties to perform 
	 * by an entity identified by the KAoSEntityDescription.
	 * 
	 * The result is returned synchronously to the caller.
	 *
	 * @param	    receiverDesc	        is the KAoSEntityDescription that identifies the entity, which is  
	 *							            requested to perform the action.  
	 *
	  * @param      actionName		        is a String representing the name of the action as defined in 
	 *                                      the associated ontology.
	 *
	 * @return		Object		            containing the final result of the action.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 
	 **/
    public Object sendHashMapRequest(KAoSEntityDescription receiverDesc, 
                                     String actionName)
        throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendHashMapRequest: received receiverDesc " + receiverDesc + "\n   action name: " 
		                      + actionName, Logger.INFO); 
		                  		                  
        String requestID = sendHashMapBasedReqMsg(receiverDesc,
						                          actionName,
						                          true);
		return getResult(requestID);
						                     
    }
    /*****************************************************************************************
       kaos.core.csi.request.HashMapRequestManager interface implementation - end
    *****************************************************************************************/
    
    /*****************************************************************************************
       kaos.core.csi.request.AIDRequestManager interface implementation - begin
    *****************************************************************************************/
     /**
	 * This method sends a request to an entity identified by the given receiverID to perform an action described  
	 * by the given ActionInstanceDescription.  Feedback from the action execution is sent to the
	 * given feedbackReceiver.
	 * 
	 * @param		feedbackReceiver	    is the FeedbackReceiver requesting the action.  
	 *							            All feedback from the action execution will be sent to it.
	 * 
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 * 
	 * @param	    aid                     is an ActionInstanceDescription that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		String		            containing the unique id of the requested action. This id may be used
	 *                                      by the FeedbackReceiver to track the status of action execution.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *				                        
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 * 
	 **/
    public String sendAIDRequest(FeedbackReceiver feedbackReceiver, 
							     String receiverID, 
                                 ActionInstanceDescription aid)
		throws ActionFailureException, NotLocatableException, TransportFailure
	{
	    if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("sendAIDRequest: received receiverID " + receiverID + "\n   AID: " 
		                     + aid, Logger.INFO);  
		validateFeedbackReceiver(feedbackReceiver);
				
        KAoSAgentDescription receiverDesc = getAgentDescription(receiverID);
		
        return sendAIDRequest(feedbackReceiver,
					          receiverDesc,
                              aid);
	}
		
	/**
	 * This method sends a request to an entity identified by the given receiverDesc to perform an action described  
	 * by the given ActionInstanceDescription.  Feedback from the action execution is sent to the
	 * given feedbackReceiver.
	 * 
	 * @param		feedbackReceiver        is the FeedbackReceiver requesting the action.  
	 *							            All feedback from the action execution will be sent to it.
	 * 
	 * @param	    receiverDesc	        is a KAosEntityDescription of the entity, which is requested
	 *								        to perform the action.
	 * 
	 * @param	    aid                     is an ActionInstanceDescription that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		String		            containing the unique id of the requested action. This id may be used
	 *                                      by the FeedbackReceiver to track the status of the action execution.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 *				                        
	 **/
    public String sendAIDRequest(FeedbackReceiver feedbackReceiver,
								 KAoSEntityDescription receiverDesc,
								 ActionInstanceDescription aid)
		throws ActionFailureException, NotLocatableException, TransportFailure
		
	{
	    if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("sendAIDRequest: received receiver description: \n" + receiverDesc
		                     + "\n  aid: " + aid, Logger.INFO);
		                 
	    validateFeedbackReceiver(feedbackReceiver);
	    String requestID = sendAIDBasedReqMsg(receiverDesc,
						                      aid,
						                      false);
		_actionRequestors.put(requestID,feedbackReceiver);
        return requestID;
	}

	/**
	 * This method sends a request described by an ActionInstanceDescription to perform an action by an entity identified 
	 * by the receiverID.
	 * The result is returned synchronously to the caller.
	 *
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 *
	 * @param	    aid                     is an ActionInstanceDescription that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		Object		            containing the final result of the action.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 
	 **/
    public Object sendAIDRequest(String receiverID, 
                                 ActionInstanceDescription aid)
        throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendAIDRequest: received receiverID " + receiverID + "\n   aid: " 
		                     + aid, Logger.INFO); 
		                     
        KAoSAgentDescription receiverDesc = getAgentDescription(receiverID);
        return sendAIDRequest(receiverDesc,
							  aid);
    }
        
    /**
	 * This method sends a request described by an ActionInstanceDescription to perform an action by an entity identified 
	 * by the KAoSEntityDescription.
	 * The result is returned synchronously to the caller.
	 *
	 * @param	    receiverDesc	        is the KAoSEntityDescription that identifies the entity, which is  
	 *							            requested to perform the action.  
	 *
	 * @param	    aid                     is an ActionInstanceDescription that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		Object		            containing the final result of the action.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 
	 **/
    public Object sendAIDRequest(KAoSEntityDescription receiverDesc, 
                                 ActionInstanceDescription aid)
        throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendAIDRequest: received receiverDesc " + receiverDesc + "\n   aid: " 
		                      + aid, Logger.INFO); 
		                  		                  
        String requestID = sendAIDBasedReqMsg(receiverDesc,
						                      aid,
		                                      true);
						                     
		return getResult(requestID);
    }

    /*****************************************************************************************
       kaos.core.csi.request.AIDRequestManager interface implementation - end
    *****************************************************************************************/
    
    /*****************************************************************************************
       kaos.core.csi.request.OWLRequestManager interface implementation - begin
    *****************************************************************************************/
    
    /**
	 * This method sends a request to an entity identified by the given receiverID to perform an action described  
	 * by the given String, which is in the OWL format.  Feedback from the action execution is sent to the
	 * given feedbackReceiver.
	 * 
	 * @param		feedbackReceiver	    is the FeedbackReceiver requesting the action.  
	 *							            All feedback from the action execution will be sent to it.
	 * 
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 * 
	 * @param	    owlActionDescription    is an OWL String that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		String		            containing the unique id of the requested action. This id may be used
	 *                                      by the FeedbackReceiver to track the status of action execution.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *				                        
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 * 
	 **/
    public String sendOWLRequest(FeedbackReceiver feedbackReceiver, 
							     String receiverID, 
                                 String owlActionDescription)
		throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("sendOWLRequest: received receiverID " + receiverID + "\n   OWL: " 
		                     + owlActionDescription, Logger.INFO);  
		validateFeedbackReceiver(feedbackReceiver);
				
        KAoSAgentDescription receiverDesc = getAgentDescription(receiverID);
		
        return sendOWLRequest(feedbackReceiver,
					          receiverDesc,
                              owlActionDescription);
    }
		
	/**
	 * This method sends a request to an entity identified by the given receiverDesc to perform an action described  
	 * by the given String, which is in the OWL format.  Feedback from the action execution is sent to the
	 * given feedbackReceiver.
	 * 
	 * @param		feedbackReceiver        is the FeedbackReceiver requesting the action.  
	 *							            All feedback from the action execution will be sent to it.
	 * 
	 * @param	    receiverDesc	        is a KAosEntityDescription of the entity, which is requested
	 *								        to perform the action.
	 * 
	 * @param	    owlActionDescription    is an OWL String that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		String		            containing the unique id of the requested action. This id may be used
	 *                                      by the FeedbackReceiver to track the status of the action execution.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 *				                        
	 **/
    public String sendOWLRequest(FeedbackReceiver feedbackReceiver,
								 KAoSEntityDescription receiverDesc,
								 String owlActionDescription)
		throws ActionFailureException, NotLocatableException, TransportFailure
    {
         if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("sendOWLRequest: received receiver description: \n" + receiverDesc
		                     + "\n  OWL Request: " + owlActionDescription, Logger.INFO);
		                 
	    validateFeedbackReceiver(feedbackReceiver);
	    String requestID = sendOWLBasedReqMsg(receiverDesc,
						                      owlActionDescription,
						                      false);
		_actionRequestors.put(requestID,feedbackReceiver);
        return requestID;
    }
		
	/**
	 * This method sends a request described by an OWL String to perform an action by an entity identified 
	 * by the receiverID.
	 * The result is returned synchronously to the caller.
	 *
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 *
	 * @param	    owlActionDescription    is an OWL String that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		Object		            containing the final result of the action.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 
	 **/
    public Object sendOWLRequest(String receiverID, 
                                 String owlActionDescription)
        throws ActionFailureException, NotLocatableException, TransportFailure
    {
         if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendOWLRequest: received receiverID " + receiverID + "\n   OWL Request: " 
		                     + owlActionDescription, Logger.INFO);  				

        KAoSAgentDescription receiverDesc = getAgentDescription(receiverID);
        return sendOWLRequest(receiverDesc,
							  owlActionDescription);
    }
        
    /**
	 * This method sends a request described by an OWL String to perform an action by an entity identified 
	 * by the KAoSEntityDescription.
	 * The result is returned synchronously to the caller.
	 *
	 * @param	    receiverDesc	        is the KAoSEntityDescription that identifies the entity, which is  
	 *							            requested to perform the action.  
	 *
	 * @param	    owlActionDescription    is an OWL String that describes the action.  
	 *							            It includes the action name and all action properties and their values.
	 * 
	 * @return		Object		            containing the final result of the action.
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 
	 **/
    public Object sendOWLRequest(KAoSEntityDescription receiverDesc, 
                                 String owlActionDescription)
        throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendOWLRequest: received receiverDesc " + receiverDesc + "\n   OWL Request: " 
		                      + owlActionDescription, Logger.INFO); 
		                  		                  
        String requestID = sendOWLBasedReqMsg(receiverDesc,
						                      owlActionDescription,
		                                      true);
		return getResult(requestID);
     }
    
    /*****************************************************************************************
       kaos.core.csi.request.OWLRequestManager interface implementation - end
    *****************************************************************************************/  
  
    /*****************************************************************************************
    kaos.core.csi.request.ObligationRequestManager interface implementation - begin
    *****************************************************************************************/
    /**
	 * 
	 * This method sends an obligation request to an entity identified by the given receiverDesc to execute an obligation action described  
	 * by the given obligation ActionInstanceDescription and the trigger ActionInstanceDescription.
	 *  
	 * @param	    receiverDesc	        is a KAosEntityDescription of the entity, which is requested
	 *								        to execute the obligation request.
	 * 
	 * @param       obligationAID		 	is an ActionInstanceDescription containing the obligation action name
	 * 										and its properties.
	 * 
	 * @param		triggerAID	            is an ActionInstanceDescription containing the trigger action name
	 * 										and its properties. 
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *				                        
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 * 
	 **/
	public void sendObligationRequest(KAoSEntityDescription receiverDesc,
									  ActionInstanceDescription obligationAID,
									  ActionInstanceDescription triggerAID)
		throws ActionFailureException, NotLocatableException, TransportFailure
	{
		if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendObligationRequest: received:\n\treceiverDesc " + receiverDesc + "\n\tobligationAID: " 
		                      + obligationAID + "\n\ttriggerAID: " + triggerAID, Logger.INFO); 
//		 Validate the arguments
        if  ( receiverDesc == null || obligationAID == null || triggerAID == null  ) {
            printDebugString("sendObligationRequest: one of the arguments is null! \n\treceiverDesc " 
            		         + receiverDesc + "\n\tobligationAID: " + obligationAID + "\n\ttriggerAID: " 
            		         + triggerAID, Logger.WARN);
            throw new IllegalArgumentException();
        }   
		        
        // Create MethodCallRequestMsg 
        MethodCallRequestMsg request = new MethodCallRequestMsg();
        String requestID = request.getSequenceId();
        
		//Add obligationAID and triggerAID 
		Vector args = new Vector(2);
	    args.addElement(obligationAID);
	    args.addElement(triggerAID);

	    //set method o request action and add args
	    request.setMethodNameArgs("processObligationRequest", args);
	
	    sendRequestMessage(receiverDesc,
		                   request,
		                   false);		
	}
    
    /**
     * 
	 * This method sends an obligation request to an entity identified by its unique ID to execute an obligation action described  
	 * by the given obligation ActionInstanceDescription and the trigger ActionInstanceDescription.
	 *
	 * @param	    receiverID	            is the String that identifies the entity, which is  
	 *							            requested to perform the action.  The receiverID is the one
	 *							            that the action executor has registered with KAoS.
	 *
	 * @param       obligationAID		 	is an ActionInstanceDescription containing the obligation action name
	 * 										and its properties.
	 * 
	 * @param		triggerAID	            is an ActionInstanceDescription containing the trigger action name
	 * 										and its properties. 
	 * 
	 * @throws	    ActionFailureException  containing the reason for failure.  
	 *				                        See ontology to determine specific reasons available for 
	 *				                        a given action
	 *
	 * @throws	    NotLocatableException   when the entity reqested to perform the action cannot be located.
	 *				                        
	 * @throws	    TransportFailure        when the request cannot be sent due to transport failure.
	 * 
	 **/
	public void sendObligationRequest(String receiverID,
									  ActionInstanceDescription obligationAID,
									  ActionInstanceDescription triggerAID)
		throws ActionFailureException, NotLocatableException, TransportFailure
	{
		if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendObligationRequest: received:\n\treceiverID " + receiverID + "\n\tobligationAID: " 
		                      + obligationAID + "\n\ttriggerAID: " + triggerAID, Logger.INFO); 
//		 Validate the arguments
        if  ( receiverID == null || obligationAID == null || triggerAID == null  ) {
            printDebugString("sendObligationRequest: one of the arguments is null! \n\treceiverID " 
            		         + receiverID + "\n\tobligationAID: " + obligationAID + "\n\ttriggerAID: " 
            		         + triggerAID, Logger.WARN);
            throw new IllegalArgumentException();
        }   
		 
        KAoSAgentDescription receiverDesc = getAgentDescription(receiverID);
        
        sendObligationRequest(receiverDesc,
				              obligationAID,
				              triggerAID);
		
	}


    /*****************************************************************************************
    kaos.core.csi.request.ObligationRequestManager interface implementation - end
    *****************************************************************************************/
    
	/*****************************************************************************************
       javax.agent.service.transport.MessageListener interface implementation - begin
    *****************************************************************************************/
    /**
     * Implements the receiveMessage method mandated by the
     * MessageListener interface.  This method prints
     * the incoming TransportMessage to System.out.
     *
     * @see MessageListener
     * @see TransportMessage
     * @see "JAS Specification, Section(s) 3.4.2, 4.7.2"
     */
    public void receiveMessage( TransportMessage tmsg )
    {
        if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("receiveMessage: Incoming transport message:\n"+ tmsg, Logger.INFO);
        
        Object msgObject= MsgUtils.unwrapObjectContent(tmsg);
        if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("receiveMessage: Object content of the message:\n"+ msgObject + "\n", Logger.INFO);

		//String robotID = null;
        Msg msgContent = MsgUtils.unwrapMsgContent(tmsg);
		
		if ( (msgContent instanceof MethodCallRequestMsg) ) {
		    satisfyRequest((MethodCallRequestMsg) msgContent);
	    }
	    else if ( msgContent instanceof MethodCallResultMsg) {
            MethodCallResultMsg content = (MethodCallResultMsg) msgContent;   
            Object result = content.getResult();
            if ( result == null ) {
                result = content.getException();
                printDebugString("receiveMessage: MethodCallResultMsg returning exception:\n"+ result + "\n", Logger.INFO);
                
            }
	        if ( _waiters.containsKey(content.getSequenceId()) ) {
                notifyWaiter(content.getSequenceId(),result);
            } 
	    }
		else {
            printDebugString(".receiveMessage: received message is of unknwon type, not processing: " + msgContent, Logger.WARN);
            return;
        }
    }

    /**
     * Two agents are equal if their respective hash codes
     * are equal.
     *
     * @return true if obj equals this AgentUsecase.
     *
     * @see #hashCode
     * @see "JAS Specification, Section(s) 4.7.2"
     */
    public boolean equals( Object obj )
    {
        if ( obj == this )
            return true;

        if ( !(obj instanceof RequestManagerImpl) )
            return false;

        return this.hashCode() == ((RequestManagerImpl)obj).hashCode();
    }

    /**
     * Determines the hash code for the AgentUsecase.  The AgentUsecase
     * hash code is simply the hash code of the represented agents
     * AgentName.  If the AgentName field has not yet been set,
     * then the hash code is found from <code>super.hashCode()
     * </code>.
     *
     * @return int the hashCode for this AgentUsecase.
     *
     * @see Object#hashCode
     * @see "JAS Specification, Section(s) 4.7.2"
     */
    public int hashCode()
    {
        return _nickname == null ? super.hashCode() : _nickname.hashCode();
    }
	/*****************************************************************************************
       javax.agent.service.transport.MessageListener interface implementation - end
    *****************************************************************************************/
    
    /*****************************************************************************************
       kaos.core.csi.command.FeedbackReceiver interface implementation - begin
    *****************************************************************************************/
    /** 
     * Forward the accept info by the action executor to the feedback receiver associated with the requested
     * action identified by the given actionID.
     *
     * @param time                       Calendar time of acceptance.
     *
     * @param acceptInfo                 Object containing additional (optional) info about the acceptance.
     *
     * @param actionID                   String containing the id of the requested action/information.
     *
     * @param actionExecutorID           String containing the name of the entity performing the action.
     *                       
     */
    public void  onAccepted(Calendar time, Object acceptInfo, String actionID, String actionExecutorID)
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString(".onAccepted: received acceptInfo: " + acceptInfo + "\n For action: " 
                              + actionID + "\n From: " + actionExecutorID, Logger.INFO);
        if ( _actionRequestors.containsKey(actionID) ) {
            FeedbackReceiver receiver = (FeedbackReceiver)_actionRequestors.get(actionID);
            if (_logger.isEnabledFor(Logger.INFO))
                printDebugString(".onAccepted: found a receiver for accepted action from action executor: " + actionExecutorID, Logger.INFO); 
            receiver.onAccepted(time, acceptInfo, actionID, actionExecutorID);
        }
    }
   
    /** 
     * Forward the reason for refusal by the action executor to the feedback receiver associated with the requested
     * action identified by the given actionID.
     *
     * @param time                   Calendar time of refusal.
     *
     * @param reason                 Object containing the reason for refusing the request.
     *
     * @param actionID               String containing the id of the requested action/information.
     *
     * @param actionExecutorID       String containing the name of the entity performing the action.
     */
    public void onRefused(Calendar time, Object reason, String actionID, String actionExecutorID)
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString(".onRefused: received reason for refusal: " + reason + "\n For action: " 
                             + actionID + "\n From: " + actionExecutorID, Logger.INFO);
        if ( _actionRequestors.containsKey(actionID) ) {
            FeedbackReceiver receiver = (FeedbackReceiver)_actionRequestors.get(actionID);
            if (_logger.isEnabledFor(Logger.INFO))
                printDebugString(".onRefused: found a receiver for refused action from robot: " + actionExecutorID, Logger.INFO); 
            _actionRequestors.remove(actionID);
            receiver.onRefused(time, reason, actionID, actionExecutorID);
        }
    }

    /** 
     * Forward the result of the action executed by the given action executor to the feedback receiver associated with the requested
     * action identified by the given actionID.
     *
     * @param time                   Calendar time of information/action completion.
     *
     * @param result                 Object containing the result of the requested action. 
     *
     * @param actionID               String containing the id of the requested action/information.
     *
     * @param actionExecutorID       String containing the name of the robot performing the action.
     *
     */
    public void onCompleted(Calendar time, Object result, String actionID, String actionExecutorID)
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString(".onCompleted: received result: " + result + "\n For action: " 
                             + actionID + "\n From: " + actionExecutorID, Logger.INFO);
        if ( _actionRequestors.containsKey(actionID) ) {
            FeedbackReceiver receiver = (FeedbackReceiver)_actionRequestors.get(actionID);
            if (_logger.isEnabledFor(Logger.INFO))
                printDebugString(".onCompleted: found a receiver for action result: " + result 
                                 + "\n From action executor: " + actionExecutorID, Logger.INFO);
                                 
            receiver.onCompleted(time, result, actionID, actionExecutorID);
            _actionRequestors.remove(actionID);
        }
    }
   
   /** 
    * Forward the action started info from the given action executor to the feedback receiver associated with the requested
    * action identified by the given actionID.
    *
    * @param time                   Calendar time of starting an action.
    *
    * @param startedInfo            Object containing additional (optional) info about the start of the action.
    *
    * @param actionID               String containing the id of the requested action/information.
    *
    * @param actionExecutorID       String containing the name of the entity performing the action.
    *
    *                       
    */
   public void onStarted(Calendar time, Object startedInfo, String actionID, String actionExecutorID)
   {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString(".onStarted: received startedInfo: " + startedInfo + "\n For action: " 
                             + actionID + "\n From: " + actionExecutorID, Logger.INFO);
        if ( _actionRequestors.containsKey(actionID) ) {
            FeedbackReceiver receiver = (FeedbackReceiver)_actionRequestors.get(actionID);
            if (_logger.isEnabledFor(Logger.INFO))
                printDebugString(".onStarted: found a receiver for started action from: " + actionExecutorID, Logger.INFO);
            receiver.onStarted(time, startedInfo, actionID, actionExecutorID);
        }
   }
   
   /** 
    * Forward the action aborted info from the given action executor to the feedback receiver associated with the requested
    * action identified by the given actionID.
    *
    * @param time                       Calendar time of aborting an action.
    *
    * @param reason                     Object containing the reason for aborting an action.
    *
    * @param actionID                   String containing the id of the requested action/information.
    *
    * @param actionExecutorID           String containing the name of the robot performing the action.
    *
    */
   public void onAborted(Calendar time, Object reason, String actionID, String actionExecutorID)
   {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString(".onAborted: received reason: " + reason + "\n For aborting action: " 
                              + actionID + "\n From: " + actionExecutorID, Logger.INFO);
        if ( _actionRequestors.containsKey(actionID) ) {
            FeedbackReceiver receiver = (FeedbackReceiver)_actionRequestors.get(actionID);
            if (_logger.isEnabledFor(Logger.INFO))
                printDebugString(".onAborted: found a receiver for aborted action from: " + actionExecutorID, Logger.INFO);
            receiver.onAborted(time, reason, actionID, actionExecutorID);
             _actionRequestors.remove(actionID);
        }
    }
    /**
     * Receive a status message describing the execution of the requested action. 
     * There maybe several status messages sent during action execution.
     *
     * @param statusMsg     Object containing the result of the requested action/information.
     *
     * @param actionID      String containing the id of the requested action/information.
     *
     * @param robotID       String containing the name of the robot performing the action.
     *
     */                                     
   public void receiveKAoSStatus(String statusMsg, String actionID, String robotID)
   {
   }
   
    /*****************************************************************************************
       kaos.core.csi.command.FeedbackReceiver interface implementation - end
    *****************************************************************************************/
    
    /////////////////////////////   PUBLIC METHODS BEGIN   
   
    /////////////////////////////   PUBLIC METHODS END
    
    /////////////////////////////   PROTECTED METHODS BEGIN
     protected MethodCallResultMsg satisfyRequest (MethodCallRequestMsg requestMsg)
    {
        MethodCallResultMsg resultMsg = null;
        try {
            
            String methodName = requestMsg.getMethodName();
            Vector args = requestMsg.getArgs();
            if ( args != null ) {
                Object argObjects[] = new Object[args.size()];
                for (int i=0; i<args.size(); i++) {
                    argObjects[i] = args.elementAt(i);
                }

                // we have to check this the "hard way" because Class.getMethod(name, args[])
                // does not work if the args are subclasses of the parameter classes

                // find the method matching the method name requested
                Method methods[] = this.getClass().getMethods();
                boolean foundMethod = false;
                for (int i=0; i<methods.length; i++) {
                    if ( methods[i].getName().equals(methodName) ) {
                        // check to see if the parameter classes match the arguments
                        Class parameters[] = methods[i].getParameterTypes();
                        boolean classesMatch = false;
                        if ( parameters.length == argObjects.length) {
                            classesMatch = true ;
                            for (int j=0; j<parameters.length; j++) {
                                if (!parameters[j].isInstance(argObjects[j])) {
                                    classesMatch = false;
                                    break;
                                }
                            }
                        }
                        // if we found the right method, execute it
                        if ( classesMatch ) {
                            foundMethod = true;
                            Object result = methods[i].invoke(this, argObjects);
                            resultMsg = new MethodCallResultMsg(requestMsg.getSequenceId(),
                                                                result);
                            break;
                        }
                    }
                }
                // if we didn't find the method, throw an exception
                if ( ! foundMethod ) {                    
                    throw new NoSuchMethodException(methodName + " args: " + args);
                }
            }
            // args == null, invoke method with no args
            else {
                Object result = this.getClass().getMethod(methodName, (Class[])null).invoke(this, (Object[])null);
                resultMsg = new MethodCallResultMsg(requestMsg.getSequenceId(),
                                                    result);
            }
        }
        catch (Exception xcp) {
            /*
            if ( xcp instanceof java.lang.reflect.InvocationTargetException ) {
                java.lang.reflect.InvocationTargetException inex = (java.lang.reflect.InvocationTargetException) xcp;
                resultMsg = new MethodCallResultMsg(requestMsg.getSequenceId(),
                                                inex.getTargetException());
            }
            else {
                resultMsg = new MethodCallResultMsg(requestMsg.getSequenceId(),
                                                xcp);
            }
            if (_logger.isEnabledFor(Logger.ERROR))
                _logger.logMessage("", xcp, Logger.ERROR);
                */
            _logger.logMessage("", xcp, Logger.ERROR);   
        }

        return resultMsg;
    }
    
    /////////////////////////////   PRIVATE METHODS BEGIN
    private void initialize(String transportName, String nickName)
        throws Exception
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("initialize: received transport name: " + transportName
		                     + " nickName: " + nickName, Logger.INFO);
		_transport = CSIFactory.getTransportSupport();
        _nickname = nickName;
        _transportName = transportName;        
        _myLocator = _transport.createLocator(_nickname, null, _transportName);
        _transport = CSIFactory.getTransportSupport();
        _waiters = new Hashtable();
         
        try {
            _transport.bindMsgListenerToTransport(this,
                                                 _nickname,
                                                 _myLocator,
                                                 _transportName);
            if (_logger.isEnabledFor(Logger.INFO))
			    printDebugString("initialize: successfully bound my locator to transport. Locator:\n " + _myLocator, Logger.INFO);
        }
        catch(Exception e) {
            e.printStackTrace();
            throw e;
        }		
	}    
    
    /**
     */
    private KAoSAgentDescription getAgentDescription(String agentID)
    throws NotLocatableException
    {
        if ( agentID == null ) {
             throw new NotLocatableException("Agent ID of the receiver null! ");
        }
        
        if ( _query == null ) {
            try {
                initQuery();
            }
            catch(Exception e) {
                e.printStackTrace();
                throw new NotLocatableException("Unable to obtain a reference to the Query service! cannot obtain locator for: "
                                                 + agentID);
            }
        }
        
        KAoSAgentDescription desc = null;
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString(".getAgentDescription: received agentID: "  + agentID, Logger.INFO);
		
        try {
			desc = _query.getActor(agentID);
        }
        catch(DirectoryFailure de) {
            de.printStackTrace();
            throw new NotLocatableException("Unable to obtain locator for: "  + agentID + " due to Directory Failure!");
                                                
        }
        catch(SearchException se) {
            se.printStackTrace();
            throw new NotLocatableException("Search for locator for agent: "  + agentID + "  failed!");
        }
        return desc;
    }
    
    /**
     */
    private void initQuery()
        throws Exception
    {
        CSIFactory.setCurrentTransport(_transportName);
        _query = CSIFactory.getQueryRegistration();
    }
    
    /**
     */
    private void validateReceiverDescription(KAoSEntityDescription receiverDesc)
        throws NotLocatableException
    {
        assertNotLocatable( receiverDesc == null, "\n   Receiver's description is null!  Cannot send a request!");
        
        Locator[] receiverLocs = receiverDesc.getLocators();
        assertNotLocatable( receiverLocs.length == 0, "\n   Receiver's Locators are empty!  Description:\n " + receiverDesc);
        
        String receiverName = receiverDesc.getEntityNameAsString();
        assertNotLocatable( receiverName == null,  "\n   Receiver's name is null!  Description:\n " + receiverDesc);
        
    }
    
    /**
     */
    private void validateFeedbackReceiver(FeedbackReceiver feedbackReceiver)
        throws ActionFailureException
    {
        if ( feedbackReceiver == null)
            throw new ActionFailureException("\n   FeedbackReceiver is null!  Will not execute requested action!");
    }
    
    /**
     * @throws NotLocatableException if exp is true.
     */
    public void assertNotLocatable( boolean exp, String msg )
        throws NotLocatableException
    {
        if ( exp )
         throw new NotLocatableException( "Assert not locatable:\n " + msg );
    }	

	/**
	    Use MethodCallRequest to send the specified request for the given action
	    to the action executor, whose KAoSEntityDescription is given.
	    
	    If blockCaller is 'true', then block the calling thread until the action executor replies.
	    
	    Return the unique ID identifying the sent request.
	 */
	private String sendHashMapBasedReqMsg(KAoSEntityDescription receiverDesc,
							              String actionName,
							              HashMap properties,
							              boolean blockCaller)
		throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendHashMapBasedReqMsg: received receiverDesc " + receiverDesc + "\n   action name: " 
		                     + actionName + "\n   action properties: " + properties, Logger.INFO); 
		                     
		// Validate the action
        if  ( actionName == null ) {
            printDebugString("requestAction: requested action is null! \n ", Logger.INFO);
            throw new IllegalArgumentException();
        }   
		        
        // Create MethodCallRequestMsg 
        MethodCallRequestMsg request = new MethodCallRequestMsg();
        String requestID = request.getSequenceId();
        
		//add action, properties and callback to args
		Vector args = new Vector(3);
	    args.addElement(actionName);
	    if ( properties == null ) {
	        args.addElement(new HashMap());
	    }
	    else {
	        args.addElement(properties);
	    }
	    
	    if ( ! blockCaller ) {
	        args.addElement(getActionCallback(requestID,receiverDesc));
	    }
		
	    //set method o request action and add args
	    request.setMethodNameArgs("processHashMapRequest", args);
	
		return sendRequestMessage(receiverDesc,
		                          request,
		                          blockCaller);
    }   
    
    /**
	    Use MethodCallRequest to send the specified request for the given action
	    to the action executor, whose KAoSEntityDescription is given.
	    
	    If blockCaller is 'true', then block the calling thread until the action executor replies.
	    
	    Return the unique ID identifying the sent request.
	 */
	private String sendHashMapBasedReqMsg(KAoSEntityDescription receiverDesc,
							              String actionName,
							              boolean blockCaller)
		throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendHashMapBasedReqMsg: received receiverDesc " + receiverDesc + "\n   action name: " 
		                     + actionName, Logger.INFO); 
		                     
		// Validate the action
        if  ( actionName == null ) {
            printDebugString("requestAction: requested action is null! \n ", Logger.INFO);
            throw new IllegalArgumentException();
        }   
		        
        // Create MethodCallRequestMsg 
        MethodCallRequestMsg request = new MethodCallRequestMsg();
        String requestID = request.getSequenceId();
        
		//add action, properties and callback to args
		Vector args = new Vector(2);
	    args.addElement(actionName);
	    if ( ! blockCaller ) {
	        args.addElement(getActionCallback(requestID,receiverDesc));
	    }
		
	    //set method o request action and add args
	    request.setMethodNameArgs("processHashMapRequest", args);
	
		return sendRequestMessage(receiverDesc,
		                          request,
		                          blockCaller);
    }   
    
    /**
	    Use MethodCallRequest to send the specified request for the given action
	    to the action executor, whose KAoSEntityDescription is given.
	    
	    If blockCaller is 'true', then block the calling thread until the action executor replies.
	    
	    Return the unique ID identifying the sent request.
	 */
	private String sendAIDBasedReqMsg(KAoSEntityDescription receiverDesc,
						              ActionInstanceDescription aid,
						              boolean blockCaller)
		throws ActionFailureException, NotLocatableException, TransportFailure
    {
      	    
		// Validate the action
        if  ( aid == null ) {
            printDebugString("sendAIDBasedReqMsg: requested action is null! \n ", Logger.INFO);
            throw new IllegalArgumentException();
        }  
        
        // Create MethodCallRequestMsg 
        MethodCallRequestMsg request = new MethodCallRequestMsg();
        String requestID = request.getSequenceId();
        
		//add action, properties and callback to args
		Vector args = new Vector(2);
	    args.addElement(aid);
	    
	    if ( ! blockCaller ) {
            args.addElement(getActionCallback(requestID,receiverDesc));	    
        }
		
	    //set method o request action and add args
	    request.setMethodNameArgs("processAIDRequest", args);
	
		return sendRequestMessage(receiverDesc,
		                          request,
		                          blockCaller);
    }   
   
    /**
	    Use MethodCallRequest to send the specified request for the given action
	    to the action executor, whose KAoSEntityDescription is given.
	    
	    If blockCaller is 'true', then block the calling thread until the action executor replies.
	    
	    Return the unique ID identifying the sent request.
	 */
	private String sendOWLBasedReqMsg(KAoSEntityDescription receiverDesc,
						              String owlRequest,
						              boolean blockCaller)
		throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.DEBUG))
            printDebugString("sendOWLBasedReqMsg: received OWL request: " + owlRequest, Logger.DEBUG); 
		                     	    
		// Validate the action
        if  ( owlRequest == null ) {
            printDebugString("sendOWLBasedReqMsg: requested action is null! \n ", Logger.INFO);
            throw new IllegalArgumentException();
        }   
		
        // Create MethodCallRequestMsg 
        MethodCallRequestMsg request = new MethodCallRequestMsg();
        String requestID = request.getSequenceId();
        
		//add action, properties and callback to args
		Vector args = new Vector(2);
	    args.addElement(owlRequest);
	    
	    if ( ! blockCaller ) {
            args.addElement(getActionCallback(requestID,receiverDesc));	    
        }
		
	    //set method o request action and add args
	    request.setMethodNameArgs("processOWLRequest", args);
		
		return sendRequestMessage(receiverDesc,
		                          request,
		                          blockCaller);
    }   
    
    /**
    */
    private RemoteActionStatusCallback getActionCallback(String requestID,
                                                         KAoSEntityDescription receiverDesc)
    {
        RemoteActionCallbackImpl statusCallback = new RemoteActionCallbackImpl(requestID,
																			    _myLocator,
																			    _nickname);
        // This is the id of the entity executing the requested action.
        statusCallback.setSenderID(receiverDesc.getEntityNameAsString()); 
        return statusCallback;
    }
    
    /**
    */
    private String sendRequestMessage(KAoSEntityDescription receiverDesc,
                                      MethodCallRequestMsg request,
                                      boolean blockCaller)
    throws ActionFailureException, NotLocatableException, TransportFailure
    {
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString("sendRequestMessage: received receiverDesc " + receiverDesc + "\n   request: " 
		                     + request, Logger.INFO); 
		                     
        validateReceiverDescription(receiverDesc);
	    
		// Validate the action
        if  ( request == null ) {
            printDebugString("sendRequestMessage: requested action is null! \n ", Logger.INFO);
            throw new IllegalArgumentException();
        }   
		
		String receiverName = receiverDesc.getEntityNameAsString();
        Locator[] receiverLocs = receiverDesc.getLocators();
        
        if ( blockCaller ) {
            OperationWaiter waiter = null;  
            synchronized(_waiters) {
                waiter = new OperationWaiter();
                _waiters.put(request.getSequenceId(), waiter);
            }
        }
         
       // The sendMessageTo() operation may throw 2 exceptions:
       // TransportFailure and NotLocatableException
	    _transport.sendMessageTo(_myLocator,
							     receiverLocs[0],
							     _nickname,
								 receiverName,
                                 FipaAclPerformative.REQUEST,
                                 request,
							     _transportName); 
							     
		if (_logger.isEnabledFor(Logger.INFO))
		    printDebugString("sendRequestMessage: message sent!", Logger.INFO);
		return request.getSequenceId();
    }
    
    /**
     */
    protected void removeWaiter(String id){
        synchronized (_waiters) {
            _waiters.remove(id);
        }
    }
   
    /**
     * This method is called when the result of the action identified by the requestID returns from the action executor. 
     * In addition, the unique requestID was used to associate the Thread invoking the request with its "waiter" object
     * blocking the Thread until the reply came.
     * 
     * Retrieve the waiter object and notify its waiting Thread to deliver action result to the caller.
     */
    private  void notifyWaiter(String requestID, Object result)
    {
        OperationWaiter waiter = (OperationWaiter) _waiters.get(requestID);
        // check if waiter still exists -- if the waiter's thread has been interrupted, then the waiter  has been removed
        if ( waiter != null) {
            waiter.doNotify(result);
        }
    }
    
    private Object getResult(String requestID)
        throws ActionFailureException
    {
        OperationWaiter waiter = (OperationWaiter) _waiters.get(requestID);
        Object result = waiter.getResult();
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString(".getResult: checking for result: " + result + " in thread: " 
                              + Thread.currentThread(), Logger.INFO);
        
        // check if the result has come already, sometimes this is true, before the waiter object is created
        if ( result != null) {
            removeWaiter(requestID);
            if ( result instanceof ActionFailureException ) throw (ActionFailureException) result;
            return result;
        }
        
        // If the result has not come back yet, then block until the action executor returns the result.
        if (_logger.isEnabledFor(Logger.INFO))
            printDebugString(".getResult: result has not come back, will block: " 
                              + Thread.currentThread(), Logger.INFO);
        
        try {
            waiter.doWait();
        }
        catch (InterruptedException e) {
            removeWaiter(requestID);
            throw new ActionFailureException(" While waiting for the result of request: " + requestID 
                                             + " thread interruption has occurred!!");
        }
        
        result = waiter.getResult();
        removeWaiter(requestID);
        
        if ( result instanceof ActionFailureException ) 
            throw (ActionFailureException) result;
            
        return result; 
    }
    //////////////////////////// PRIVATE METHODS END
	
    // Utility methods ///////////////////////////////////////////
      
	protected void printDebugString (String msg, int loggerDisplayLevel)
    {
   	    _logger.logMessage(msg, loggerDisplayLevel);
    }

	///////////////////////////////////////////////////////////////////////////////
    public static void main(String [] args)
      throws Exception      
    {
        String transport = null;
        String name = null;
        if (args.length > 0) {
	        int i = 0;
	        while (i < args.length) {
	            if (args[i].equalsIgnoreCase ("-name")) {
	                if ((i+1) < args.length) {
    	                name = args[i+1];
    	                i++;
    	            }
    	            else {
    	                System.out.println ("Must specify name with the -name keyword.");
    	                System.exit(0);
    	            }
	            }
	            else if (args[i].equalsIgnoreCase ("-transport")) {
	                if ((i+1) < args.length) {
    	                transport = args[i+1];
    	                i++;
    	            }
    	            else {
    	                System.out.println ("Must specify transport with the -transport keyword.  Valid transpors: corba, grid, tcp.");
    	                System.exit(0);
    	            }
	            }
	            i++;
	        }
	    }

        System.out.println ("\n     Requested name:      " + name);
        System.out.println ("\n     Requested transport: " + transport);
        
        RequestManager mgr = RequestManagerImpl.getInstance(transport);
        
        System.out.println ("\n Got mgr instance" + mgr + "\n\n");
		
        synchronized ( Thread.currentThread() ) {
            Thread.currentThread().wait();

        }
   }
   
   //////////////////////////////////////   STATIC CLASS DEFINITION
   /**
     * This class is used as a helper class to block Threads invoking the synchronous 'requestAction' methods.  
     * After the action request is sent to the action executor, the calling Thread has to wait until
     * the action executor returns a reply message containing the action result.
     * 
     */
    protected static class OperationWaiter {
        private boolean     _operationCompleted;
        private Object      _result = null;
        
        public OperationWaiter(){
            _operationCompleted = false;
        }
    
        synchronized void doWait()
            throws InterruptedException 
        {
            if (_logger.isEnabledFor(Logger.INFO))
                printDebugString("OperationWaiter.doWait: entered! _operationCompleted: " + _operationCompleted, Logger.INFO);
            try {
                while ( ! _operationCompleted ){
                    wait();
                }
            }
            catch(InterruptedException e){
                if ( ! _operationCompleted ) {
                    _operationCompleted = true;
                    throw e;
                }
                else{
                }
            }
            if (_logger.isEnabledFor(Logger.INFO))
                printDebugString("OperationWaiter.doWait: exiting! \n      _operationCompleted: " + _operationCompleted
                                 + "\n     result: " + _result, Logger.INFO);
            
        }
        
        synchronized void doNotify(Object result)
        {
            if (_logger.isEnabledFor(Logger.INFO))
                printDebugString(".doNotify: got result: " + result, Logger.INFO);
            if ( _operationCompleted ) {
                return;
            }
            else {
                if (_logger.isEnabledFor(Logger.INFO))
                    printDebugString(".doNotify: just before notifying...", Logger.INFO);
                
                _operationCompleted = true;
                _result = result;
                notify();
            }
        }
        
        Object getResult(){
            return _result;
       }
       
       protected void printDebugString (String msg, int loggerDisplayLevel) {
        _logger.logMessage("**** " + this.getClass().getName() + msg, loggerDisplayLevel);
        }
        private Logger _logger = Logger.getLogger(this);
       
    }
}

