/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2012/03/05 04:43:46 $
 */

package kaos.core.csi.transport;

import kaos.core.KAoSLocator;
import kaos.core.service.KAoSServiceRoot;
import kaos.core.service.boot.KAoSGuardBootstrapper;
import kaos.core.service.directory.KAoSAgentDescription;
import kaos.core.util.KAoSConstants;
import kaos.core.util.Logger;
import kaos.core.util.MsgUtils;
import ri.JasBeanImpl;
import ri.acr.FipaAclPerformative;

import javax.agent.JasBean;
import javax.agent.Locator;
import javax.agent.TransportMessage;
import javax.agent.service.transport.*;
import java.io.Serializable;
import java.util.List;

/**
 * This class defines the operations to bind an entity, whose KAoSAgentDescription or Locator is given to the specified transport
 * and to send messages via a given transport.
 * NOTE: This class is NOT an implementation of a specific transport, it just provides access to ANY transport
 * for binding and message sends, simplifying some of the complexities of interfacing with a transport.
 * 
 * @author KAoS Team
 * $Revision: 1.12 $
 */
public class TransportSupportImpl implements TransportSupport
{
    private static final TransportSupportImpl __INSTANCE = new TransportSupportImpl();
    private Logger                            _logger    = Logger.getLogger(this);

    /*(
     * Retrieve an instance of Registration.
     * 
     * @return a singleton, Registration.
     */
    public static TransportSupport getInstance()
    {
        return __INSTANCE;
    }

    private TransportSupportImpl()
    {
    }

    /**
     * Create a Locator using the specified entity name and transport name. If there is ay other additional info required by the given transport, pass it in the given JasBean.
     * 
     * @param name String containing the name to be used by the given transport, or null, if the transport does not require a name for its registry/lookup service.
     * @param agentInitInfo JasBean encapsulating additional transport info, or null.
     * @param transportName String containing the name of the requested transport.
     * @return Locator that can be used to bind an entity to the given transport.
     */
    public Locator createLocator(String name, JasBean agentInitInfo, String transportName)
    {
        if (agentInitInfo == null)
            agentInitInfo = new JasBeanImpl();

        if (!agentInitInfo.containsKey(KAoSConstants.NAME))
        	agentInitInfo.set(KAoSConstants.NAME, name);

        KAoSServiceRoot sr = getServiceRoot(transportName);

        if (sr == null)
            return null;

        MessageTransportService mts = sr.getPreferredMessageTransportService();
        Locator locator = null;

        try
        {
            locator = mts.newLocalLocator(agentInitInfo);
        }
        catch (Exception e)
        {
            logMessage(".createLocator: Unable to obtain Locator for agent:  " + name, e, Logger.ERROR);

            return null;
        }

        logMessage(".createLocator: Obtained Agent Locator: " + locator + "\n", Logger.INFO);

        return locator;
    }

    /**
     * Bind the specified MessageListener, whose KAoSAgentDescription is given to the given transport. Depending on the transport, the binding process will register the entity with
     * the transport's lookup service, or create a server socket to accept client sockets, or what has to be done in order to communicate with other entities.
     * 
     * @param listener MessageListener to receive messages using the Locator, retrieved from the given KAoSAgentDescription.
     * @param listenerName String containing the name of the receiver.
     * @param listenerDesc KAoSAgentDescription to be used to retrieve the Locator of the MessageListener and the name of the transport to bind to.
     * 
     * @throws NotLocatableException if locator is not a local locator.
     * @throws TransportFailure if the binding process fails
     */
    public void bindMsgListenerToTransport(MessageListener listener, String listenerName, KAoSAgentDescription listenerDesc) throws NotLocatableException, TransportFailure
    {
        Locator receiverLocator = listenerDesc.getLocators()[0];

        if ( ! (receiverLocator instanceof KAoSLocator))
            throw new NotLocatableException("Locator of MessageListener: " + listenerName + " not instance of KAoSLocator!  "
                    + "Unable to bind to any transport. Listener described by:\n " + listenerDesc);

        KAoSLocator loc = (KAoSLocator)receiverLocator;
        String transportName = loc.getTransportName();

        bindMsgListenerToTransport(listener, listenerName, receiverLocator, transportName);
    }

    /**
     * Bind the specified MessageListener, whose Locator is given to the given transport. Depending on the transport, the binding process will register the entity with the
     * transport's lookup service, or create a server socket to accept client sockets, or what has to be done in order to communicate with other entities.
     * 
     * @param listener MessageListener to receive messages via the specified transport using the given Locator.
     * @param listenerName String containing the name of the receiver. 
     * @param locator Locator to be used to bind to the transport. 
     * @param transportName String indicating which transport to use. 
     * @throws NotLocatableException if locator is not a local locator.
     * 
     * @throws TransportFailure if the binding process fails
     */
    public void bindMsgListenerToTransport(MessageListener listener, String listenerName, Locator locator, String transportName) throws NotLocatableException, TransportFailure
    {
        KAoSServiceRoot sr = getServiceRoot(transportName);

        if (sr == null)
            throw new TransportFailure("Unable to obtain transport for: " + transportName);

        MessageTransportService mts = sr.getPreferredMessageTransportService();
        MessageReceiver rcvr;

        if (listenerName == null)
            rcvr = mts.newMessageReceiver();
        else
        {
            JasBean agentInitInfo = new JasBeanImpl();
            agentInitInfo.set(KAoSConstants.NAME, listenerName);
            rcvr = mts.newMessageReceiver(agentInitInfo);
        }

        // all the messaging Callback
        rcvr.addMessageListener(listener);
        rcvr.bindToLocalLocator(locator);
    }

    /**
     * Bind an entity, whose name and Locator are given to the transport, whose name is specified. As a result of binding, the entity will receive the MessageReceiver in order to
     * get messages from it (MessageReceiver.getMessage()), as opposed to listening for messages via the MessageListener interface.
     * 
     * @param receiverName String containing the name of the entity binding to transport.
     * @param transportName String indicating which transport to use
     * 
     * @return a MessageReceiver, from which to retrieve messages.
     * 
     * @throws NotLocatableException if locator is not a local locator.
     * 
     * @throws TransportFailure if the binding process fails
     */
    public MessageReceiver bindToTransport(String receiverName, Locator receiverLocator, String transportName) throws NotLocatableException, TransportFailure
    {
        KAoSServiceRoot sr = getServiceRoot(transportName);

        if (sr == null)
            throw new TransportFailure("Unable to obtain transport for: " + transportName);

        MessageTransportService mts = sr.getPreferredMessageTransportService();
        MessageReceiver rcvr;

        if (receiverName == null)
            rcvr = mts.newMessageReceiver();
        else
        {
            JasBean agentInitInfo = new JasBeanImpl();
            agentInitInfo.set(KAoSConstants.NAME, receiverName);
            rcvr = mts.newMessageReceiver(agentInitInfo);
        }

        rcvr.bindToLocalLocator(receiverLocator);
        return rcvr;
    }

    /**
     * Send the given message content to the receiver, from whom the receivedMsg was received. This method is defined to be used for entities, who are replying to the receivedMsg.
     * 
     * @param senderDesc KAoSAgentDescription of the sender. Based on the description, the sender Locator and name will be determined.
     * @param receivedMsg TransportMessage that was received by the sender. Based on that the receiver Locator and name will be determined.
     * @param content Serializable content of the message to be sent out.
     * 
     * @throws NotLocatableException if the receiver cannot be located.
     * @throws TransportFailure if transport fails during message send.
     */
    public void sendMessageTo(KAoSAgentDescription senderDesc, TransportMessage receivedMsg, Serializable content) throws NotLocatableException, TransportFailure
    {
        sendMessageTo(senderDesc.getLocators()[0], receivedMsg.getSender(), senderDesc.getAgentNickname(), MsgUtils.getSender(receivedMsg), FipaAclPerformative.INFORM, content);
    }

    /**
     * Send the given message content to the receiver, whose KAoSAgentDescription is given. This method is defined to be used for entities looking-up other entities in the Directory
     * Service in order to send messages.
     * 
     * @param senderDesc KAoSAgentDescription of the sender. Based on that the sender Locator and name will be determined.
     * @param receiverDesc KAoSAgentDescription of the receiver. Based on that the receiver Locator and name will be determined.
     * @param content Serializable content of the message to be sent out.
     * 
     * @throws NotLocatableException if the receiver cannot be located.
     * @throws TransportFailure if transport fails during message send.
     */
    public void sendMessageTo(KAoSAgentDescription senderDesc, KAoSAgentDescription receiverDesc, Serializable content) throws NotLocatableException, TransportFailure
    {
        sendMessageTo(senderDesc.getLocators()[0], receiverDesc.getLocators()[0], senderDesc.getAgentNickname(), receiverDesc.getAgentNickname(), FipaAclPerformative.REQUEST,
                content);
    }

    /**
     * Send the given message content to the receiver, whose Locator is given.
     * 
     * @param senderLocator Locator of the sender.
     * @param receiverLocator Locator of the receiver.
     * @param senderName String containing the name of the sender.
     * @param receiverName String containing the name of the receiver.
     * @param fipaPerformative String containing the FIPA performative describing the message intent.
     * @param content Serializable containing the message content.
     * @param transportName String containing the name of the transport to be used.
     * 
     * @throws NotLocatableException if the receiver cannot be located.
     * @throws TransportFailure if transport fails during message send.
     */
    public void sendMessageTo(Locator senderLocator, Locator receiverLocator, String senderName, String receiverName, String fipaPerformative, Serializable content,
            String transportName) throws NotLocatableException, TransportFailure
    {
        if (transportName == null)
            throw new NotLocatableException(" Transport name is null, message: " + content + "\n cannot be sent out!");

        KAoSServiceRoot sr = getServiceRoot(transportName);

        if (sr == null)
            throw new TransportFailure("   *** Unable to obtain KAoSServiceRoot! Cannot send messages out!");

        MessageTransportService mts = sr.getPreferredMessageTransportService();

        if (mts.isSupported(senderLocator) && mts.isSupported(receiverLocator))
        {
            if (senderName == null)
                senderName = "UNKNOWN";

            if (receiverName == null)
                receiverName = "UNKNOWN";

            if (fipaPerformative == null)
                fipaPerformative = FipaAclPerformative.INFORM;

            TransportMessage messageToSend = MsgUtils.wrapObject(senderLocator, receiverLocator, senderName, receiverName, fipaPerformative, content);
            mts.sendMessage(messageToSend);
        }
        else
            throw new NotLocatableException(" Sender and/or Receiver Locators not supported by the Message Transport System!");
    }

    /**
     * Send the given message content to the receiver, whose Locator is given.
     * 
     * @param senderLocator Locator of the sender.
     * @param receiverLocator Locator of the receiver.
     * @param senderName String containing the name of the sender.
     * @param receiverName String containing the name of the receiver.
     * @param fipaPerformative String containing the FIPA performative describing the message intent.
     * @param content Serializable containing the message content.
     * 
     * @throws NotLocatableException if the receiver cannot be located.
     * @throws TransportFailure if transport fails during message send.
     */
    public void sendMessageTo(Locator senderLocator, Locator receiverLocator, String senderName, String receiverName, String fipaPerformative, Serializable content)
            throws NotLocatableException, TransportFailure
    {
        if (!(receiverLocator instanceof KAoSLocator))
            throw new NotLocatableException("Receiver Locator not instance of KAoSLocator!  Unable to send a message to: " + "\n   Receiver name: receiverName"
                    + "\n   Receiver locator: " + receiverLocator + "\n   Message to send: " + content);

        KAoSLocator loc = (KAoSLocator)receiverLocator;
        String transportName = loc.getTransportName();

        sendMessageTo(senderLocator, receiverLocator, senderName, receiverName, fipaPerformative, content, transportName);
    }

    /**
     * Return the names of the currently supported transports.
     * 
     * @return a List containing Strings representing names of supported transports.
     */
    public List getSupportedTransportsNames()
    {
        return KAoSGuardBootstrapper.supportedTransports();
    }

    /**
     * Answer true if the given transport is supported, otherwise, answer false.
     * 
     * @param name String containing the name of the transport in question.
     * 
     * @return a boolean indicating if the given transport is supported.
     */
    public boolean isTransportSupported(String name)
    {
        return KAoSGuardBootstrapper.isTransportSupported(name);
    }

    /**
     * Return KAoSServiceRoot for the specified transportName.
     * 
     * @param transportName String containing the name of the transport to create a KAoSServiceRoot for.
     * @return a KAoSServiceRoot specific for the transportName.
     */
    public KAoSServiceRoot getServiceRoot(String transportName)
    {
        KAoSServiceRoot sr = null;

        try
        {
            sr = KAoSGuardBootstrapper.getKAoSServiceRootForTransport(transportName);
        }
        catch (Exception e)
        {
            logMessage("\n  *** Unable to obtain KAoSServiceRoot for transport:  " + transportName + "!", e, Logger.ERROR);

            return null;
        }

        return sr;
    }

    // ----------- Utility methods

    protected void logMessage(String msg, int logLevel)
    {
        if (_logger.isEnabledFor(logLevel))
            _logger.logMessage(msg, logLevel);
    }

    protected void logMessage(String msg, Exception e, int logLevel)
    {
        if (_logger.isEnabledFor(logLevel))
            _logger.logMessage(msg, e, logLevel);
    }
}