/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2012/03/21 14:22:35 $
 */
package kaos.core.csi.usecase;

import kaos.core.csi.ActionInstanceDescription;
import kaos.core.csi.ActionInstanceDescriptionImpl;
import kaos.core.csi.CSIFactory;
import kaos.core.csi.policy.KAoSSecurityException;
import kaos.core.csi.policy.PolicyChecking;
import kaos.core.csi.query.QueryRegistration;
import kaos.core.csi.registration.Registration;
import kaos.core.csi.transport.TransportSupport;
import kaos.core.service.directory.KAoSAgentDescription;
import kaos.core.util.Logger;
import kaos.core.util.MsgUtils;
import kaos.ontology.vocabulary.ActionConcepts;
import kaos.ontology.vocabulary.GroupConcepts;

import javax.agent.TransportMessage;
import javax.agent.service.ServiceFailure;
import javax.agent.service.directory.AlreadyRegisteredException;
import javax.agent.service.directory.DirectoryFailure;
import javax.agent.service.transport.MessageListener;
import javax.agent.service.transport.NotLocatableException;
import javax.agent.service.transport.TransportFailure;
import javax.agent.service.directory.SearchException;

import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * $Revision: 1.7 $
 */
public class SimpleGuardedSender
    implements  MessageListener
    
 {   
    /**
        This class llustrates an implementation of a SimpleGuardedSender: an entity registered with KAoS via its Guard
        and capable of sending messages to other entities registered in the KAoS Directory Service.  Each message send is authorized according to the current set of policies
        stored by the Guard.
        
        Configuration data includes: nickname, the name of the entity to send a message to, names of kaos domains it will 
        register in, name of a transport to use for communication within KAoS.
        
        Bootstrapping steps:
	        - obtains a reference to CSI Registration
	        - binds itself to CSI Transport 
	        - asks CSI Registration  to create a KAoSAgentDescription for  itself
	        - asks CSI Registration to register with KAoS

        Running
	        - looks up another agent, whose name is given
	        - sends a message to it
    */
    protected KAoSAgentDescription      _agentDescription = null;
    protected TransportSupport          _transportSupport;
    protected QueryRegistration         _query; 
    protected PolicyChecking			_policyChecking;
    private Hashtable                   _receiverDescriptions;
    
    public SimpleGuardedSender( )
    {
    }

    //////////////////////////////////////////////////////////////

    public void initialize(String name, String transportName, Vector domainNames)
    throws Exception
    {
    	// Set the message transport (e.g. tcp) and get a kaos registration object
        Registration registration = null;
        try {
            CSIFactory.setCurrentTransport(transportName);
            registration = CSIFactory.getRegistration();
        }
        catch(Exception e){
            e.printStackTrace();
            System.out.println ("  **** Unable to obtain CSI services: Registration. Exiting...");
            return;
        }
        
        // register the domains to which the agent belongs before registering the agent
        // it's ok if the domain is already registered
        Enumeration domainsEnum = domainNames.elements();
        while ( domainsEnum.hasMoreElements() ) {
            String domainName  = (String) domainsEnum.nextElement();
            System.out.println("\n     <>  Will register domain: " + domainName + "\n");
            
            try {
                registration.registerDomain(domainName);
            }
            catch(DirectoryFailure de) {
                System.out.println("\n   Due to Directory Failure unable to register domain: " + domainName);
                de.printStackTrace();
                return;
            }
            catch(AlreadyRegisteredException ae) {
                System.out.println("    The following domain has already been registered: " + domainName);
            }
        }
        
        // Ask KAoS registration to create a KAoSAgentDescription for the sender.
        _agentDescription = registration.createKAoSAgentDescription(name,       // use nickname as guid
        															name,       // nickname
                                                                    domainNames,  // domains to register in
                                                                    transportName,
                                                                    null);      // no extra transport info necessary
                                                                    
        // Bind this agent to the KAoS transport (necessary to use kaos messaging)
        _transportSupport = CSIFactory.getTransportSupport();
        _transportSupport.bindMsgListenerToTransport(this, name, _agentDescription);
        
        registration.registerEntity(_agentDescription,
                                    true,       // guarded
                                    false);     // replace registration 
        
        // Obtain a reference to the KAoS registration query service, used later by sendMessageTo() to lookup the receiver agent
        _query = CSIFactory.getQueryRegistration();  
        // Obtain a reference to the KAoS policy checking service, used later by authorizeCommunicationAction() to check policies
        _policyChecking = CSIFactory.getPolicyChecking();
        
        // Initialize a variable to cache the agentDescription of the reciever(s) that are returned by the query service
        _receiverDescriptions = new Hashtable(); 
       
    }
    
    //////////////////////////////////////////////////////////////
    // Implementation of the MessageListener interface

    /**
     * Implements the receiveMessage method mandated by the
     * MessageListener interface.  This method prints
     * the incoming TransportMessage to System.out.
     *
     */
    public void receiveMessage( TransportMessage tmsg )
    {
        System.out.println( "\n   *** SimpleGuardedSender: received transport message:\n" + tmsg);
        Object msgContent = MsgUtils.unwrapObjectContent(tmsg);

        if ( msgContent == null ) {
            System.out.println("\n   *** SimpleGuardedSender: \n receiveMessage: could not retrieve the content from: " + tmsg);
            return;
        }
                    
       String senderName = MsgUtils.getSender(tmsg);
       
       System.out.println("\n   *** SimpleGuardedSender: received the following content from " + senderName + ": " + msgContent);
    }

    /**
     * Two agents are equal if their respective hash codes
     * are equal.
     *
     * @return true if obj equals this SimpleGuardedSender.
     */
    public boolean equals( Object obj )
    {
        if ( obj == this )
            return true;

        if ( !(obj instanceof SimpleGuardedSender) )
            return false;

        return this.hashCode() == ((SimpleGuardedSender)obj).hashCode();
    }

    /**
     * Determines the hash code for the SimpleGuardedSender.  The SimpleGuardedSender
     * hash code is simply the hash code of the represented agents
     * AgentName.  If the AgentName field has not yet been set,
     * then the hash code is found from <code>super.hashCode()
     * </code>.
     *
     * @return int the hashCode for this SimpleGuardedSender.
     *
     */
    public int hashCode()
    {
        return _agentDescription.getEntityNameAsString() == null ? super.hashCode() : _agentDescription.getEntityNameAsString().hashCode();
    }

    //////////////////////////////////////////////////////////////
    
    public void sendMessageTo(String receiverName)
        throws SearchException, NotLocatableException, KAoSSecurityException, NullPointerException, ServiceFailure
    {   	
        // First, lookup the receiver with the given receiverName.
        KAoSAgentDescription receiverDesc = null;
        if ( _receiverDescriptions.containsKey(receiverName)) {
        	// use the cached receiver agentDescription
            receiverDesc = (KAoSAgentDescription) _receiverDescriptions.get(receiverName);
        }
        else {
        	// not in the cache of receiver agentDescriptions, so query KAoS registration for it
            receiverDesc = _query.getActor(receiverName);
            // cache the result
            _receiverDescriptions.put(receiverName, receiverDesc);
        }
        
        System.out.println("\n   *** SimpleGuardedSender: Requested description for: " +  receiverName 
                          +"\n                  Received: \n\n    " + receiverDesc);
        if ( receiverDesc == null ) return;     
        
	    authorizeCommunicationAction(_agentDescription.getEntityNameAsString(), receiverDesc.getEntityNameAsString(), null, null, "tcp");

        
        // Then ask the transport support service to send a messager to the receiver.
        _transportSupport.sendMessageTo(_agentDescription,
                                        receiverDesc,
                                        "\n\n   Hello World!");
       
       System.out.println("\n   *** SimpleGuardedSender: Sent the following message to " +  receiverName + ": \n\n   Hello World!");
    }
    
    /**
     * Authorize a communication action described by properties expressed by the arguments to this method.
     * Authorization is based on the current set of policies stored in the Guard.
     * All arguments, except for the dsTransportName, contain the values of properties with the same name
     * as the name of the given argument.  Those names are defined in the ontology for CommunicationAction.
     *
     * @param performedBy           String containing the name of the actor performing the CommunicationAction.
     *
     * @param hasDestination        String containing the name of the entity receiving the CommunicationAction.
     *
     * @param carriesMessage        String containing the content of the CommunicationAction.
     *
     * @param hasPartner            String containing the name of the partner in the CommunicationAction.
     *
     * @param dsTransportName       String containing the name of the transport used by KAoS Policy service
     *                              and other KAoS services to communicate with each other.
     *
     * @throws KAoSSecurityException    when the CommunicationAction is forbidden by the current policies.
     *
     * @throws NullPointerException     when performedBy is null.  
     *
     * @throws ServiceFailure           when the Policy service cannot be obtained, or fails durin action authorization.
     */
    
    public void authorizeCommunicationAction(String performedBy, 
                                             String hasDestination,
                                             String carriesMessage,
                                             String hasPartner,
                                             String dsTransportName)
    throws KAoSSecurityException, NullPointerException, ServiceFailure
    {
        if ( performedBy == null ){
            System.out.println("authorizeCommunicationAction: actorID null, unable to authorize for:\n destination:\n "
   			                       + hasDestination + "\n carriesMessage: " + carriesMessage + "\n hasPartner: " 
   			                       + hasPartner);
   			throw new NullPointerException();
        }
        
        HashMap actionProperties = new HashMap();
        
        if ( hasDestination != null ) {
            actionProperties.put(ActionConcepts.hasDestination(), hasDestination);
        }
        
        if ( carriesMessage != null ) {
            actionProperties.put(ActionConcepts.carriesMessage(), carriesMessage);
        }
        
        if ( hasPartner != null ) {
            actionProperties.put(GroupConcepts.hasPartner(), hasPartner);
        }
        
        try {
        	ActionInstanceDescription aid = new ActionInstanceDescriptionImpl(null,
        																	 ActionConcepts.CommunicationAction(),
        																	 performedBy,
        			   												         actionProperties);
   	
            _policyChecking.checkPermission(aid,null); 
        }
        catch(KAoSSecurityException kse){ 
            kse.printStackTrace();
            throw new KAoSSecurityException("\n\n  !!!! SECURITY VIOLATION !!!!\n\n  CommunicationAction NOT ALLOWED BETWEEN: \n\n  SENDER:   "
                                             + performedBy + "\n  RECEIVER: " + hasDestination + "\n\n");
        }
        catch(NullPointerException npe){
            throw npe;
        }
        catch(ServiceFailure sfe){ 
            throw sfe;
        }
        catch(URISyntaxException urie){ 
        	urie.printStackTrace();
            throw new ServiceFailure();
        }
    }

   //////////////////////////////////////////////////////////////
   // Entry point...

   public static void main(String [] args)
      throws Exception      
    {
        String sender = null;
        String transport = null;
        String receiver = null;
        Vector domainNames = null;
        if (args.length > 0) {
	        int i = 0;
	        while (i < args.length) {
	            if (args[i].equalsIgnoreCase ("-sender")) {
	                if ((i+1) < args.length) {
    	                sender = args[i+1];
    	                i++;
    	            }
    	            else {
    	                System.out.println ("Must specify Entity Name with -sender");
    	                System.exit(0);
    	            }
	            }
	            else if (args[i].equalsIgnoreCase ("-receiver")) {
	                if ((i+1) < args.length) {
    	                receiver = args[i+1];
    	                i++;
    	            }
    	            else {
    	                System.out.println ("Must specify  transport with -transport.  Valid transports: corba, grid, tcp.");
    	                System.exit(0);
    	            }
	            }
	            else if (args[i].equalsIgnoreCase ("-transport")) {
	                if ((i+1) < args.length) {
    	                transport = args[i+1];
    	                i++;
    	            }
    	            else {
    	                System.out.println ("Must specify  transport with -transport.  Valid transports: corba, grid, tcp.");
    	                System.exit(0);
    	            }
	            }
	            else if (args[i].equalsIgnoreCase ("-domains")) {
                    if ((i+1) < args.length) {
                        String domainsStr = args[i+1];
                        domainNames = new Vector();
                        StringTokenizer tokenizer = new StringTokenizer(domainsStr, ",");
                        while (tokenizer.hasMoreTokens()) {
                            domainNames.add(tokenizer.nextToken());
                        }
                    }
                }// else if
                 i++;
	        }// while
	    }
	    
        if ( sender == null) {
            System.out.println ("\n     *** Must specify Sender Name with -sender\n");
            return;
        }
        
        if ( receiver == null) {
            System.out.println ("\n     *** Must specify Receiver Name with -receiver\n");
            return;
        }
        
        if ( transport == null) {
            System.out.println ("\n     *** Must specify  transport with -transport.  Valid transports: corba, grid, tcp.");
            return;
        }
        
        if ( domainNames == null ) {
            domainNames = new Vector();
            domainNames.add("SenderDomain");
            System.out.println (" <>    Warning: No Domain Names specified with -domains, using default: " + domainNames);
        }
        
        
        System.out.println ("\n     <>  Sender name                 " + sender);
        System.out.println ("\n     <>  Receiver name               " + receiver);
	    System.out.println ("\n     <>  Requested transport:        " + transport);
	    System.out.println ("\n     <>  Requested domains:          " + domainNames + "\n");
	    
        
        SimpleGuardedSender usecase = new SimpleGuardedSender();
        
        try {
            usecase.initialize(sender, transport, domainNames);
        }
        catch(Exception e) {
            e.printStackTrace();
            return;
        }
       
       while (true) {
            usecase.sendMessageTo(receiver);
            try {
                Thread.currentThread().sleep(5000);
            }
            catch(Exception e){
                e.printStackTrace();
                break;
            }
       }
       
        synchronized ( Thread.currentThread() ) {
         Thread.currentThread().wait();

      }

   }
}

