/**
 * Last Modified by: $Author: mjohnson $
 * On: $Date: 2010/05/23 02:51:19 $
 */
package kaos.core.csi.usecase;

import kaos.core.csi.CSIFactory;
import kaos.core.csi.registration.Registration;
import kaos.core.csi.transport.TransportSupport;
import kaos.core.service.directory.DirectoryEvent;
import kaos.core.service.directory.DirectoryEventListener;
import kaos.core.service.directory.KAoSAgentDescription;
import kaos.core.util.DirectoryEventMsg;
import kaos.core.util.MsgUtils;

import javax.agent.TransportMessage;
import javax.agent.service.directory.DirectoryFailure;
import javax.agent.service.directory.AlreadyRegisteredException;
import javax.agent.service.transport.MessageListener;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.Vector;

/**
 * $Revision: 1.7 $
 */
public class SimpleReceiver
    implements  MessageListener, DirectoryEventListener

 {
    /**
        This class illustrates an implementation of a SimpleReceiver: an entity capable of receiving
        messages.

        Configuration data includes: nickname, names of kaos domains it will register in, name of a transport
        to use for communication within KAoS.

        Bootstrapping steps:
	        - obtains a reference to CSI Registration
	        - binds itself to CSI Transport
	        - asks CSI Registration to register with KAoS using given KAoSAgentDescription

        Running
	        - waits for messages
	        - acknowledges received messages by sending a reply to sender

    */
    protected KAoSAgentDescription      _agentDescription = null;
    protected TransportSupport          _transportSupport;

    public SimpleReceiver( )
    {
    }

    //////////////////////////////////////////////////////////////

    public void initialize(String name, KAoSAgentDescription agentDescription, String transportName)
        throws Exception
    {
        CSIFactory.setCurrentTransport(transportName);
        Registration registration = CSIFactory.getRegistration();

        _agentDescription = agentDescription;

        _transportSupport = CSIFactory.getTransportSupport();

        // Ask TransportSupport to bind this MessageListener to the given transport, so the receiver
        // will be able to receive messages.
        _transportSupport.bindMsgListenerToTransport(this, name, _agentDescription);
        registration.addRemoteEventListener(_agentDescription.getLocators()[0]);

        registration.registerEntity(_agentDescription,
                                    true,       // guarded
                                    false);     // replace registration

        System.out.println("\n   *** SimpleReceiver: registered in KAoS using KAoSAgentDescription:\n " + _agentDescription);

    }

    //////////////////////////////////////////////////////////////
    // Implementation of the MessageListener interface

    /**
     * Implements the receiveMessage method mandated by the
     * MessageListener interface.
     *
     * Sends an acknowledgement back to the receiver.
     */
    public void receiveMessage(TransportMessage tmsg)
    {
        System.out.println( "\n   *** SimpleReceiver: received transport message:\n"+ tmsg);
        Object msgContent = MsgUtils.unwrapObjectContent(tmsg);

        if ( msgContent == null ) {
            System.out.println("\n   *** SimpleReceiver: \n receiveMessage: could not retrieve the content from: " + tmsg);
            return;
        }
        if (msgContent instanceof DirectoryEventMsg)
        {
        	DirectoryEventMsg msg = (DirectoryEventMsg)msgContent;
        	receiveDirectoryEvent(msg.getDirectoryEvent());
        }
        String senderName = MsgUtils.getSender(tmsg);

        System.out.println("\n   *** SimpleReceiver: received the following content from " + senderName + ": " + msgContent);
        String ack =  "\n\n   Acknowledgement from " + _agentDescription.getAgentNickname()
                      + " that the message:\n " + msgContent + "\n\n   has been received.";
        try {
            _transportSupport.sendMessageTo(_agentDescription,
                                            tmsg,
                                            ack);
        }
        catch(Exception e){
            e.printStackTrace();
        }

        System.out.println("\n   *** SimpleReceiver: sent the following acknowledgement:" + ack);
    }

    /**
     * Two agents are equal if their respective hash codes
     * are equal.
     *
     * @return true if obj equals this SimpleReceiver.
     *
     */
    public boolean equals( Object obj )
    {
        if ( obj == this )
            return true;

        if ( !(obj instanceof SimpleReceiver) )
            return false;

        return this.hashCode() == ((SimpleReceiver)obj).hashCode();
    }

    public int hashCode()
    {
        return _agentDescription.getEntityNameAsString() == null ? super.hashCode() : _agentDescription.getEntityNameAsString().hashCode();
    }

    //////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////
    // Entry point...

   public static void main(String [] args)
      throws Exception
    {
        String name = null;
        String transport = null;
        Vector domainNames = null;
        if (args.length > 0) {
	        int i = 0;
	        while (i < args.length) {
	            if (args[i].equalsIgnoreCase ("-name")) {
	                if ((i+1) < args.length) {
    	                name = args[i+1];
    	                i++;
    	            }
    	            else {
    	                System.out.println ("Must specify Entity Name with -name");
    	                System.exit(0);
    	            }
	            }
	            else if (args[i].equalsIgnoreCase ("-transport")) {
	                if ((i+1) < args.length) {
    	                transport = args[i+1];
    	                i++;
    	            }
    	            else {
    	                System.out.println ("Must specify  transport with -transport.  Valid transports: corba, grid, tcp.");
    	                System.exit(0);
    	            }
	            }
	            else if (args[i].equalsIgnoreCase ("-domains")) {
                    if ((i+1) < args.length) {
                        String domainsStr = args[i+1];
                        domainNames = new Vector();
                        StringTokenizer tokenizer = new StringTokenizer(domainsStr, ",");
                        while (tokenizer.hasMoreTokens()) {
                            domainNames.add(tokenizer.nextToken());
                        }
                    }
                }// else if
                 i++;
	        }// while
	    }

        if ( name == null) {
            System.out.println ("\n     *** Must specify Entity Name with -name\n");
            return;
        }

        if ( transport == null) {
            System.out.println ("\n     *** Must specify  transport with -transport.  Valid transports: corba, grid, tcp.");
            return;
        }

        if ( domainNames == null ) {
            domainNames = new Vector();
            domainNames.add("TestDomain");
            System.out.println (" <>    Warning: No Domain Names specified with -domains, using default: " + domainNames);
        }

        System.out.println ("\n     <>  Name of entity to receive messages: " + name);
	    System.out.println ("\n     <>  Requested transport:                " + transport);
	    System.out.println ("\n     <>  Requested domains:                  " + domainNames + "\n");

        Registration registration = null;
        // Obtain Registration in order to register domains.
        try {
            CSIFactory.setCurrentTransport(transport);
            registration = CSIFactory.getRegistration();
        }
        catch(Exception e){
            e.printStackTrace();
            System.out.println ("  **** Unable to obtain CSI services: Registration. Exiting...");
            return;
        }

        System.out.println ("\n     <>  Obtained registration " + registration + "\n\n");
        Enumeration domainsEnum = domainNames.elements();
        while ( domainsEnum.hasMoreElements() ) {
            String domainName  = (String) domainsEnum.nextElement();
            System.out.println("\n     <>  Will register domain: " + domainName + "\n");

            try {
                registration.registerDomain(domainName);
            }
            catch(DirectoryFailure de) {
                System.out.println("\n   Due to Directory Failure unable to register domain: " + domainName);
                de.printStackTrace();
                return;
            }
            catch(AlreadyRegisteredException ae) {
                System.out.println("    The following domain has already been registered: " + domainName);
            }
        }

        // Ask Registration to create a KAoSAgentDescription for the receiver.
        KAoSAgentDescription agentDescription = registration.createKAoSAgentDescription(name,       // use nickname as guid
                                                                                        name,       // nickname
                                                                                        domainNames,
                                                                                        transport,
                                                                                        null);      // no extra transport info necessary


        SimpleReceiver usecase = new SimpleReceiver();

        try {
            usecase.initialize(name, agentDescription, transport);
        }
        catch(Exception e) {
            e.printStackTrace();
            return;
        }

        System.out.println ("\n\n   " + name + " ready to receive messages!\n\n" );

        synchronized ( Thread.currentThread() ) {
            Thread.currentThread().wait();
        }

   }

	public boolean isListeningForEvent(DirectoryEvent event)
	{
		return true;
	}

	public void receiveDirectoryEvent(DirectoryEvent event)
	{
		System.out.println("Received DirectoryEvent " + event.getName() + "\n" + event.getDescription());
	}
}

