/**
 * Last Modified by: $Author: rjeffers $
 * On: $Date: 2006/08/23 00:56:02 $
 */
package kaos.core.csi.usecase.policy;

import kaos.core.csi.ActionInstanceDescription;
import kaos.core.csi.CSIFactory;
import kaos.core.csi.OntPropertyDescription;
import kaos.core.csi.query.QueryState;
import kaos.core.csi.request.ObligationRequestReceiverImpl;
import kaos.core.util.Logger;
import kaos.notification.NotifyAction;
import kaos.notification.NotifyRequest;
import kaos.notification.email.EmailClient;
import kaos.notification.gui.NotifyWnd;
import kaos.notification.ontologies.NotifiableHumanConcepts;
import kaos.notification.ontologies.NotificationActionConcepts;
import kaos.notification.ontologies.NotificationEntityConcepts;
import kaos.notification.ontologies.NotificationEntityInstances;
import kaos.notification.pager.PagerClient;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import javax.mail.MessagingException;

/**
 * This class is an example of an actor, performing obligations for NotificationAction 
 * defined by: http://ontology.ihmc.us/Notification/NotificationAction.owl#NotificationAction.
 * 
 * To support the functionality of ObligationActorForNotificationAction, the classe has been
 * placed in the following hierarchy:
 * 
 * KAoSActorImpl
 * 		|
 * ObligationRequestReceiverImpl
 * 		|
 * ObligationActorForNotificationAction
 *  
 * Obligation actors are required to implement the kaos.core.csi.request.ObligationRequestReceiver 
 * defining the method for receiving and procesing obligations: 
 * 
 * 		processObligationRequest(ActionInstanceDescription obligationAID, ActionInstanceDescription triggerAID)
 * 
 * The above method is called remotely by the guard as a part of the obligation policy enforcement.
 * 
 * ObligationActorForNotificationAction's superclass, ObligationRequestReceiverImpl, implements this
 * method, however, it relies on the subclasses to complete the obligation enforcement.
 * In particular, ObligationRequestReceiverImpl only initiates processing of obligations in its implementation
 * of "processObligationRequest": it accepts the obligation in the form of the ActionInstanceDescription,
 * it retrieves the name of the obligation action and then, to complete obligation enforcement, 
 * it calls a method with the action name and the following arguments:
 * 
 *  	obligationActionName(ActionInstanceDescription obligationAID, ActionInstanceDescription triggerAID)
 * 
 * In the case of ObligationActorForNotificationAction, the method is:
 * 
 * 		NotificationAction(ActionInstanceDescription obligationAID, ActionInstanceDescription triggerAID)
 * 
 * The NotificationAction() method is the starting point to the implementation of NotificationAction enforcement. 
 *
 * For messaging, ObligationActorForNotificationAction implements the javax.agent.service.transport.MessageListener, implemented
 * by  KAoSActorImpl.
 * 
 * @author rjeffers
 * @author lbunch
 *
 */
public class ObligationActorForNotificationAction 
	extends ObligationRequestReceiverImpl		 
 {	

	protected EmailClient 	_emailClient = null;
	protected PagerClient 	_pagerClient = null;
	protected Hashtable 	_sentNotifications = null;
	protected NotifyWnd 	_notifyWnd = null;
    protected QueryState	_queryState = null;

	/**
	 * The command line arguments required to bootstrap an instance of ObligationActorForNotificationAction are:
	 * 
	 * -name NotifyAgent -guid  NotifyAgent  -transport tcp -domains NotifyDomain 
	 * 
	 * The -name keyword is followed by the agent's nickname, NotifyAgent. Required.
	 * 
	 * The -guid keyword is followed by the agent's guid, NotifyAgent. Ooptional. In tis case, the nickname
	 * and the guid are the same to allow re-using the same policy snapshot.  In most cases
	 * the guid would be assigned by the registration service in KAoS.
	 * 
	 * The -transport keyword is followed by the name of the requested transport. Optional, the default is "tcp".
	 * 
	 * The -domains keyword is followed by the name of the domain to register the NotifyAgent in. Required.
	 * 
	 * This constructor instantiates a new KAoS-aware ObligationActorForNotificationAction
	 * using JAS services for transport, naming and agent directory. During the bootstrapping steps the instance:
	 *  	- obtains a reference to CSI Registration
	 *    	- binds itself to CSI Transport
	 *   	- asks CSI Registration to create a KAoSAgentDescription for itself
	 * 
	 * @param args - containing the arguments described above.
	 * 
	 * @throws Exception if some of the required arguments are missing or KAoS Registration or transport
	 * 					 services are not available.
	 */
	public ObligationActorForNotificationAction(String[] args) throws Exception {
		super(args);
	}
	
	/**
	 * In this method, this obligation actor continues its policy- and application-specific instantiation:
	 *		- specifies ontological type of a NotificationAgent in its KAoSAgentDescription
	 *      - registers with KAoS
	 *      - initializes its vars
	 *      - creates its gui
	 *      - obtains a reference the CSI Query Service needed to obtain addresses to
	 *        send notifications via email
	 *    	- initializes the notification modes
	 *           
	 * @throws Exception if the connection to the Directory Service cannot be established, or
	 * 					 if the registering instance has been already registered in the Directory Service, or
	 * 					 if the QueryState service cannot be obtained.
	 */
	public void start()
		throws Exception
    {	
	    boolean emailFailed = true;
	    boolean pagerFailed = true;
	    
		_actorDesc.setEntityOntologicalType(NotificationEntityConcepts.NotificationAgent());
				
		registerWithKAoS();
				
		_sentNotifications = new Hashtable();
		_notifyWnd = new NotifyWnd();
		
		try	{
			_notifyWnd.setStatus("Registered with KAoS...");
			_queryState = CSIFactory.getQueryState();					
		}
		catch(Exception e)	{
			_notifyWnd.setStatus("Obtaining Query service failed!");
			logMessage("\n\n\t*** Failed to obtain Query service. ", e, Logger.WARN);
			removeAgent();
			throw(e);
		}
							  
		_notifyWnd.setStatus("Initializing notification modes...");

		try {
			_emailClient = new EmailClient();
			emailFailed = false;
		}
		catch (Exception e) {
			logMessage("\n\n\t***Error starting email mode.", e, Logger.INFO);
		}
							
		try {
			_pagerClient = new PagerClient();
			pagerFailed = false;
		}
		catch(Exception e) {
			logMessage("\n\n\t***Error starting pager mode.", e, Logger.INFO);
		}

		if ( emailFailed && pagerFailed ) {
			_notifyWnd.setStatus("Notification modes failed to initialize.");
			logMessage("\n\n\t***Notification modes failed to initialize!", Logger.WARN);
			removeAgent();
			throw new Exception("***Notification modes failed to initialized.  Terminating...");
		}
			
		_notifyWnd.setStatus("Ready.");				
	}
	 
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	///
	///  METHODS SUPPORTING OBLIGATION POLICY ENFORCEMENT 
	///
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * This method is called by the superclass, ObligationRequestReceiverImpl, in order
	 * to complete the enforcement of an obligation, which in this case is the NotificationAction.
	 * 
	 * @param       obligationAID		 	is an ActionInstanceDescription containing the NotificationAction
	 * 										and its properties.
	 * 
	 * @param		triggerAID	            is an ActionInstanceDescription containing the trigger action name
	 * 										and its properties. The trigger action provides more information about the context,
	 *  									in which the trigger action has occurred.
	 */
	public void NotificationAction(ActionInstanceDescription obligationAID, ActionInstanceDescription triggerAID)
	{		
		notify(new NotifyRequest(triggerAID,obligationAID));
	}
	 
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	///
	///  METHODS IMPLEMENTING NOTIFICATIONS
	///
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////
	 /**
	  * Entry point to the notification process.
	  * 
	  * @param nReq NotifyRequest containing ActionInstanceDescriptions for trigger action and obligation action.
	  * 
	  */
	 private void notify(NotifyRequest nReq)
	 {
		 if ( nReq == null ) return;
			
		 logMessage("notify: received\n\n\tNotifyRequest:\n" + nReq, Logger.INFO);
			
		 ActionInstanceDescription notifyAID = nReq.get_notifyAction();
		 
		 if ( notifyAID == null ) {
			 logMessage("notify: AID for notify Action is null! No notifications will be sent...", Logger.INFO);
			 return;
		 }
				
		 logMessage("notify: obligation AID: \n\tNotification Action:" + notifyAID, Logger.INFO);
			
		 NotifyAction na = new NotifyAction(notifyAID, nReq);
			
		 logMessage("notify: NotifyAction\n\t" + na, Logger.INFO);
		
		 sendNotification(na);						
	 }
			
	 /**
	  * Sends a notification message to a user based on the properties of the given NotifyAction.
	  * The NotifyAction contains the ActionInstanceDescription for a NotificationAction.
	  * 
	  * @param nAction
	  */
	 private void sendNotification(NotifyAction nAction)
	 {	
		 logMessage("sendNotification: got: " + nAction, Logger.INFO);
		 OntPropertyDescription recipientProp = nAction._notifyActionID.getProperty(NotificationActionConcepts.hasRecipientRestriction());
		 OntPropertyDescription modeProp = nAction._notifyActionID.getProperty(NotificationActionConcepts.hasNotificationMode());
		 OntPropertyDescription focusProp = nAction._notifyActionID.getProperty(NotificationActionConcepts.hasFocusOfAttention());
	
		 logMessage("sendNotification: \nrecipientProp:\n\t" + recipientProp
	        		+ "\nmodeProp:\n\t" + modeProp +  "\nfocusProp:\n\t" + focusProp, Logger.INFO);
	        	        
		 if ( modeProp != null ) {	
		    // extract notification mode from policy
			 Set modeSet = modeProp.getValues();
			 Iterator mIter = modeSet.iterator();
			 if (mIter.hasNext())
				 nAction._mode = (String)mIter.next();
		 }

		 // extract focus of attention from policy
		 if ( focusProp != null ) {
			 Set foaSet = focusProp.getValues();
			 Iterator fIter = foaSet.iterator();
			 if ( fIter.hasNext() )	{
				 nAction._focusOfAttention = (String)fIter.next();
			 }
		 }
			
		 if ( recipientProp != null )	{
			 // extract new recipient info from policy
			 Set recipientSet = recipientProp.getValues();
			 Iterator rIter = recipientSet.iterator();
			 if ( rIter.hasNext() )	{
				 String recipientStr = (String)rIter.next();
				 // if a role was specified, this will be the first instance in that role
				 nAction._recipientID = recipientStr;
			 }
		 }
		 
		 if ( nAction._recipientID != null )	{
			 // lookup the recipient's address
			 nAction._recipientAddress = getRecipientAddress(nAction._recipientID, nAction._mode);
		 }

		 boolean priamryFocus = false;
		 if ( nAction._focusOfAttention != null && nAction._focusOfAttention.equals(NotificationEntityInstances.ShiftToPrimary()) )
			 priamryFocus = true;

		 try {
			 if ( nAction._mode.indexOf("Email") > 0 ) {
	         	_emailClient.send(nAction._recipientAddress, nAction.get_subject(false), nAction.get_body(false), priamryFocus);
	         	_notifyWnd.addSentNotification("Email", nAction._recipientAddress, nAction.get_subject(false));
	         }
	         else if ( nAction._mode.indexOf("TextPager") > 0 ) {
	         	_pagerClient.send(nAction._recipientAddress, nAction.get_subject(true), nAction.get_body(true), priamryFocus);
	         	_notifyWnd.addSentNotification("Text Pager", nAction._recipientAddress, nAction.get_subject(false));
	         }
	      }
	      catch (MessagingException e) {
	         e.printStackTrace();
	      }			
	 }
	 
	/**
	 * Gets the recipient's address (e.g. lbunch@ihmc.us) for the given mode(e.g. email)
	 * 
	 * @param recipientID - daml string containing Human instance
	 * 
	 * @param mode - daml string containing Notification Mode instance
	 * 
	 * @return recipient's address for the given mode
	 */
	 private String getRecipientAddress(String recipientID, String mode)
	 {
		 String recipientAddr = recipientID;
		 try {
			 String recipientAddrPropertyStr = null;
			 if (mode.indexOf("Email") > 0)
				 recipientAddrPropertyStr = NotifiableHumanConcepts.hasEmailAddress();
			 else if (mode.indexOf("InstantMessage") > 0)
				 recipientAddrPropertyStr = NotifiableHumanConcepts.hasInstantMessageAddress();
			 else if (mode.indexOf("TextPager") > 0)
				 recipientAddrPropertyStr = NotifiableHumanConcepts.hasTextPagerAddress();
						
			 recipientAddr = _queryState.getUniqueValueForProperty(recipientAddrPropertyStr, recipientID);					
		} 
		 catch (Exception e1) {
			 e1.printStackTrace();
		}
		return recipientAddr;
	 }
		
		
	 /**
	  * 
	  * Called to perform cleanup activities before the agent is terminated.	
	  * 
	  */
	 private void removeAgent()
	 {
		 logMessage("Shutdown requested. Deregistering from KAoS, disconnecting the service.", Logger.INFO);
		 
		 try {
			 deregisterFromKAoS();
		 }
		 catch(Exception e){
			 e.printStackTrace();		 
		 }
	 }
			 
	/**
	 * The command line arguments required to instantiate the ObligationActorForNotificationAction are:
	 * 
	 * -name NotifyAgent -guid  NotifyAgent  -transport tcp -domains NotifyDomain 
	 * 
	 * The -name keyword is followed by the agent's nickname, NotifyAgent - required.
	 * 
	 * The -guid keyword is followed by the agent's guid, NotifyAgent - optional. In tis case, the nickname
	 * and the guid are the same to allow re-using the same policy snapshot.  In most cases
	 * the guid would be assigned by the registration service in KAoS.
	 * 
	 * The -transport keyword is followed by the name of the requested transport - optional, the default is "tcp".
	 * 
	 * The -domains keyword is followed by the name of the domain to register the NotifyAgent in - required.
	 * 
	 * @param args - described above
	 */
	public static void main(String[] args) {

		ObligationActorForNotificationAction obligationActor = null;
		
		try {
			obligationActor = new ObligationActorForNotificationAction(args);
			obligationActor.start();	  
		}
		catch(Exception e){
			e.printStackTrace();
			return;		
		}
	}

}
