package kaos.metrics;

import java.text.DateFormat;
import java.util.GregorianCalendar;
//import org.apache.log4j.Category;
 
/**
 * From: http://techrepublic.com.com/html/tr/sidebars/1045673-3.html
 *
 * Represents an individual metric. Each metric stores a long value or
 * date value (which is a long).  The metrics are kept and managed by
 * the Metric Manager.
 *
 * @see MetricManager
 */
public class Metric {
   /**
    * Identifies the generic TimeStamp metric.  The TimeStamp metric
    * is stored as a Gregorian Calendar in UTC time.
    */
   public static final int TIME_STAMP_METRIC = 1;
   /**
    * Identifies the generic Elapsed Time metric.  The Elapsed
    * Time metric is a TimeStamp metric that returns the number of milliseconds
    * elapsed since the TimeStamp.
    */
   public static final int ELAPSED_TIME_METRIC = 2;
   /**
    * Identifies the generic Indicator metric.  An Indicator metric
    * is just a numeric total.
    */
   public static final int INDICATOR_METRIC = 3;
   /**
    * Identifies the generic Static metric. A Static metric is a
    * numeric value that replaces existing values with the new values.
    */
   public static final int STATIC_METRIC = 4;
   /**
    * Identifies the generic Average metric. An Average metric is a
    * numeric value that also keeps a count of the number of datapoints and returns
    * the average as a float.
    */
   public static final int AVERAGE_METRIC = 5;
  
//   static Category log = Category.getInstance(Metric.class.getName());
  
   private int mMetricType = INDICATOR_METRIC;
   private String mMetricId;
   private String mMetricName;
   private boolean mCanReset = false;
   private String mMetricDescription;
   private long mMetricValue = 0;
   private long mMetricDataPoints = 0;
   private GregorianCalendar mMetricDate;
   private long historicHigh = Long.MIN_VALUE;
   private long historicLow = Long.MAX_VALUE;
 
 
   /**
    * Construct a default Metric object. Default type is Indicator.
    */
   public Metric() {
//      log.debug("Metric created: Defaults used.");
   }
  
   /**
    * Construct a default Metric object using the specified values.
    *
    * @param      pMetricType                          Metric type number.
    * @param      pMetricId                     The metric id used to find this metric.
    * @param      pMetricName                          Descriptive name of metric.
    * @param         pCanReset                         True or False to control whether value can be reset.
    * @param      pMetricDescription     Descriptive text in addition to the name.
    */
   public Metric(int pMetricType, String pMetricId, String pMetricName,
                  boolean pCanReset, String pMetricDescription) {
      mMetricId = pMetricId;
      mMetricName = pMetricName;
      mCanReset = pCanReset;
      mMetricDescription = pMetricDescription;
      // Validate the metric type before setting.  Todo: this could be more object oriented.
      if ( pMetricType > 0 && pMetricType < 6 ) {
         mMetricType = pMetricType;
      }              
//      log.debug("Metric created: Type=" + mMetricType + "; Id=" + mMetricId
//                  + "; Name=" + mMetricName + ".");
   }
  
   /**
    * @return     Metric id as a string.
    */
   public String getId() {
      return mMetricId;
   }
  
   /**
    * @return     Name used for this Metric.
    */
   public String getName() {
      return mMetricName;
   }
  
   /**
    * @return        Description for this Metric.
    */
   public String getDescription() {
      return mMetricDescription;
   }
  
   /**
    * @return        int type for this Metric; should match one of the _METRIC constants.
    */
   public int getType() {
      return mMetricType;
   }
  
   /**
    * @return              true if metric can be reset, false if it cannot be reset.
    */
   public boolean getCanReset() {
              return mCanReset;
       }
      
   /**
    * Add method increases or sets a metrics value.
    *
    * @param      pMetricValue      value to be added to the current Metric.
    */
   public void add(int pMetricValue) {
      if ( mMetricType == STATIC_METRIC ) {
         mMetricValue = pMetricValue;
      }
      else {
         mMetricValue += pMetricValue;
      }
      mMetricDataPoints++;
   }
  
   /**
    * Add method increases or sets a metrics value.
    *
    * @param      pMetricValue      value to be added to the current Metric.
    */
   public void add(long pMetricValue) {
      if(pMetricValue < historicLow)
         historicLow = pMetricValue;
      if(pMetricValue > historicHigh)
         historicHigh = pMetricValue;

      if ( mMetricType == STATIC_METRIC ) {
         mMetricValue = pMetricValue;
      }
      else {
         mMetricValue += pMetricValue;
      }
      mMetricDataPoints++;
   }
  
   /**
    * Add method increases or sets a metrics value.
    *
    * @param      pMetricValue      date value to be applied to the current Metric.
    */
   public void add(java.util.Date pMetricValue) {
      mMetricDate = new GregorianCalendar();
      mMetricDate.setTime(pMetricValue);
   }
  
   /**
    * Add method increases or sets a metrics value.
    *
    * @param      pMetricValue      calendar value to be applied to the current Metric.
    */
   public void add(GregorianCalendar pMetricValue) {
      mMetricDate = pMetricValue;
   }
  
   /**
    * Returns the metric's value.
    *
    * @return     Value of the metric, 0 is returned if the metric is not an
    *             indicator or static metric type.
    */
   public long getValue() {
      switch(mMetricType) {
         case INDICATOR_METRIC:
         case STATIC_METRIC:
            return mMetricValue;
         case TIME_STAMP_METRIC:
            if ( mMetricDate == null ) {
               return 0;
            }
            else {
               return mMetricDate.getTime().getTime();
            }
         case ELAPSED_TIME_METRIC:
           if ( mMetricDate == null ) {
               return 0;
            }
            else {
               return new GregorianCalendar().getTime().getTime() - mMetricDate.getTime().getTime();
            }
         case AVERAGE_METRIC:
            if ( mMetricDataPoints > 0 ) {
               return (long)mMetricValue / mMetricDataPoints;
            }
            else {
               return 0;
            }
         default:
            return 0;
      }
   }
  
   /**
    * Returns the metric's average value.
    *
    * @return     Value of the metric, 0 is returned if the metric is not an
    *             indicator or static metric type.
    */
   public float getAverage() {
      switch(mMetricType) {
         case INDICATOR_METRIC:
         case STATIC_METRIC:
         case AVERAGE_METRIC:
            if ( mMetricDataPoints > 0 ) {
               return (float)mMetricValue / mMetricDataPoints;
            }
            else {
               return 0;
            }
         case TIME_STAMP_METRIC:
         case ELAPSED_TIME_METRIC:
         default:
            return 0;
      }
   }

   public long getHistoricLow()
   {
      switch(mMetricType)
      {
         case AVERAGE_METRIC:
            return historicLow;
         default:
            return 0;
      }
   }
        
   public long getHistoricHigh()
   {
      switch(mMetricType)
      {
         case AVERAGE_METRIC:
            return historicHigh;
         default:
            return 0;
      }
   }

   /**
    * Returns the metric's TimeStamp value.
    *
    * @return     The Date of the metric or null.
    */
   public GregorianCalendar getTimeStamp() {
      switch(mMetricType) {
         case TIME_STAMP_METRIC:
            return mMetricDate;
         case ELAPSED_TIME_METRIC:
         case INDICATOR_METRIC:
         case STATIC_METRIC:
         case AVERAGE_METRIC:
         default:
            return null;
      }
   }
  
   /**
    * Returns the metric's count of data points.
    *
    * @return     Count of datapoints added to this metric.
    */
   public long getDataPointCount() {
      return mMetricDataPoints;
   }
  
   /**
    * The Metric value is returned as a string for logs or screen output.
    *
    * @return     String representation of the metric value.
    */
   public String toString() {
      StringBuffer stringValue = new StringBuffer();
      switch(mMetricType) {
         case TIME_STAMP_METRIC:
            if ( mMetricDate == null ) {
               stringValue.append("null");
            }
            else {
               DateFormat sd = DateFormat.getDateTimeInstance(DateFormat.LONG, DateFormat.LONG);
              sd.setCalendar(mMetricDate);
               stringValue.append(sd.format(mMetricDate.getTime()));
            }
            break;
         case ELAPSED_TIME_METRIC:
            if ( mMetricDate == null ) {
               stringValue.append("null");
            }
            else {
               stringValue.append(new GregorianCalendar().getTime().getTime() - mMetricDate.getTime().getTime());
            }
            break;
         case INDICATOR_METRIC:
         case STATIC_METRIC:
            stringValue.append(mMetricValue);
            break;
         case AVERAGE_METRIC:
             stringValue.append("avg ");
            if ( mMetricDataPoints == 0 ) {
               stringValue.append("0");
            }
            else {
               stringValue.append((float)mMetricValue / mMetricDataPoints);
               stringValue.append(" (" + mMetricDataPoints + " total)");
            }
            break;
         default:
            stringValue.append("n/a");
            break;
      }
      return stringValue.toString();
   }
  
   /**
    * Resets all values to zero or null.
    */
   public void reset() {
              if ( mCanReset ) {
     mMetricValue = 0;
     mMetricDataPoints = 0;
     historicHigh = Long.MIN_VALUE;
     historicLow = Long.MAX_VALUE;
     mMetricDate = null;
              }
   }
}
