package kaos.metrics;

import kaos.core.util.KAoSConfigurationParameters;
import kaos.core.util.Logger;

import java.io.FileInputStream;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;
//import org.apache.log4j.Category;
 
/**
 * This class is a singleton that provides tracking and logging of
 * metrics related to the listener.  The metrics are tracked by this
 * class and logged at intervals defined in the properties file.  The
 * available metrics are defined in the properties file.
 *
 * //@see com.telemics.listener.Listener
 */
public class MetricManager implements TimerCallBack {
   private static MetricManager mInstance; // The single instance
//   static Category log = Category.getInstance(MetricManager.class.getName());
   private Hashtable mMetrics = new Hashtable();
   private Hashtable localMetrics = new Hashtable();
   private long mTimerInterval = 60000;
   private String mHost;
   private TaskTimer mTimer;

   private final static Logger _logger = Logger.getLogger(MetricManager.class.getName());
   /**
    * Returns the single instance, creating one if it's the
    * first time this method is called.
    *
    * @return      The single instance of MetricManager.
    */
   static synchronized public MetricManager getInstance() {
      if ( mInstance == null ) {
         mInstance = new MetricManager();
      }
      return mInstance;
   }

   private boolean _resetAfterLog;
  
   /**
   * A private constructor since this is a Singleton.  The constructor reads
   * the properties file and sets up the defaults.  The Metrics to be collected
   * are setup from data in the properties file as well.
   */
   private MetricManager() {
      String filename = System.getProperty(KAoSConfigurationParameters.METRICS_CONFIG_FILE);
      if (filename == null) {
          _logger.logMessage(KAoSConfigurationParameters.METRICS_CONFIG_FILE + " not set, no metrics will be logged", Logger.DEBUG);
          return;
      }

      Properties pProps = new Properties();
      try {
         FileInputStream is = new FileInputStream(filename);
         pProps.load(is);
         is.close();
      }
      catch (Exception e) {
         _logger.logMessage("Unable to read " + filename + ".properties, using defaults.", Logger.ERROR);
         return;
      }
      
      boolean enabled = Boolean.parseBoolean(pProps.getProperty("enabled"));
      if (!enabled) {
          _logger.logMessage("metrics disabled, no metrics will be logged", Logger.INFO);
          return;
      }
      
      try {
         mHost = pProps.getProperty("host", InetAddress.getLocalHost().getHostName());
      }
      catch(UnknownHostException ue) {
         _logger.logMessage("Unable to retrieve host name of local machine, using 'unknown'", ue, Logger.ERROR);
      }
      try {
         mTimerInterval = Long.parseLong(pProps.getProperty("logInterval", "60000"));
      }
      catch(Exception e) {
         _logger.logMessage("Unable to retrieve logInterval property, using default.", Logger.ERROR);
     }
      _resetAfterLog = Boolean.parseBoolean(pProps.getProperty("resetAfterLog"));
      loadMetrics(pProps);
     
      mTimer = new TaskTimer(this, "Metric Timer", mTimerInterval);
     
      mTimer.start();
     
      _logger.logMessage("MetricManager started: Log Interval is " + mTimerInterval / 1000 + ".", Logger.DEBUG);
   }
   
   /**
    * Loads and registers all the metrics.
    *
    * @param     pProps      The properties for MetricManager.
    */
   private void loadMetrics(Properties pProps) {
      String metricId;
      int metricType = 0;
      String metricName;
      boolean metricReset = true;
      String metricDescription;
     
      String metricsProperty = pProps.getProperty("metrics");
      StringTokenizer metricIds = new StringTokenizer(metricsProperty, ",");
 
      while (metricIds.hasMoreElements()) {
         metricId = metricIds.nextToken().trim();
         try {
            metricType = Integer.parseInt(pProps.getProperty(metricId + ".type", String.valueOf(Metric.INDICATOR_METRIC)));
         }
         catch (NumberFormatException ne) {
            metricType = Metric.INDICATOR_METRIC;
         }
         metricName = pProps.getProperty(metricId + ".name", "");
         metricReset = Boolean.valueOf(pProps.getProperty(metricId + ".reset", "true")).booleanValue();
         metricDescription = pProps.getProperty(metricId + ".desc", "");
         try {
            mMetrics.put(metricId, new Metric(metricType, metricId, metricName,
                                                                                 metricReset, metricDescription));
            _logger.logMessage("Registered Metric " + metricId + ".", Logger.DEBUG);
         }
        catch (Exception propsError) {
            _logger.logMessage("Can't register Metric: " + metricId, propsError, Logger.ERROR);
         }
      }
   }
  
   /**
    * Allows a client to explicity shutdown the metric manager
    * causing all metrics to be logged, the timer thread to be
    * stopped, and the internal reference for the singleton pattern
    * to be cleared.
    */
   public void shutdown() {
      mTimer.killTimer();
      mTimer.interrupt();
      mTimer = null;
      logMetrics();
      mMetrics.clear();
      mInstance = null;
   }
 
   /**
    * Called during each expiration of the timer in order to log current
    * metric totals.
    */
   public void timerExpired() {
      this.logMetrics();
   }
  
   /**
    * Adds to a metrics value, accounting for the type of metric.
    *
    * @param      pMetricId      String id of the metric to add value to.
    * @param      pMetricValue   Value to be added to this metric.
    */
   public void add(String pMetricId, int pMetricValue ) {
      Metric toAdd = (Metric)mMetrics.get(pMetricId);
      if ( toAdd == null ) {
//                   log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
      }
      else {
         toAdd.add(pMetricValue);
      }
      synchronized (_localMetrics) {
          for (TaggedMetrics localMetrics : _localMetrics.values()) {
             toAdd = (Metric)localMetrics._metrics.get(pMetricId);
             if ( toAdd == null ) {
                 //log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
             }
             else {
                 toAdd.add(pMetricValue);
             }
          }
       }
   }
 
   /**
    * Adds to a metrics value, accounting for the type of metric.
    *
    * @param      pMetricId      String id of the metric to add value to.
    * @param      pMetricValue   Value to be added to this metric.
    */  
   public void add(String pMetricId, long pMetricValue) {
      Metric toAdd = (Metric)mMetrics.get(pMetricId);
      if ( toAdd == null ) {
//                   log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
              }
              else {
         toAdd.add(pMetricValue);
      }
      synchronized (_localMetrics) {
          for (TaggedMetrics localMetrics : _localMetrics.values()) {
             toAdd = (Metric)localMetrics._metrics.get(pMetricId);
              if ( toAdd == null ) {
    //                       log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
                      }
                      else {
                toAdd.add(pMetricValue);
            }
           }
       }
   }
 
   /**
   * Adds to a metrics value, accounting for the type of metric.
    *
    * @param      pMetricId      String id of the metric to add value to.
    * @param      pMetricValue   Value to be added to this metric.
    */  
   public void add(String pMetricId, java.util.Date pMetricValue) {
      Metric toAdd = (Metric)mMetrics.get(pMetricId);
      if ( toAdd == null ) {
//                   log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
              }
              else {
         toAdd.add(pMetricValue);
      }
      synchronized (_localMetrics) {
          for (TaggedMetrics localMetrics : _localMetrics.values()) {
             toAdd = (Metric)localMetrics._metrics.get(pMetricId);
              if ( toAdd == null ) {
    //                       log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
                      }
                      else {
                toAdd.add(pMetricValue);
            }
           }
       }
   }
 
   /**
    * Adds to a metrics value, accounting for the type of metric.
    *
    * @param      pMetricId      String id of the metric to add value to.
    * @param      pMetricValue   Value to be added to this metric.
    */  
   public void add(String pMetricId, GregorianCalendar pMetricValue) {
      Metric toAdd = (Metric)mMetrics.get(pMetricId);
      if ( toAdd == null ) {
//                   log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
              }
              else {
         toAdd.add(pMetricValue);
      }
      synchronized (_localMetrics) {
          for (TaggedMetrics localMetrics : _localMetrics.values()) {
             toAdd = (Metric)localMetrics._metrics.get(pMetricId);
              if ( toAdd == null ) {
    //                       log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
                      }
                      else {
                toAdd.add(pMetricValue);
            }
           }
       }
   }
 
   /**
    * Returns the specified metric's value.
    *
    * @param      pMetricId      String id of the metric to add value to.
    *
    * @return     Value of the metric, 0 is returned if the metric is not an
   *             indicator or static metric type.
    */  
   public long getValue(String pMetricId) {
      Metric toRetrieve = (Metric)mMetrics.get(pMetricId);
      if ( toRetrieve == null ) {
//                   log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
                     return 0;
              }
              else {
         return toRetrieve.getValue();
      }
   }
  
   /**
    * Returns the specified metric's average value.
    *
    * @param      pMetricId      String id of the metric to retrieve.
    *
    * @return     Value of the metric, 0 is returned if the metric is not an
    *             indicator or static metric type.
    */
   public float getAverage(String pMetricId) {
      Metric toRetrieve = (Metric)mMetrics.get(pMetricId);
      if ( toRetrieve == null ) {
//                   log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
                     return 0;
              }
              else {
         return toRetrieve.getAverage();
      }
   }
 
   /**
    * Returns the specified metric's TimeStamp value.
   *
    * @param      pMetricId      String id of the metric to retrieve.
    *
    * @return     The Date of the metric or null.
    */        
   public GregorianCalendar getTimeStamp(String pMetricId) {
      Metric toRetrieve = (Metric)mMetrics.get(pMetricId);
      if ( toRetrieve == null ) {
//                   log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
                     return null;
              }
              else {
         return toRetrieve.getTimeStamp();
      }
   }
 
   /**
    * Returns the specified metric's count of data points.
    *
    * @param      pMetricId      String id of the metric to retrieve.
    *
    * @return     Count of datapoints added to this metric.
    */  
   public long getDataPointCount(String pMetricId) {
      Metric toRetrieve = (Metric)mMetrics.get(pMetricId);
      if ( toRetrieve == null ) {
//                   log.warn("Metric Id (" + pMetricId + " does not exist.", new InvalidMetricException());
                     return 0;
              }
              else {
         return toRetrieve.getDataPointCount();
      }
   }
  
   /**
    * Logs all the metrics to the file.  This routine is called by another
    * thread that is operating on a timer.  The thread controls when this
    * method is called.
    */
   public void logMetrics() {
      Metric toLog;
      _logger.logMessage("*** Logging Metrics ***", Logger.INFO);
      _logger.logMessage(mHost + " -> Total Memory : " + Runtime.getRuntime().totalMemory() + " bytes.", Logger.INFO);
      _logger.logMessage(mHost + " -> Free Memory : " + Runtime.getRuntime().freeMemory() + " bytes.", Logger.INFO);
      for ( Enumeration allMetrics = mMetrics.elements() ; allMetrics.hasMoreElements() ;) {
         toLog = (Metric)allMetrics.nextElement();
         _logger.logMessage(mHost + " -> " + toLog.getName() + " : " + toLog.toString(), Logger.INFO);
      }

      ArrayList<TaggedMetrics> taggedMetricsCopy;
      synchronized (_taggedMetrics) {
        taggedMetricsCopy = new ArrayList(_taggedMetrics);
        _taggedMetrics.clear();
      }
      for (TaggedMetrics taggedMetrics : taggedMetricsCopy) {
          _logger.logMessage("LOCAL METRICS: " + taggedMetrics._tag, Logger.INFO);
          for ( Enumeration allMetrics = taggedMetrics._metrics.elements() ; allMetrics.hasMoreElements() ;) {
             toLog = (Metric)allMetrics.nextElement();
             _logger.logMessage(mHost + " -> " + toLog.getName() + " : " + toLog.toString(), Logger.INFO);
          }
          _logger.logMessage("LOCAL METRICS END: " + taggedMetrics._tag, Logger.INFO);
      }
      _logger.logMessage(mHost + " all metrics logged.", Logger.DEBUG);
      if (_resetAfterLog) {
          reset();
      }
   }


   /**
    * Logs specified text to the file.  For example, adding a comment to the file
    */
   public void logString(String text) {
      _logger.logMessage(text, Logger.INFO);
   }

   /**
    * Logs specified value to the file.  For example, printing out a certain variable
    */
   public void logMetric(String metricName) {
      for ( Enumeration allMetrics = mMetrics.elements() ; allMetrics.hasMoreElements() ;) {
         Metric metric = (Metric)allMetrics.nextElement();
         if(metric.getName().equals(metricName))
         {
            _logger.logMessage(metricName + " = " + metric.toString(), Logger.INFO);
            if(metric.getType() == Metric.AVERAGE_METRIC && metric.getDataPointCount() > 1)
            {
               _logger.logMessage(metricName + "high = " + metric.getHistoricHigh(), Logger.INFO);
               _logger.logMessage(metricName + "low = " + metric.getHistoricLow(), Logger.INFO);                                 
            }
            break;
         }
      }
   }

   // start logging local metrics
   public void startLocalMetrics (String localMetricId, String note, String[] metricsToLog)
   {
       Hashtable metricsTable = new Hashtable();
       for (int i=0; i<metricsToLog.length; i++) {
           Metric m = (Metric) mMetrics.get(metricsToLog[i]);
           if (m == null) continue;
           metricsTable.put(metricsToLog[i], new Metric(m.getType(), m.getId(), m.getName(), m.getCanReset(), m.getDescription()));
       }
       synchronized (_localMetrics) {
           _localMetrics.put(localMetricId, new TaggedMetrics(note, metricsTable));
       }
   }

   public void stopLocalMetrics (String localMetricId)
   {
       TaggedMetrics tm;
       synchronized (_localMetrics) {
           tm = _localMetrics.remove(localMetricId);
       }
       if (tm != null) {
           synchronized (_taggedMetrics) {
            _taggedMetrics.add(tm);
           }
       }
   }

   private Hashtable<String,TaggedMetrics> _localMetrics = new Hashtable();

   private ArrayList<TaggedMetrics> _taggedMetrics = new ArrayList();

   public static class TaggedMetrics
   {
       public String _tag;
       public Hashtable _metrics;

       public TaggedMetrics (String tag, Hashtable metrics)
       {
           _tag = tag;
           _metrics = metrics;
       }
   }

   /**
    * Resets all metrics to zero or null 
    */
   public void reset() {
      Metric toReset;
      //this.logMetrics();  // log all the metrics
      for ( Enumeration allMetrics = mMetrics.elements() ; allMetrics.hasMoreElements() ;) {
         toReset = (Metric)allMetrics.nextElement();
         toReset.reset();
      }
      resetLocalMetrics();
      _logger.logMessage(mHost + " reset all metrics.", Logger.DEBUG);
   }
  
   public void resetLocalMetrics() {
      Metric toReset;
      //this.logMetrics();  // log all the metrics
      for ( Enumeration allMetrics = localMetrics.elements() ; allMetrics.hasMoreElements() ;) {
         toReset = (Metric)allMetrics.nextElement();
         toReset.reset();
      }
      _logger.logMessage(mHost + " reset local metrics.", Logger.DEBUG);
   }

   /**
    * Resets a metric to zero or null
    *
    * @param      @pMetricId     Metric id to reset
    */
   public void reset(String pMetricId) {
      Metric toReset = (Metric)mMetrics.get(pMetricId);
      if ( toReset != null ) {
         toReset.reset();
      }
      _logger.logMessage(mHost + " reset Metric: " + pMetricId + ".", Logger.DEBUG);
   }
}
 
