package kaos.metrics;

import java.util.ListIterator;
import java.util.Vector;
//import org.apache.log4j.Category;
 
/**
 * This is a generic timer used by other classes.  An instance of TaskTimer
 * is given a callback (any object implementing the TimerCallBack interface)
 * which is called everytime the timer expires.  The length of the timer is
 * a millisecond value set in the constructor or directly.  The timer uses
 * a separate thread for execution by extending the default Thread.
 *
 * @see  TimerCallBack
 */
public class TaskTimer extends Thread {
//   static Category log = Category.getInstance(TaskTimer.class.getName());
   private Vector mTimerCallBackList = new Vector();
   private long mTimerInterval = 60000;  // milliseconds between calls.
   private boolean mKeepRunning = true;
   private String mThreadName;
 
   /**
    * The minimum allowed value of the timer interval.  The timer must call
    * each call back object each interval of the timer and therefore must have
    * enough time to make the appropriate calls.  Since the timer value is milliseconds
    * a minimum value assures the timer doesn't spend all it's time making
    * calls to the TimerCallBack objects.
    */
   public static final long MINIMUM_TIMER_INTERVAL = 5000;
  
   /**
    * Default constructor creates a TaskTimer with not established call back objects.
    */
   public TaskTimer() {
   }
  
   /**
    * Construct a TaskTimer object with the specified timer setting.
    *
    * @param      pTimerInterval     millisecond interval value.
    */
   public TaskTimer(long pTimerInterval) {
      this.setTimerInterval(pTimerInterval);
   }
  
   /**
    * Construct a TaskTimer object with the specified timer name and timer setting.
    *
    * @param      pThreadName       name for this thread.
    * @param      pTimerInterval     millisecond interval value.
    */
   public TaskTimer(String pThreadName, long pTimerInterval) {
      this.setTimerInterval(pTimerInterval);
      mThreadName = pThreadName;
   }
  
   /**
    * Construct a TaskTimer object with the specified call back, name, and timer setting.
    *
    * @param      pCallBack         Object implementing the TimerCallBack interface that is
    *                               to be called each time the timer expires.
    * @param      pThreadName       name for this thread.
    * @param      pTimerInterval    millisecond interval value.
    */
   public TaskTimer(TimerCallBack pCallBack, String pThreadName, long pTimerInterval) {
      this.setTimerInterval(pTimerInterval);
      mThreadName = pThreadName;
      this.addTimerCallBack(pCallBack);
   }
  
   /**
    * Construct a TaskTimer object with the specified call back.
    *
    * @param      pCallBack         Object implementing the TimerCallBack interface that is
    *                               to be called each time the timer expires.
    */
   public TaskTimer(TimerCallBack pCallBack) {
      this.addTimerCallBack(pCallBack);
   }
  
   /**
    * @return     The current timer interval setting.
    */
   public long getTimerInterval() {
      return mTimerInterval;
   }
  
   /**
    * Set the timer interval.
    *
    * @param      pTimerInterval    Interval value in milliseconds, if less than
    *                               MINIMUM_TIMER_INTERVAL then timer will be set equal
    *                               to the minimum.
    */
   public void setTimerInterval(long pTimerInterval) {
      if ( pTimerInterval < MINIMUM_TIMER_INTERVAL ) {
         mTimerInterval = MINIMUM_TIMER_INTERVAL;
      }
      else {
         mTimerInterval = pTimerInterval;
      }
   }
  
   /**
    * Adds an object that will be called each iteration of the timer.
    *
    * @param      pCallBack      Object to be called each time the timer expires.
    */
   public synchronized void addTimerCallBack(TimerCallBack pCallBack) {
      mTimerCallBackList.add(pCallBack);
   }
  
   /**
   * Removes an existing timer call back object from the available call
    * back list.
    *
    * @param      pCallBack      Object to be called each time the timer expires.
    */
   public synchronized void removeTimerCallBack(TimerCallBack pCallBack) {
      mTimerCallBackList.remove(pCallBack);
   }
  
   /**
    * Main timer loop runs until shutdown command is received. The thread will sleep
    * for the designated timer interval and then call back to all registered objects.
    */
   public void run() {
              if ( mThreadName != null ) {
         Thread.currentThread().setName(mThreadName);
      }
      while ( mKeepRunning ) {
         try {
            Thread.currentThread().sleep(mTimerInterval);
            if ( mKeepRunning ) {  // Recheck whether we should be running.
               ListIterator callBackList = mTimerCallBackList.listIterator();
               while ( callBackList.hasNext() ) {
                  TimerCallBack callBackObject = (TimerCallBack) callBackList.next();
                  callBackObject.timerExpired();
               }
            }
         }
         catch (InterruptedException ie) {
            ;  // The thread was interrupted so we loop and start again.
         }
      }
   }
  
   /**
    * This stops the timer, after the current interval expires.
    */
   public void stopTimer() {
      mKeepRunning = false;
   }
  
   /**
    * Stops the timer and clears all references to call back items.
    */
   public synchronized void killTimer() {
              mKeepRunning = false;
              mTimerCallBackList.clear();
              notifyAll();
       }
}
