/**
 * Last Modified by: $Author: mvignati $
 * On: $Date: 2012/01/26 19:30:01 $
 */

package kaos.ontology.util;

import kaos.core.util.Logger;

import java.util.List;
import java.util.Set;

public interface KAoSOntClassBuilder extends KAoSOntBuilder
{
    /*
     * Get the name of the currently build class
     * @return the name of the class
     * @exception ValueNotSet if value of the class name is not known
     */
    public String getClassName() throws ValueNotSet;

    /*
     * Sets the main superclass of the constructed class
     * @param mainSuperClass the name of the superclass
     * @exception ClassNameNotSet if the class name has not been set before
     */
    public void setMainSuperClass(String mainSuperClass) throws ClassNameNotSet;
    
    /*
     * Get the main superclass of the constructed class
     * @return the name of the main superclass of the constructed class
     * @exception ValueNotSet if superclasses is not known
     */
    public String getMainSuperClass() throws ValueNotSet;
    
    /*
     * Add this class as a superclass to the class being built.
     * @param baseClassName the name of the base class
     * @exception ClassNameNotSet if the class name has not been set before
     * @exception DisjointClassWithAlreadyAdded if the provided class is disjoint with the classes already added
     */
    public void addSuperlass(String superclassName) throws ClassNameNotSet, DisjointClassWithAlreadyAdded;

    /*
     * Remove the class from the list of super classes of the built class.
     * @param superlassName the name of the base class
     * @exception ClassNotSuperClass if the class was not previously added to the list of superclasses
     */
    public void removeSuperclass(String superclassName) throws ClassNotSuperClass;

    /*
     * Get all superclasses of the built class.
     * @return the list containing names of the superclasses
     * @exception ValueNotSet if super classes are not known
     */
    public List<String> getAllSuperclasses() throws ValueNotSet;
    
    /**
     * Get names of all properties for which restrictions have been defined in this class
     * @return list of property names
     */
    public List<String> getPropertyNames();

    /**
     * Get the type of the restrictions defined on the given property 
     * @param propertyName the name of the property for which the type of the restriction has to be returned
     * @return the type of the restriction
     */
    public String getRestrictionTypeName(String propertyName);

    /*
     * Check if the range of the specified property is class based or instance based.
     * @return true if based on a class, false if based on instances
     * @exception ValueNotSet if the range is not set
     */
    public boolean isPropertyRangeBasedOnClass(String propertyName) throws ValueNotSet;

    /*
     * Set the range of the given property to the specified class. If the class is already set it will replace it with the provided one.
     * Uses restriction allValuesFromRestriction
     * @param propertyName      the name of the property
     * @param propertyClassName the name of the class
     * @exception ClassNameNotSet if the class name has not been set before
     */
    public void setPropertyRangeClass(String propertyName, String propertyClassName) throws ClassNameNotSet;

    /*
     * Set the range of the given property to the specified class. If the class is already set it will replace it with the provided one.
     * @param propertyName      the name of the property
     * @param propertyClassName the name of the class
     * @param restrictionType   type of the restriction allValuesFromRestriction
     * @exception ClassNameNotSet if the class name has not been set before
     */
    public void setPropertyRangeClass(String propertyName, String propertyClassName, String restrictionType) throws ClassNameNotSet;

    /*
     * Get the name of the parent class of the class on which the range of the given property is set.
     * @param propertyName the name of the property
     * @return the name of the parent class
     */
    public String getComplementParentClassForPropertyRange(String propertyName);

    /*
     * Get the name of the class on which the range of the given property is set.
     * @param propertyName the name of the property
     * @return the name of the class
     * @exception ValueNotSet if the class is not known
     * @exception RangeIsBasedOnInstances if the range is based on instances
     */
    public String getBasePropertyRangeClass(String propertyName) throws ValueNotSet, RangeIsBasedOnInstances;

    /*
     * Get the name of the class on which the range of the given property is set.
     * The difference between this method and getBasePropertyRangeClass() is that this method returns the name of the instance holder class even 
     * if the range is based on instances.
     * @param propertyName the name of the property
     * @return the name of the class
     * @exception ValueNotSet if the class is not known
     */
    public String getCurrentPropertyRangeClass(String propertyName) throws ValueNotSet;

    /**
     * @param propertyName
     * @param instanceName
     * @throws ClassNameNotSet
     * @throws RangeIsBasedOnAClass
     */
    public void addPropertyRangeInstance(String propertyName, String instanceName) throws ClassNameNotSet, RangeIsBasedOnAClass;

    /*
     * Add the provided instance to the class being the range of the specified property. 
     * If this is a first call to this method a new class is created if this is a subsequent call the provided instance will be added 
     * to the list of instances of the already created class.
     * @param propertyName the name of the property
     * @param instanceName the name of the instance to define a range of possible values for this property
     * @exception ClassNameNotSet if the class name has not been set before
     * @exception RangeIsBasedOnAClass if the target range was already set to class
     */
    public void addPropertyRangeInstance(String propertyName, String instanceName, String type) throws ClassNameNotSet, RangeIsBasedOnAClass;

    /*
     * Remove the provided instance from the class being the range of the specified property. If as the result the class became empty 
     * it is removed and the property range is not set.
     * @param propertyName the name of the property
     * @param instanceName the name of the instance to define a range of possible values for this property
     * @exception RangeIsBasedOnAClass if the property range was already set to class
     * @exception InstanceNotInThisRange if the instance was not previously added
     */
    public void removePropertyRangeInstance(String propertyName, String instanceName) throws RangeIsBasedOnAClass, InstanceNotInThisRange;

    /*
     * Get the instances defining the range of the specified property.
     * @param propertyName the name of the property
     * @return a String array containing names of the instances
     * @exception ValueNotSet if property range instances are not known
     * @exception RangeIsBasedOnAClass if the range is based on class
     */
    public Set<String> getPropertyRangeInstance(String propertyName) throws ValueNotSet, RangeIsBasedOnAClass;

    /*
     * Add the provided instance to the hasValue property of restriction on the specified property. 
     * If this is a first call to this method a new restriction is created, if this is a subsequent call the provided instance will be added 
     * to the restriction already created.
     * @param propertyName the name of the property
     * @param instanceName the name of the instance to define a range of possible values for this property
     * @exception ClassNameNotSet if the class name has not been set before
     */
    public void addRequiredValueOnProperty(String propertyName, String instanceName) throws ClassNameNotSet;

    /*
     * Remove the provided instance from the hasValue property of restriction on the specified property.
     * @param propertyName the name of the property
     * @param instanceName the name of the instance to define a range of possible values for this property
     * @exception ClassNameNotSet if the class name has not been set before
     * @exception ValueNotSet if the specified value is not found
     */
    public void removeRequiredValueOnProperty(String propertyName, String instanceName) throws ClassNameNotSet, ValueNotSet;

    /*
     * Make the range of the property complement of the current range.
     * @param propertyName the name of the property
     * @exception AlreadyComplement if property range was already made complement
     */
    public void makeRangeComplement(String propertyName, String propertyClassName) throws AlreadyComplement;

    /*
     * Check if the range of the property is a complement.
     * @param propertyName the name of the property
     * @return true if complement, false if not
     */
    public boolean isRangeComplement(String propertyName);

    /*
     * Prepare representation of the class to be printed to log
     */
    public void showClass(Logger log, int level);

    /*
     * Return class string in XML format
     */
    public String toString();
    
    // ---- Exceptions 
    
    public class AlreadyComplement extends Exception
    {
        private static final long serialVersionUID = 5022738655337418448L;

        public AlreadyComplement()
        {
            super();
        }

        public AlreadyComplement(String s)
        {
            super(s);
        }
    }    
}